#line 7 "/home/dodo/btstack/src/avdtp-test.u"

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <errno.h>
#include "socket.h"
#include "aread.h"
#include "awrite.h"
#include "util.h"
#include "sdp-lib.h"
#include "mp3sbc.h"

uint8_t phone_addr[6]= { 0xbc,0x1d,0x01,0x55,0x38,0x30 };

uint8_t hs_addr[6]= { 0x59,0x37,0x11,0x67,0x11,0x00 };

int endpoint_suitable(avdtp_conn_t *A, uint8_t id,
                      uint8_t *Rmin_bitpool, uint8_t *Rmax_bitpool)
{
  uint8_t oct0, oct1;
  if (avdtp_get_sbc_capabilities(A, id, &oct0, &oct1,
                                Rmin_bitpool, Rmax_bitpool))
  {
    printf("Can't get capabilities for endpoint %d\n", id);
    return 0;
  }
  return 1;
}

int get_audio_sink(avdtp_conn_t *A,uint8_t *Rid,
                   uint8_t *min_bitpool,uint8_t *max_bitpool)
{
  uint8_t *endpoints;
  uint32_t s_endpoints;  
  int i;

  if (avdtp_discover_endpoints(A, &endpoints, &s_endpoints))
  {
    printf("Could not get endpoint list.\n");
    return 1;
  }
  
  printf("Endpoints\n");
  for(i=0;i<s_endpoints;i+=2)
  {
     uint8_t id, in_use, media_type, tsep;
     id= endpoints[i] >> 2;
     in_use= (endpoints[i] >> 1) &1;
     media_type= endpoints[i+1] >> 4;
     tsep= (endpoints[i+1]>>3) &1;
     printf("  EP: %d in_use: %d media_type: %d tsep: %d\n",
                  id, in_use, media_type, tsep);

           // sink,  audio
     if (tsep==1 && media_type==0 && 
         endpoint_suitable(A, id,min_bitpool,max_bitpool))
     {
       *Rid=  id;
       return 0;
     }
  }

  return 1;
}

uint8_t max8(uint8_t a,uint8_t b) { return a>b?a:b; }
uint8_t min8(uint8_t a,uint8_t b) { return a<b?a:b; }


int main(int argc,char **argv)
{
  avdtp_conn_t *A,*B;
  uint8_t max_bitpool, min_bitpool;
  uint8_t reid;
  uint8_t bp;
  uint8_t oct0, oct1;
  int DL;


  if (argc>1) {
    DL= atoi(argv[1]);
  } else DL= 100;

  A= avdtp_open(hs_addr);

  if (A->err)
    exit(fprintf(stderr,"avdtp-test(connect): %s\n", A->errstr));

  if (get_audio_sink(A, &reid, &min_bitpool, &max_bitpool))
    exit(fprintf(stderr,"can't find an audio sink on device.\n"));

  printf("Suitable EP: %d min_bp= %d max_bp= %d\n", reid,
                       min_bitpool,
                       max_bitpool);

  // 53 is joint stereo high quality 44.100 hz
  // 35 is the same for middle quality

  if (53>=min_bitpool && 53<=max_bitpool) bp= 53;
  else if (35>=min_bitpool && 35<=max_bitpool) bp= 35;
  else bp= ((uint32_t) max_bitpool+(uint32_t) min_bitpool) /2;
  printf("Selected bitpool: %d\n", bp);

  // now we can configure our stream.

  oct0= (1<<5) | (1<<0) ; // 44100 Hz, Joint Stereo
  oct1= (1<<4) | (1<<2) | (1<<0); //blocklength= 16, subbands=8, alloc=loudness

  B= avdtp_open_sbc_stream(A, reid, 1, oct0, oct1, min_bitpool, max_bitpool);
  if (!B)
  {
     printf("Can't open sbc stream.\n");
     goto fail;
  }

  if (mp3sbc_start("test.mp3", 44100, 16, 8, 2, 'L', bp))
  {
     printf("Can't start mp3sbc: %s\n", mp3sbc_error());
     goto fail;
  }

  uint32_t fdur;
  fdur= mp3sbc_frame_duration();

  uint8_t mp[2048];

  uint16_t seqno;
  uint32_t ts;
  uint32_t ssrc;
  seqno= 0;
  ts= 0;
  ssrc= 100;

  usleep(100*1000);

#if 0
  printf("ready to start. press enter\n");
  getchar();
#endif

  printf("frame duration= %d\n", fdur);
  while(1)
  {
    uint8_t *frame; size_t framelen;
    int mpl;
    int i;
#define NFRA 2

    mpl= bpack(mp, 2048, "bbwddb", 2,0, seqno, ts, ssrc,
                                                  NFRA, NULL);

    for(i=0;i<NFRA;i++) {
    if (mp3sbc_get_frame(&frame, &framelen))
    {
      printf("mp3sbc decoding/encoding error: %s\n", mp3sbc_error());
      goto fail;
    }
    if (!framelen)
    {
      printf("EOF\n");
      goto finish;
    }
    memcpy(mp+mpl, frame, framelen);
     mpl+= framelen;
    }

   
    if (send_packet(B, mp, mpl))
    {
      printf("could not send packet %d\n", seqno);
    }

    seqno++;
    ts+= 512 * NFRA;


    struct timespec req,rem;

    req.tv_sec= 0;
    req.tv_nsec= 1000;           // 1000*us= nanoseconds
    req.tv_nsec*= (fdur-DL)*NFRA;

    nanosleep(&req,&rem);
  }

finish:
  printf("transmitted %d packets\n", seqno);
  avdtp_close(B);

fail:
  avdtp_close(A);
  printf("all done.\n");
  return 0;
}

