#line 149 "/home/dodo/btstack/src/awrite.u"

#include <stdint.h>
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include "socket.h"

#line 8 "/home/dodo/btstack/src/awrite.u"

typedef struct awrite_buffer {
  uint8_t *buf, *obuf;
  uint16_t len;
  struct awrite_buffer *next;
  uint8_t hdr[2];
} awriter_buffer_t;

typedef struct {
  socket_t *socket;
  int error;
  char errstr[2000];

  uint16_t len;
  uint8_t *buf, *obuf;

  awriter_buffer_t *last,*first;
} awriter_t;

#line 28 "/home/dodo/btstack/src/awrite.u"

awriter_t *awrite_init(socket_t *socket)
{
  awriter_t *aw;
  aw= calloc(1,sizeof(*aw));
  aw->socket= socket;
  return aw;
}

#line 44 "/home/dodo/btstack/src/awrite.u"

void awrite_put(awriter_t *aw, uint8_t *buf, uint16_t len,uint8_t *obuf)
{
  awriter_buffer_t *B,*L;

  if (len==0) { if (obuf) free(obuf); return; }

  B= malloc(sizeof(*B));

  if (obuf && (buf-obuf>=2))
  {
    buf-= 2;
    buf[0]= len&0xff;
    buf[1]= len>>8;
    len+= 2;
    L= B;
  }
  else
  {
    L= malloc(sizeof(*L));
    L->buf= L->hdr;
    L->obuf= NULL;
    L->len= 2;
    L->next= B;
    L->hdr[0]= len&0xff;
    L->hdr[1]= len>>8;
  }

  B->buf= buf;
  B->len= len;
  B->obuf= obuf;
  B->next= NULL;

  if (aw->last)
    aw->last->next= L;
  else
    aw->first= L;

  aw->last= B;
}


int awrite_go(awriter_t *aw)
{
  int R;
  awriter_buffer_t *F,*N;

  F= aw->first;

  if (!F) return 1;

  R= send(aw->socket->fd, F->buf, F->len,
                        MSG_DONTWAIT | MSG_NOSIGNAL);

  if (R<=0)
  {
    if (errno==EINTR || errno==EWOULDBLOCK || errno==EAGAIN)
      return 0;
    aw->error= errno;
    aw->errstr[sizeof(aw->errstr)-1]= 0;
    snprintf(aw->errstr, sizeof(aw->errstr), "could not send data over "
                     "socket errno= %d: (%s)", errno, strerror(errno));
    return 1;
  }

  F->len-= R;
  F->buf+= R;

  if (aw->first->len==0)
  {
    if (F->obuf) free(F->obuf);
    N= F->next;
    free(F);
    aw->first= N;
    if (!aw->first) aw->last= NULL;
  }
  return 0;
}

void awrite_free(awriter_t *aw)
{
  awriter_buffer_t *B,*N;
  for(B=aw->first;B;B=N)
  {
    N= B->next;
    if (B->obuf) free(B->obuf);
    free(B);
  }
  free(aw);
}

#line 160 "/home/dodo/btstack/src/awrite.u"

