#line 320 "/home/dodo/btstack/mp3/mp3sbc.u"

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <mpg123.h>
#include <stdarg.h>
#include <sbc.h>

#line 7 "/home/dodo/btstack/mp3/mp3sbc.u"

static mpg123_handle *mp3h;
static sbc_t sbc;
static char errstr[200];

static int sbc_init_ok, mp3_init_ok;

FILE *inf;

#line 120 "/home/dodo/btstack/mp3/mp3sbc.u"

static char *mp3_lib_err;
static int mp3_lib_init;

#line 152 "/home/dodo/btstack/mp3/mp3sbc.u"

static size_t sbc_inp_size, sbc_out_size;
static size_t sbc_inp_alloc, sbc_out_alloc;
static unsigned char *sbc_inp, *sbc_out;

static size_t mp3_leftover_len;
static unsigned char* mp3_leftover;

#line 265 "/home/dodo/btstack/mp3/mp3sbc.u"

static off_t mp3_offset;

#line 330 "/home/dodo/btstack/mp3/mp3sbc.u"

#line 41 "mp3sbc.c"

int mp3sbc_start
  (char *fn, int freq, int blocks, int subbands,
             int channels, int allocation, int bitpool);
int mp3sbc_get_frame(unsigned char** out, size_t *len);
void mp3sbc_end();
char *mp3sbc_error();
uint32_t mp3sbc_frame_duration();
static int init_sbc
  (int freq, int blocks, int subbands,
             int channels, int allocation, int bitpool);
static int init_mp3(char *fn,int freq,int channels);
static int check_mp3_lib();
static void alloc_buffers();
static size_t min_size(size_t a,size_t b) ;
static int mserr(const char *fmt, ...);
#line 335 "/home/dodo/btstack/mp3/mp3sbc.u"

#define FROM_FILE 0
#line 17 "/home/dodo/btstack/mp3/mp3sbc.u"

int mp3sbc_start
  (char *fn, int freq, int blocks, int subbands,
             int channels, int allocation, int bitpool)
{
#if FROM_FILE
  inf= fopen("zz.sbc", "rb");
#endif

  sbc_init_ok= 0; mp3_init_ok= 0;

  if (init_sbc(freq,blocks,subbands, channels,allocation, bitpool))
     return 1;
  if (init_mp3(fn, freq, channels))
     return 1;
  alloc_buffers();
  return 0;
}
#line 200 "/home/dodo/btstack/mp3/mp3sbc.u"

int mp3sbc_get_frame(unsigned char** out, size_t *len)
{
   int R;
   int sbc_inp_len;


#if FROM_FILE
   if (fread(sbc_out, 1, 119, inf)<119)
     { *len= 0; return 0; }
   *len= 119;
   *out= sbc_out;
    return 0;
#endif

   sbc_inp_len= 0;

   while(sbc_inp_len<sbc_inp_size) {
     if (mp3_leftover_len)
     {
       size_t cp;
       cp= min_size(sbc_inp_size-sbc_inp_len, mp3_leftover_len);
       memcpy(sbc_inp+sbc_inp_len, mp3_leftover, cp);
       mp3_leftover_len-= cp;
       mp3_leftover+= cp;
       sbc_inp_len+= cp;
       continue;
     }
     R= mpg123_decode_frame(mp3h, &mp3_offset, &mp3_leftover,
                                              &mp3_leftover_len);
                  // It's OK, we already have only one format.
     if (R==MPG123_NEW_FORMAT) continue;
     if (R==MPG123_DONE) break; // EOF
     if (R!=MPG123_OK) goto decode_error;
   }

   if (sbc_inp_len==sbc_inp_size) // we have a full input frame
   {
      ssize_t consumed;
      ssize_t written;

      consumed= sbc_encode(&sbc, sbc_inp, sbc_inp_size, 
                                 sbc_out, sbc_out_size, &written);
  
      if (consumed<0) { R=consumed; goto encode_error; }
      *out= sbc_out;
      *len= written;
      return 0;
   } 
   else
   {
      *len= 0;
      return 0;
   }

decode_error:
   return mserr("mpg123_decode_frame: %s", mpg123_strerror(mp3h));
encode_error:
   return mserr("sbc_encode: %d (%s)", -R, strerror(-R));
}
#line 272 "/home/dodo/btstack/mp3/mp3sbc.u"

void mp3sbc_end()
{
  if (sbc_init_ok) sbc_finish(&sbc);
  if (mp3_init_ok) mpg123_close(mp3h);
  if (mp3h)        mpg123_delete(mp3h);
  mp3h= NULL;
  sbc_init_ok= 0;
  mp3_init_ok= 0;
}

#line 299 "/home/dodo/btstack/mp3/mp3sbc.u"

char *mp3sbc_error()
{
  return errstr;
}

#line 343 "/home/dodo/btstack/mp3/mp3sbc.u"

uint32_t mp3sbc_frame_duration()
{
  return sbc_get_frame_duration(&sbc);
  
}
#line 41 "/home/dodo/btstack/mp3/mp3sbc.u"

static int init_sbc
  (int freq, int blocks, int subbands,
             int channels, int allocation, int bitpool)
{
  sbc_init(&sbc,0);
  switch(freq)
  {
  case 48000: sbc.frequency= SBC_FREQ_48000; break;
  case 44100: sbc.frequency= SBC_FREQ_44100; break;
  case 32000: sbc.frequency= SBC_FREQ_32000; break;
  case 16000: sbc.frequency= SBC_FREQ_16000; break;
  default:
    return mserr("sbc_init: bad frequency %d", freq);
  }
  switch(blocks)
  {
  case 4: sbc.blocks= SBC_BLK_4; break;
  case 8: sbc.blocks= SBC_BLK_8; break;
  case 12:sbc.blocks= SBC_BLK_12; break;
  case 16:sbc.blocks= SBC_BLK_16; break;
  default:
    return mserr("sbc_init: bad blocks value %d", blocks);
  }
  switch(subbands)
  {
  case 4: sbc.subbands= SBC_SB_4; break;
  case 8: sbc.subbands= SBC_SB_8; break;
  default:
    return mserr("sbc_init: bad subbands value %d", subbands);
  }
  switch(channels)
  {
  case 1: sbc.mode= SBC_MODE_MONO; break;
  case 2: sbc.mode= SBC_MODE_JOINT_STEREO; break;
  default: 
    return mserr("sbc_init: bad channels value %d", channels);
  }
  switch(allocation)
  {
  case 'L': sbc.allocation= SBC_AM_LOUDNESS; break;
  case 'S': sbc.allocation= SBC_AM_SNR; break;
  default:
    return mserr("sbc_init: bad allocation value %d", allocation);
  }
  sbc.bitpool= bitpool;
  sbc_init_ok= 1;
  return 0; 
}

#line 95 "/home/dodo/btstack/mp3/mp3sbc.u"

static int init_mp3(char *fn,int freq,int channels)
{
  int R;
  int mpchan;

  if (check_mp3_lib()) return 1;

  if (channels==1) mpchan= MPG123_MONO;
             else  mpchan= MPG123_STEREO;
  mp3h= mpg123_new(0, &R);
  if (!mp3h)
    return mserr("mpg123_new: %s", mpg123_plain_strerror(R));
  if (mpg123_format_none(mp3h)!=MPG123_OK)
    return mserr("mpg123_format_none: %s", mpg123_strerror(mp3h));
  if (mpg123_format(mp3h, freq, mpchan, MPG123_ENC_SIGNED_16)!=MPG123_OK)
    return mserr("mpg123_format: %s", mpg123_strerror(mp3h));
  if (mpg123_open(mp3h, fn)!=MPG123_OK)
    return mserr("mpg123_open: %s", mpg123_strerror(mp3h));

  mp3_init_ok= 1;
  return 0;
}

#line 125 "/home/dodo/btstack/mp3/mp3sbc.u"

static int check_mp3_lib()
{
  int R;
  switch(mp3_lib_init)
  {
  case 0:
    R= mpg123_init();
    if (R!=MPG123_OK)
    { 
      mp3_lib_init=2;
      mp3_lib_err= strdup(mpg123_plain_strerror(R));
      return mserr("mpg123_init: %s", mp3_lib_err);
    }
    mp3_lib_init= 1;
    break;
  case 2:
    return mserr("mpg123_init: %s", mp3_lib_err);
  case 1:
    break;
  }
  return 0;
}
#line 166 "/home/dodo/btstack/mp3/mp3sbc.u"

static void alloc_buffers()
{
  sbc_inp_size= sbc_get_codesize(&sbc);

  printf("SBC Input Size= %ld (%ld ms)\n",sbc_inp_size, 
                          sbc_inp_size*1000/2/16/44100);
  sbc_out_size= sbc_get_frame_length(&sbc);
  if (sbc_inp_size>sbc_inp_alloc)
  {
    if (sbc_inp) free(sbc_inp);
    sbc_inp= malloc(sbc_inp_size);
    sbc_inp_alloc= sbc_inp_size;
  }
  if (sbc_out_size>sbc_out_alloc)
  {
    if (sbc_out) free(sbc_out);
    sbc_out= malloc(sbc_out_size);
    sbc_out_alloc= sbc_out_size;
  }

  mp3_leftover_len= 0;
}

#line 196 "/home/dodo/btstack/mp3/mp3sbc.u"

static size_t min_size(size_t a,size_t b) { return a<b?a:b; }

#line 287 "/home/dodo/btstack/mp3/mp3sbc.u"

static int mserr(const char *fmt, ...)
{
  va_list ap;
  va_start(ap,fmt);
  vsnprintf(errstr, sizeof(errstr)-1, fmt, ap);
  va_end(ap);
  errstr[sizeof(errstr)-1]= 0;
  return 1;
}

#line 339 "/home/dodo/btstack/mp3/mp3sbc.u"

