#line 101 "/home/dodo/btstack/gui/player_main.u"

#line 7 "/home/dodo/btstack/gui/player_device.u"
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <errno.h>
#include "socket.h"
#include "aread.h"
#include "awrite.h"
#include "util.h"
#include "sdp-lib.h"
#include "mp3sbc.h"


#line 103 "/home/dodo/btstack/gui/player_main.u"
#include <poll.h>
#include "mbox.h"
#include "gui.h"
#include "msg_worker.h"
#include "msg_dpy.h"

#line 26 "player_main.c"

static void *worker_func(void *dummy);
void worker_quit();
void worker_play_song(char *fn);
void worker_pause() ;
void worker_resume() ;
void worker_close_file();
static int poll_msg();
int endpoint_suitable(avdtp_conn_t *A, uint8_t id,
                      uint8_t *Rmin_bitpool, uint8_t *Rmax_bitpool);
int get_audio_sink(avdtp_conn_t *A,uint8_t *Rid,
                   uint8_t *min_bitpool,uint8_t *max_bitpool);
uint8_t max8(uint8_t a,uint8_t b) ;
uint8_t min8(uint8_t a,uint8_t b) ;
static int connect_device();
static void send_frame();
static void get_device_address();
static int ishex(int a);
static int hexv(int a);
#line 112 "/home/dodo/btstack/gui/player_main.u"

#line 27 "/home/dodo/btstack/gui/player_main.u"

gui_main_t* main_window;
static char *home_dir;
static pthread_t worker_thread;
mbox_t *worker_mbox, *dpy_mbox;

#line 61 "/home/dodo/btstack/gui/player_main.u"

static int quit_all;

#line 93 "/home/dodo/btstack/gui/player_main.u"

static int playing, buffer_ok;
static uint32_t fdur;
static uint16_t seqno;
static uint32_t ts;


#line 81 "/home/dodo/btstack/gui/player_device.u"

static int conn_state;
static avdtp_conn_t *conn_sig, *conn_data;
static uint8_t hs_addr[6]= { 0x59,0x37,0x11,0x67,0x11,0x00 };
static uint8_t bitpool;
static uint8_t device_addr[6];

#line 163 "/home/dodo/btstack/gui/player_main.u"

char *get_home_dir() { return home_dir; }
void play_sound_file(char *fn) 
{
  mbox_put(worker_mbox, worker_PLAY_SONG(fn));
}

void close_sound_file() { mbox_put(worker_mbox, worker_CLOSE_FILE()); }
void pause_sound() { mbox_put(worker_mbox, worker_PAUSE()); }
void resume_sound() { mbox_put(worker_mbox, worker_RESUME()); }
void close_player() { mbox_put(worker_mbox, worker_QUIT()); }

#line 41 "/home/dodo/btstack/gui/player_main.u"

static void *worker_func(void *dummy)
{
  while(!quit_all)
  {
    if (playing)
    {
      if (poll_msg(0))
        { worker_dispatch(mbox_get(worker_mbox)); continue; }
      send_frame();
      continue;
    }
    worker_dispatch(mbox_get(worker_mbox));
  }
  pthread_exit(dummy);
}
#line 65 "/home/dodo/btstack/gui/player_main.u"

void worker_quit()
{
  quit_all= 1;
  // FIXME: properly disconnect everything
}

#line 76 "/home/dodo/btstack/gui/player_main.u"

void worker_play_song(char *fn)
{
  if (connect_device()) return ;
  if (buffer_ok)
    worker_close_file();
  if (mp3sbc_start(fn, 44100, 16, 8, 2, 'L', bitpool))
  {
    fprintf(stderr,"Can't start mp3sbc: %s\n", mp3sbc_error());
    return ;
  }
  fdur= mp3sbc_frame_duration();
  buffer_ok= 1;
  playing= 1;
}

#line 125 "/home/dodo/btstack/gui/player_main.u"

void worker_pause() { playing= 0; }
void worker_resume() { if (buffer_ok) playing= 1; }
void worker_close_file()
{
  if (buffer_ok) mp3sbc_end();
  buffer_ok= 0;
}


#line 140 "/home/dodo/btstack/gui/player_main.u"

static int poll_msg()
{
  struct pollfd  pfd;
  struct timespec tv;
  pfd.fd= mbox_fd(worker_mbox);
  pfd.events= POLLIN;
  pfd.revents= 0;
   // FIXME: 2 is NFRA, make it a top-level definition somewhere
   // also, 100 is the delay 
  tv.tv_nsec= (fdur-100) * 1000 *2;
  tv.tv_sec= 0;
  if (ppoll(&pfd, 1, &tv, NULL)>0)
     return 1;
  return 0;
}

#line 23 "/home/dodo/btstack/gui/player_device.u"

int endpoint_suitable(avdtp_conn_t *A, uint8_t id,
                      uint8_t *Rmin_bitpool, uint8_t *Rmax_bitpool)
{
  uint8_t oct0, oct1;
  if (avdtp_get_sbc_capabilities(A, id, &oct0, &oct1,
                                Rmin_bitpool, Rmax_bitpool))
  {
    printf("Can't get capabilities for endpoint %d\n", id);
    return 0;
  }
  return 1;
}

int get_audio_sink(avdtp_conn_t *A,uint8_t *Rid,
                   uint8_t *min_bitpool,uint8_t *max_bitpool)
{
  uint8_t *endpoints;
  uint32_t s_endpoints;  
  int i;

  if (avdtp_discover_endpoints(A, &endpoints, &s_endpoints))
  {
    printf("Could not get endpoint list.\n");
    return 1;
  }
  
  printf("Endpoints\n");
  for(i=0;i<s_endpoints;i+=2)
  {
     uint8_t id, in_use, media_type, tsep;
     id= endpoints[i] >> 2;
     in_use= (endpoints[i] >> 1) &1;
     media_type= endpoints[i+1] >> 4;
     tsep= (endpoints[i+1]>>3) &1;
     printf("  EP: %d in_use: %d media_type: %d tsep: %d\n",
                  id, in_use, media_type, tsep);

           // sink,  audio
     if (tsep==1 && media_type==0 && 
         endpoint_suitable(A, id,min_bitpool,max_bitpool))
     {
       *Rid=  id;
       return 0;
     }
  }

  return 1;
}

uint8_t max8(uint8_t a,uint8_t b) { return a>b?a:b; }
uint8_t min8(uint8_t a,uint8_t b) { return a<b?a:b; }

#line 89 "/home/dodo/btstack/gui/player_device.u"

static int connect_device()
{
  uint8_t max_bitpool, min_bitpool;
  uint8_t reid;
  uint8_t bp;
  uint8_t oct0, oct1;

  if (conn_state==2) return 1;
  if (conn_state==1) return 0;
  get_device_address();
  conn_sig= avdtp_open(device_addr);

  if (conn_sig->err) {
    fprintf(stderr,"avdtp-test(connect): %s\n", conn_sig->errstr);
    goto failed;
  }

  if (get_audio_sink(conn_sig, &reid, &min_bitpool, &max_bitpool)) {
    fprintf(stderr,"can't find an audio sink on device.\n");
    goto failed;
  }

  printf("Suitable EP: %d min_bp= %d max_bp= %d\n",
             reid, min_bitpool, max_bitpool);


  if (53>=min_bitpool && 53<=max_bitpool) bp= 53;
  else if (35>=min_bitpool && 35<=max_bitpool) bp= 35;
  else bp= ((uint32_t) max_bitpool+(uint32_t) min_bitpool) /2;
  printf("Selected bitpool: %d\n", bp);

  bitpool= bp;

  // now we can configure our stream.

  oct0= (1<<5) | (1<<0) ; // 44100 Hz, Joint Stereo
  oct1= (1<<4) | (1<<2) | (1<<0); //blocklength= 16, subbands=8, alloc=loudness

  conn_data= avdtp_open_sbc_stream(conn_sig, reid, 1, oct0, oct1,
                                   min_bitpool, max_bitpool);
  if (!conn_data)
  {
     fprintf(stderr,"Can't open sbc stream.\n");
     goto failed;
  }

  conn_state= 1;
  return 0;
failed:
  conn_state= 2;
  return 1;
}

#line 147 "/home/dodo/btstack/gui/player_device.u"

static void send_frame()
{
  uint8_t mp[2048];
  uint32_t ssrc;
  ssrc= 100;
  uint8_t *frame; size_t framelen;
  int mpl;
  int i;
#define NFRA 2

  mpl= bpack(mp, 2048, "bbwddb", 2,0, seqno, ts, ssrc, NFRA, NULL); 
  for(i=0;i<NFRA;i++)
  {
    if (mp3sbc_get_frame(&frame, &framelen))
    {
      printf("mp3sbc decoding/encoding error: %s\n", mp3sbc_error());
      goto finish;
    }
    if (!framelen)
    {
      printf("EOF\n");
      goto finish;
    }
    memcpy(mp+mpl, frame, framelen);
     mpl+= framelen;
  }
   
  if (send_packet(conn_data, mp, mpl))
  {
     printf("could not send packet %d\n", seqno);
  }

  seqno++;
  ts+= 512 * NFRA;

  return ;
finish:
  playing= 0;
  worker_close_file();
  mbox_put(dpy_mbox, dpy_SONG_FINISHED());
  return ;
}

#line 192 "/home/dodo/btstack/gui/player_device.u"

static void get_device_address()
{
  char *v;
  char *k;
  int i;
  v= getenv("BTH_DEVICE_ADDR");
  if (!v) goto fail;

  k= malloc(strlen(v));
  i= 0;
  while(v[0])
  {
    if (ishex(*v)) k[i++]= v[0];
    v++;
  }
  k[i]= 0;
  if (i!=12) goto fail;
  for(i=5;i>=0;i--)
  {
    device_addr[i]= (hexv(k[0]) << 4) | hexv(k[1]);
    k+= 2;
  }
  return ;

fail:
  memcpy(device_addr, hs_addr, 6);
}

static int ishex(int a)
{
  if (a>='0'&&a<='9') return 1;
  if (a>='a'&&a<='f') return 1;
  if (a>='A'&&a<='F') return 1;
  return 0;
}

static int hexv(int a)
{
  if (a>='0'&&a<='9') return a-'0';
  if (a>='a'&&a<='f') return a-'a'+10;
  if (a>='A'&&a<='F') return a-'A'+10;
  return 0;
}

#line 116 "/home/dodo/btstack/gui/player_main.u"

#line 6 "/home/dodo/btstack/gui/player_main.u"


int main(int argc,char **argv)
{
  home_dir= getenv("BTP_HOME_DIR");
  if (!home_dir) home_dir= getenv("HOME");
  if (!home_dir) home_dir= "/home/dodo/Music";
  home_dir= strdup(home_dir);
  gui_init(&argc,argv);
  worker_mbox= mbox_new();
  dpy_mbox= mbox_new();
  main_window= gui_c_main();
  gui_set_current_dir(main_window->files, home_dir);
  pthread_create(&worker_thread, NULL, worker_func);
  gui_run();
  pthread_join(worker_thread, NULL);
  return 0;
}


#line 118 "/home/dodo/btstack/gui/player_main.u"
#include "msg_worker.c"
#include "msg_dpy.c"

