#line 167 "/home/dodo/btstack/gui/playlist.u"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "actions.h"
#include "gui.h"

extern gui_main_t *main_window;

#line 26 "/home/dodo/btstack/gui/playlist.u"

typedef struct mfile
{
  struct mfile *prev, *next;
  int number;
  char *file;
} mfile_t;

#line 35 "/home/dodo/btstack/gui/playlist.u"

static mfile_t files= { &files,&files };
static int max_number;

#line 47 "/home/dodo/btstack/gui/playlist.u"

static int playing_song;
static int is_playing;

#line 143 "/home/dodo/btstack/gui/playlist.u"

static int is_paused;

#line 34 "playlist.c"

static mfile_t *next_song();
static mfile_t *prev_song();
static mfile_t *find_song(int nr);
static mfile_t *first_song();
static void play_song(mfile_t *M);
static int has_sfx(const char *src,const char *sfx);
#line 52 "/home/dodo/btstack/gui/playlist.u"

static mfile_t *next_song()
{
  mfile_t *F,*found;
  if (!is_playing) return first_song();
  // we now find the minimal song number among the ones greater
  // than the current one.
  found= NULL;
  for(F=files.prev;F!=&files;F=F->prev)
  {
     if (F->number<=playing_song) break;
     found= F;
  }
  return found;
}


static mfile_t *prev_song()
{
  mfile_t *F,*found;
  if (!is_playing) return first_song();
  found= NULL;
  for(F=files.next;F!=&files;F=F->next)
  {
     if (F->number>=playing_song) break;
     found= F;
  }
  return found;
}

#line 87 "/home/dodo/btstack/gui/playlist.u"

static mfile_t *find_song(int nr)
{
  mfile_t *F;
  for(F=files.next;F!=&files;F=F->next)
    if (F->number==nr) return F;
  return NULL;
}

static mfile_t *first_song()
{
  if (files.next==&files) return NULL;
  return files.next;
}


#line 147 "/home/dodo/btstack/gui/playlist.u"

static void play_song(mfile_t *M)
{
  if (!M) return ;
  if (is_playing) {
    if (!is_paused)
    {
       pause_sound();
       is_paused= 1;
    } 
    close_sound_file();
  } 
  is_playing= 1;
  is_paused= 0;
  playing_song= M->number;
  play_sound_file(M->file);
}


#line 189 "/home/dodo/btstack/gui/playlist.u"

static int has_sfx(const char *src,const char *sfx)
{
  int Ls,Lf;
  Ls= strlen(src); Lf= strlen(sfx);
  if (Ls<Lf) return 0;
  return !strcasecmp(src+Ls-Lf,sfx);
}

#line 10 "/home/dodo/btstack/gui/playlist.u"

void playlist_add_file(const char *name, const char *file)
{
  mfile_t *M;
  if (!has_sfx(name,".mp3")) return ;
  M= malloc(sizeof(*M));
  M->number= max_number++;
  M->file= strdup(file);
  M->next= &files;
  M->prev= files.prev;
  files.prev->next= M;
  files.prev= M;
  gui_list_append(main_window->playlist, name, M->number);
}

#line 104 "/home/dodo/btstack/gui/playlist.u"

void play_file(int nr) { play_song(find_song(nr)); } 
void play_next() { play_song(next_song()); }
void play_prev() { play_song(prev_song()); }


#line 114 "/home/dodo/btstack/gui/playlist.u"

void playlist_remove_file(int nr)
{
  mfile_t *F;
  F= find_song(nr);
  if (!F) return ;
  F->prev->next= F->next;
  F->next->prev= F->prev;
  free(F->file);
  free(F);
  gui_remove(main_window->playlist, nr);
}

void playlist_remove_all()
{
  mfile_t *F,*N;
  for(F=files.next;F!=&files;F=N)
  {
     gui_remove(main_window->playlist, F->number);
     N= F->next; free(F->file); free(F);
  }
  files.next= files.prev= &files;
}

#line 203 "/home/dodo/btstack/gui/playlist.u"


void action_play() {
  if (is_playing)
  {
     if (is_paused)
     {
       resume_sound();
       is_paused= 0;
     }
     return ;
  }
  play_song(first_song());
}
void action_pause() {
  if (!is_playing || is_paused) return ;
  pause_sound();
  is_paused= 1;
}

void dpy_song_finished()
{
  play_song(next_song());
}

#line 183 "/home/dodo/btstack/gui/playlist.u"


