#line 272 "/home/dodo/btstack/src/sdp-lib.u"
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <sys/socket.h>
#include <errno.h>
#include <stdarg.h>
#include "socket.h"
#include "bpack.h"
#include "awrite.h"
#include "aread.h"
#include "comnames.h"
#include "util.h"
#include "sdp-lib.h"

#line 20 "/home/dodo/btstack/src/sdp-lib.u"

#define SDP_SERVICE_SEARCH_REQ  2
#define SDP_SERVICE_SEARCH_REP  3
#define SDP_SERVICE_ATTRIBUTE_REQ 4
#define SDP_SERVICE_ATTRIBUTE_REP 5

#define SDP_DESEQ_B      "blW[", (6<<3) | 6 
#define SDP_DESEQ_E      "]"
#define SDP_UUID(X)      "bW", (3<<3) | 1, X

#line 28 "sdp-lib.c"

int sdp_send_bytes(sdp_conn_t *S, uint8_t *buf, int len);
int sdp_recv_bytes(sdp_conn_t *S, uint8_t *buf, int len);
int sdp_send(sdp_conn_t *S, ...);
int sdp_recv(sdp_conn_t *S);
void sdp_disconnect(sdp_conn_t *S);
static sdp_protocol_t **sdp_interp_protodesc(uint8_t *buffer);
int sdp_scan_typelen(uint8_t *buffer, uint8_t *type, uint32_t *len);
int sdp_parse_single_protodesc(sdp_protocol_t **Rtop, uint8_t *buffer);
int sdp_parse_single_protocol(sdp_protocol_t **Rs, uint8_t *buffer);
static int avdtp_wrd(avdtp_conn_t *A, uint8_t *cmd, uint32_t len);
static void print_hex(char *pmt, uint8_t *B, int L);
#line 14 "/home/dodo/btstack/src/sdp-attrib.u"

uint8_t 
TEST_OUT[]=
{
0x05,0x00,0x02,0x00,0x23,0x00,0x20,0x36,0x00,0x1d,
0x09,0x00,0x04,0x35,0x18,0x35,0x06,0x19,0x01,0x00,
0x09,0x00,0x0f,0x35,0x0e,0x19,0x00,0x0f,0x09,0x01,
0x00,0x35,0x06,0x09,0x86,0xdd,0x09,0x08,0x06,0x00
};

#line 61 "/home/dodo/btstack/src/avdtp-lib.u"
static int avdtp_trans;

#line 146 "/home/dodo/btstack/src/sdp-lib.u"

int sdp_send_bytes(sdp_conn_t *S, uint8_t *buf, int len)
{
  int R;
  while(len)
  {
    R= send(S->socket->fd, buf, len, MSG_NOSIGNAL);
    if (R<=0)
    {
      if (errno==EINTR) continue;
      return 1;
    }
    buf+= R;
    len-= R;
  }

  return 0;
}

int sdp_recv_bytes(sdp_conn_t *S, uint8_t *buf, int len)
{
  int R;
  while(len)
  {
    R= recv(S->socket->fd, buf, len, 0);
    if (R<=0)
    {
      if (errno==EINTR) continue;
      return 1;
    }
    buf+= R;
    len-= R;
  }

  return 0;
}

#line 188 "/home/dodo/btstack/src/sdp-lib.u"

int sdp_send(sdp_conn_t *S, ...)
{
  int bl;
  va_list ap;

again:
  va_start(ap, S);
  bl= bpack(S->obuf, S->obufsiz, "lw[a]", ap, NULL);
  va_end(ap);

  if (bl>S->obufsiz)
  {
     S->obufsiz= bl + bl/2;
     free(S->obuf);
     S->obuf= malloc(S->obufsiz);
     goto again;  
  }
  
  S->obuflen= bl;

  int i;
  printf("SENDING SDP PACKET\n");
  for(i=0;i<bl;i++) printf(" %02x", S->obuf[i]);
  printf("\n");

  return sdp_send_bytes(S, S->obuf, S->obuflen);
}

int sdp_recv(sdp_conn_t *S)
{
  uint8_t *B;
  uint8_t opcode;
  uint16_t trans, len, errcode;

  if (sdp_recv_bytes(S, S->ibuf, 2)) return 1;
  S->ibuflen= S->ibuf[0] | (S->ibuf[1]<<8);

  printf("new inbound message length= %d\n", S->ibuflen);

  if (S->ibuflen<4) return 1;
  if (sdp_recv_bytes(S, S->ibuf, S->ibuflen)) return 1;

  int i;
   printf("INBOUND MESSAGE\n");
  for(i=0;i<S->ibuflen;i++)
    printf(" %02x", S->ibuf[i]);
  printf("\n");

  B= S->ibuf;
  B+= bscan(B,"bWW", &opcode, &trans, &len);
  if (opcode==0x01) // error response
  {
    bscan(B,"W", &errcode);
    printf("SDP command failed with code %04x\n",errcode);
    return 1;
  }
  return 0;
}
#line 251 "/home/dodo/btstack/src/sdp-lib.u"

void sdp_disconnect(sdp_conn_t *S)
{
  socket_destroy(S->socket);
  free(S->ibuf);
  free(S->obuf);
  free(S);
}


#line 40 "/home/dodo/btstack/src/sdp-proto.u"

static sdp_protocol_t **sdp_interp_protodesc(uint8_t *buffer)
{
  uint8_t T; uint32_t L;
  sdp_protocol_t *P;
  sdp_protocol_t **protos;
  int n_protos, s_protos;
  uint8_t *B, *end;
  sdp_protocol_t **R;
  int i;
  uint8_t *V;

  pvecini(&protos, &s_protos, &n_protos);

  B= buffer;
  B+= sdp_scan_typelen(B, &T, &L);
  if (T!=6 && T!=7)
  {
    printf("invalid type %d in protocol descriptor list.\n", T);
    goto done;
                // bad type, it should've been 7=alternative
                //                          or 6= sequence
  }

  if (T==6) // sequence
  {
    B= buffer;
    B+= sdp_parse_single_protodesc(&P, B);
    if (P)
       pvecapd(protos, P);
  }
  else 
  {
    end= B+L;
    while(B<end)
    {
      B+= sdp_parse_single_protodesc(&P, B);
      if (P)
         pvecapd(protos,P);
      else
         break;
    }
  }

done:
  pvecapd(protos, 0);
  
  R= malloc(n_protos*sizeof(void*));
  memcpy(R, protos, n_protos*sizeof(void*));
  pvecfre(protos);
  return R;
}

int sdp_scan_typelen(uint8_t *buffer, uint8_t *type, uint32_t *len)
{
  uint8_t T;
  uint8_t *B;
  uint8_t lB; uint16_t lW; 

  B= buffer;
  B+= bscan(B, "b", &T);
  switch(T&0x07)
  {
  case 0: if (T==0) *len= 0; else *len= 1; break;
  case 1: *len= 2; break;
  case 2: *len= 4; break;
  case 3: *len= 8; break;
  case 4: *len= 16; break;
  case 5: B+= bscan(B, "b", &lB); *len= lB; break;
  case 6: B+= bscan(B, "W", &lW); *len= lW; break;
  case 7: B+= bscan(B, "D", len); break;
  }
  *type= T>>3;
  return B-buffer;
}

#line 126 "/home/dodo/btstack/src/sdp-proto.u"

int sdp_parse_single_protodesc(sdp_protocol_t **Rtop, uint8_t *buffer)
{
  uint8_t *B,*end;
  uint8_t T; uint32_t L;
  sdp_protocol_t *P;

  B= buffer;

  *Rtop= NULL;
  B+= sdp_scan_typelen(B, &T, &L);
  end= B+L;
  if (T!=6) goto done;


  while(B<end)
  {
     B+= sdp_parse_single_protocol(&P, B);
     if (!P) goto done;
     P->lower= *Rtop;
     *Rtop= P;
  }

done:
  return end-buffer;
}

#line 158 "/home/dodo/btstack/src/sdp-proto.u"

int sdp_parse_single_protocol(sdp_protocol_t **Rs, uint8_t *buffer)
{
  uint8_t *B,*end;
  uint8_t T; uint32_t L;

  uint16_t uuid;
  uint8_t *pstart;
  sdp_protocol_t *P;

  B= buffer;
  P= NULL;
  B+= sdp_scan_typelen(B, &T, &L);
  end= B+L;
  if (T!=6) goto done;

  if (B>=end) goto done;
  B+= sdp_scan_typelen(B, &T, &L);
  if (T!=3 || L!=2) goto done; // not a 16-bit uuid we don't support this

  B+= bscan(B, "W", &uuid);
  
  P= malloc(sizeof(*P));
  P->uuid= uuid;
  P->paramslen= end-B;
  P->params= malloc(P->paramslen);
  memcpy(P->params, B,P->paramslen);
  P->lower= NULL;

done:
  *Rs= P;
  return end-buffer;
}

#line 84 "/home/dodo/btstack/src/avdtp-lib.u"

static int avdtp_wrd(avdtp_conn_t *A, uint8_t *cmd, uint32_t len)
{
  uint8_t SH;

  print_hex("avdtp-out:", cmd, len);

  if (A->err) return 1;
  awrite_put(A->wr, cmd, len, NULL);
  while(A->wr->first)
   if (awrite_go(A->wr))
   {
     snprintf(A->errstr, sizeof(A->errstr),
              "can't write command to daemon: %s", A->sock->errstr);
     A->err= 1;
     return 1;
   }

  aread_start(A->rd);
  while(!aread_go(A->rd))
     ;

  if (A->rd->error)
  {
    snprintf(A->errstr, sizeof(A->errstr),
              "can't read response from remote device: %s",A->sock->errstr);
    A->err= 1;
    return 1;
  }

  print_hex("avdtp-in :", A->rd->buffer, A->rd->len);

  SH= A->rd->buffer[0];
  if ((SH>>4) != avdtp_trans)
  {
    snprintf(A->errstr, sizeof(A->errstr),
              "different transcation id was returned for avdtp "
              "command. was expecting %x, got %x", avdtp_trans & 0xf, SH>>4);
    A->err= 1;
    return 1;
  }
  SH &= ~0xf0;
  if (SH&(3<<2))
  {
    snprintf(A->errstr, sizeof(A->errstr),
              "unsupported packet type %d. we only support single packets in "
              "avdtp", (SH>>2));
    A->err= 1;
    return 1;
  }
  SH &= ~(3<<2);

  switch(SH)
  {
   case 2: return 0;
   case 0: snprintf(A->errstr, sizeof(A->errstr), 
                    "the command was responded by another command? invalid.");
           break;
   case 1: snprintf(A->errstr, sizeof(A->errstr), 
                    "the command was responded by invalid message type.");
           break;
   case 3: snprintf(A->errstr, sizeof(A->errstr), 
                    "the command was rejected.");
           break;
  }
  
  A->err= 1;
  return 1;
}

#line 309 "/home/dodo/btstack/src/avdtp-lib.u"

static void print_hex(char *pmt, uint8_t *B, int L)
{
  int i;
  printf("%s", pmt);
  for(i=0;i<L;i++)
    if (i) printf(":%02x",B[i]);
    else printf("%02x", B[i]);
  printf("\n");
}




#line 31 "/home/dodo/btstack/src/sdp-lib.u"


uint32_t *sdp_service_list(sdp_conn_t *SC, int nsea, uint16_t *sea)
{
  uint8_t cont[32];
  int trans;
  int maxc=0xff; // for testing. I will make it bigger later.
  uint32_t *services;
  int nservices, nreceived;
  int i;

  uint16_t n, nc,cp;
  uint8_t *B;
  uint8_t uuids[12*3];

  cont[0]= 0;
  services= NULL;
  nservices= nreceived= 0;

  B= uuids;
  for(i=0;i<nsea;i++)
    B+= bpack(B, 3,SDP_UUID(sea[i]), NULL);


again:
  trans= SC->trans++;
  if (sdp_send(SC, "bWlW[", SDP_SERVICE_SEARCH_REQ, trans,
           SDP_DESEQ_B, 
              "p", 3*nsea, uuids,
           SDP_DESEQ_E,
           "W", maxc,  // Max Service Record Count
           "bp", cont[0], cont[0], cont,"]", NULL))
     goto done;

  if (sdp_recv(SC)) // I don't use the transaction ID at all,
                    // this is wrong.
     goto done;


  B= SC->ibuf+ 5; // skip the PDU header
  B+= bscan(B, "WW", &n, &nc);
  if (services && n!=nservices)
     goto done; // different results in different responses

  if (!services)
  {
    nservices= n;
    services= malloc(4*(nservices+1));
  }

  cp= nservices-nreceived;

  if (cp>nc) cp= nc;
  for(i=0;i<cp;i++)
    B+= bscan(B, "D", services+nreceived+i);
  nreceived+= cp;

  if (nreceived==nservices) goto done;

  memcpy(cont, B, B[0]+1);
  goto again; 

done:
  if (services)
   services[nreceived]= 0x0000; // sentinel
  return services;
}

#line 103 "/home/dodo/btstack/src/sdp-lib.u"

sdp_conn_t *sdp_connect(uint8_t *addr)
{
   sdp_conn_t *S;
   S= malloc(sizeof(*S));
   S->socket= socket_connect_local(60712);
   if (S->socket->state==SOCKET_ERROR)
   {
      snprintf(S->errstr, sizeof(S->errstr),
               "can't connect to daemon: %s", S->socket->errstr);
      S->state= 1;
      return S;
   }
   S->obufsiz= S->ibufsiz= 1024;
   S->ibuflen= S->obuflen= 0;
   S->obuf= malloc(S->obufsiz);
   S->ibuf= malloc(S->ibufsiz);
   S->trans= 1; // last transaction id
   S->state= 0;

   // send the connection request
   if (sdp_send(S, "pw", 6, addr, 0x0001, NULL))
   {
      snprintf(S->errstr,sizeof(S->errstr), "can't send request: %s",
                      strerror(errno));
      S->state= 1;
      return S;
   }
   
   if (sdp_recv_bytes(S,S->ibuf,3))
   {
      snprintf(S->errstr,sizeof(S->errstr), "connection to device failed");
      S->state= 1;
      return S;
   }

   return S;
}

#line 311 "/home/dodo/btstack/src/sdp-lib.u"

ctrl_t *ctrl_connect()
{
  ctrl_t *C;
  char buf[8];
  C= calloc(1,sizeof(*C));
  C->socket= socket_connect_local(60712);
  if (C->socket->state==SOCKET_ERROR)
  {
     snprintf(C->errstr, sizeof(C->errstr),
              "can't connect to daemon: %s", C->socket->errstr);
     C->error= 1;
     return C;
  }
  C->wr= awrite_init(C->socket);
  C->rd= aread_init(C->socket,0);
  memset(buf,0,8);

  awrite_put(C->wr, buf, 8, NULL);
  while(C->wr->first)
    if (awrite_go(C->wr))
    {
     snprintf(C->errstr, sizeof(C->errstr),
              "can't write initial command to daemon: %s", C->socket->errstr);
     C->error= 1;
     return C;
    }

  return C;
}

uint64_t *ctrl_device_list(ctrl_t *C,int inq)
{
  uint64_t *devices;
  uint16_t N;
  uint8_t buffer[2];
  int i; uint8_t *B;

  if (inq)
    bpack(buffer,2, "w", COM_INQUIRE,NULL);
  else
    bpack(buffer,2, "w", COM_DEVICES,NULL);
   
  if (ctrl_send(C, buffer,2)) return 0;
  if (ctrl_recv(C)) return 0;
  B= C->rd->buffer;
  B+= bscan(B, "w", &N, NULL);
  
  devices= malloc(sizeof(uint64_t)*(N+1));
  for(i=0;i<N;i++)
   devices[i]= read_addr(B+i*6);
  devices[i]= 0;
  return devices;
}

uint64_t read_addr(uint8_t *B)
{
  uint32_t L; uint16_t H;
  uint64_t R;

  bscan(B, "dw", &L, &H,NULL);
  R= H;
  R= (R<<32) | L;
  return R;
}

int write_addr(uint8_t *dest, uint64_t addr)
{
  int i;
  for(i=0;i<6;i++) { dest[i]= addr; addr>>= 8; }
  return 6;
}

int ctrl_send(ctrl_t *C, uint8_t *buf, int len)
{
  awrite_put(C->wr, buf, len, NULL);
  while(C->wr->first)
    if (awrite_go(C->wr))
      return 1;
  return 0;
}

void print_addr(uint64_t addr)
{
  int i;
  for(i=5;i>=0;i--)
   printf("%s%02x", (i!=5)? ":" : "", (uint8_t) ((addr>>(i*8))&0xff));
}


int ctrl_recv(ctrl_t *C)
{
  aread_start(C->rd);
  while(!aread_go(C->rd))
     ;

  if (C->rd->error)
  {
     snprintf(C->errstr, sizeof(C->errstr),
              "device_list: can't read from daemon: %s", C->socket->errstr);
     C->error= 1;
     return 1;
  }
  return 0;
}

char *ctrl_device_name(ctrl_t *C, uint64_t addr)
{
  uint8_t buffer[200];
  int i;
  uint8_t *B;

  B= buffer;
  B+= bpack(B, 2, "w", COM_NAME, NULL);
  B+= write_addr(B, addr);

  if (ctrl_send(C, buffer, B-buffer)) return strdup("");
  if (ctrl_recv(C)) return strdup("");

  return strdup(C->rd->buffer);
}

void ctrl_close(ctrl_t *C)
{
  aread_free(C->rd);
  awrite_free(C->wr);
  socket_destroy(C->socket);
  free(C);
}


#line 25 "/home/dodo/btstack/src/sdp-attrib.u"

int sdp_get_attribs(sdp_conn_t *S, uint32_t svc, ...)
{
  va_list ap;
  int N;
  int i;
  uint8_t **rp;
  uint32_t an;
  uint16_t *names; uint8_t *namep; void **resp;
  int rv;

  uint8_t *B, *V,*end,*C;
  uint8_t T; uint32_t L;
  uint16_t nb;
  uint16_t trans;

  rv= 1;

#line 113 "/home/dodo/btstack/src/sdp-attrib.u"
  va_start(ap, svc);
  N= 0;
  while(1)
  {
    rp= va_arg(ap, uint8_t**);
    if (!rp) break;
    *rp= NULL;
    an= va_arg(ap, int);
    N++;   
  }
  va_end(ap);


#line 44 "/home/dodo/btstack/src/sdp-attrib.u"

  names= malloc(2*N);
  resp= malloc(sizeof(void*)*N);

#line 127 "/home/dodo/btstack/src/sdp-attrib.u"
  va_start(ap, svc);
  N= 0;
  while(1)
  {
    rp= va_arg(ap, uint8_t**);
    if (!rp) break;
    an= va_arg(ap, int);
    names[N]= an; resp[N]= rp;
    N++; 
  }
  va_end(ap);

#line 140 "/home/dodo/btstack/src/sdp-attrib.u"
  while(1) {
    int change;
    uint16_t Tw; void *Tp;
    change= 0;
    for(i=0;i<N-1;i++)
      if (names[i]>names[i+1])
      {
        Tw= names[i]; names[i]= names[i+1]; names[i+1]= Tw;
        Tp= resp[i];  resp[i]= resp[i+1]; resp[i+1]= Tp;
        change= 1;
      }
    if (!change) break;
  }


#line 50 "/home/dodo/btstack/src/sdp-attrib.u"

  namep= malloc(3*N);
#line 156 "/home/dodo/btstack/src/sdp-attrib.u"

  for(i=0;i<N;i++) {
    namep[3*i]=   (1<<3) | 1 ; // uint16
    namep[3*i+1]= names[i] >> 8;
    namep[3*i+2]= names[i] & 0xff;
  }


#line 53 "/home/dodo/btstack/src/sdp-attrib.u"

  trans= S->trans++;
  if (sdp_send(S, "bWlW[", SDP_SERVICE_ATTRIBUTE_REQ, trans, 
                     "D", svc,
                     "W", 0xffff,
                     SDP_DESEQ_B,
                       "p", 3*N, namep,
                     SDP_DESEQ_E,
                   "b]", 0, NULL)) // continuation = 0 FIXME
     goto fail;
  if (sdp_recv(S))
     goto fail;


  B= S->ibuf + 5;
  B+= bscan(B, "W", &nb);
  end= B+nb;
  if (nb<2) goto fail;

  // FIXME: we should make sure we never go beyond end during 
  // FIXME: scan_typelen. this is a problem
  B+= sdp_scan_typelen(B, &T, &L);
  if (T!=6) goto fail; 
  // FIXME: give an error message here

  while(B<end)
  {
    B+= sdp_scan_typelen(B, &T, &L);
    if (T!=1 || L!=2) 
    {
      B+= L;
      B+= sdp_scan_typelen(B, &T, &L);
      B+= L;
    } else {
      B+= bscan(B, "W", &an);
      V= B;
      B+=sdp_scan_typelen(B, &T,&L);
      B+= L;
      for(i=0;i<N;i++)
         if (names[i]==an)
         {
           rp= resp[i];
           C= malloc(B-V);
           memcpy(C, V, B-V);
           *rp= C;
           break;
         }
    }
  }
  rv= 0;

fail:
  free(names);
  free(namep);
  free(resp);
  return rv;
}


#line 198 "/home/dodo/btstack/src/sdp-proto.u"

sdp_protocol_t **sdp_protocol_descriptor_list
      (sdp_conn_t *S, uint32_t svc)
{
  uint8_t *value;
  sdp_protocol_t **R;
  if (sdp_get_attribs(S, svc, &value, 4, NULL)) // 4= protocol descriptor list
       goto fail;

  if (!value) goto fail;
  R= sdp_interp_protodesc(value);
  free(value);
  return R;

fail:
  R= malloc(sizeof(void*)*1);
  R[0]= 0;
  return R;
}

#line 222 "/home/dodo/btstack/src/sdp-proto.u"

void print_proto_descs(sdp_protocol_t **P)
{
  int i,j;
  for(i=0;P[i];i++)
  {
     sdp_protocol_t *L;
     printf("Alt %d\n", i);
     for(L=P[i];L;L=L->lower)
     {
       printf("  proto: %04x params:", L->uuid);
       for(j=0;j<L->paramslen;j++)
          printf(" %02x", L->params[j]);
       printf("\n");
     }
  }
}
#line 17 "/home/dodo/btstack/src/avdtp-lib.u"

avdtp_conn_t *avdtp_open(uint8_t *addr)
{
  avdtp_conn_t *A;
  char buf[200];
  int L;
  A= calloc(1,sizeof(*A));

  memcpy(A->addr, addr, 6);
  A->sock= socket_connect_local(60712);
  if (A->sock->state==SOCKET_ERROR)
  {
    snprintf(A->errstr, sizeof(A->errstr), "can't connect to daemon :%s", 
                             A->sock->errstr);
    A->err= 1;
    return A;
  }
  A->wr= awrite_init(A->sock);
  A->rd= aread_init(A->sock,0);

  L= bpack(buf,200, "pw", 6, addr, 0x0019, NULL);
  awrite_put(A->wr, buf, L,NULL);
  while(A->wr->first)
    if (awrite_go(A->wr))
    {
       snprintf(A->errstr, sizeof(A->errstr),
              "can't write initial command to daemon: %s", A->sock->errstr);
       A->err= 1;
       return A;
    }
  while(!aread_go(A->rd))
     ;

  if (A->rd->error)
    {
       snprintf(A->errstr, sizeof(A->errstr),
              "can't connect to remote device: %s", A->sock->errstr);
       A->err= 1;
       return A;
    }
  return A;
}

#line 64 "/home/dodo/btstack/src/avdtp-lib.u"

int avdtp_discover_endpoints(avdtp_conn_t *A,uint8_t **R, uint32_t *L)
{
  uint8_t B[2];
  avdtp_trans++;
  B[0] = (avdtp_trans<<4) ; // Packet type = 0:single msg type=0 : command
  B[1]= 0x01; // discover

  if (avdtp_wrd(A, B,2))
    return 1;

  if (A->rd->len==2) { *L= 0; return 0; }

  *R= malloc(A->rd->len-2);
  memcpy(*R, A->rd->buffer+2, A->rd->len-2);
  *L= A->rd->len-2;
  return 0;
}

#line 155 "/home/dodo/btstack/src/avdtp-lib.u"

void avdtp_close(avdtp_conn_t *A)
{
  aread_free(A->rd);
  awrite_free(A->wr);
  socket_destroy(A->sock);
  free(A);
}

#line 172 "/home/dodo/btstack/src/avdtp-lib.u"

int avdtp_get_capabilities(avdtp_conn_t *A, uint8_t reid,
                           uint8_t **R, uint32_t *L)
{
  uint8_t B[200];
  avdtp_trans++;
  B[0] = (avdtp_trans<<4) ; // Packet type = 0:single msg type=0 : command
  B[1]= 0x02; // get_capabilities
  B[2]= reid << 2;

  if (avdtp_wrd(A, B, 3))
    return 1;

  if (A->rd->len==2) { *L= 0; return 0; }

  *R= malloc(A->rd->len-2);
  memcpy(*R, A->rd->buffer+2, A->rd->len-2);
  *L= A->rd->len-2;
  return 0;
}

#line 197 "/home/dodo/btstack/src/avdtp-lib.u"

int avdtp_get_sbc_capabilities
     (avdtp_conn_t* A, uint8_t reid, uint8_t *oct0, uint8_t *oct1,
      uint8_t *min_bitpool, uint8_t *max_bitpool)
{
  uint8_t *caps;
  uint32_t capsl;
  uint8_t *end,*cur;

  if (avdtp_get_capabilities(A, reid, &caps, &capsl))
     return 1;

  cur= caps; end= caps+capsl;
  while(cur<end)
  {
    uint8_t sc, losc;
    sc= *(cur++); if (cur==end) break;
    losc= *(cur++); if (cur==end) break;
    if (sc!=7) // codec capabilities
       { cur+= losc; continue; }
    if (cur>=end-1) break;
    if (cur[0]!=0 || cur[1]!=0) // media type=0 , codec= sbc
       { cur+= losc; continue; }
    cur+= 2;
    if (cur>=end-3) break;
    *oct0= cur[0];
    *oct1= cur[1];
    *min_bitpool= cur[2];
    *max_bitpool= cur[3];
    free(caps);
    return 0;
  }
  free(caps);
  return 1;
}

#line 238 "/home/dodo/btstack/src/avdtp-lib.u"

int send_packet(avdtp_conn_t *A, uint8_t *b, uint32_t l)
{
  if (A->err) return 1;
  awrite_put(A->wr, b, l, NULL);
  while(A->wr->first)
   if (awrite_go(A->wr))
   {
     snprintf(A->errstr, sizeof(A->errstr),
              "can't write media packet to daemon: %s", A->sock->errstr);
     A->err= 1;
     return 1;
   }

  return 0;
}

#line 261 "/home/dodo/btstack/src/avdtp-lib.u"

avdtp_conn_t *avdtp_open_sbc_stream
  (avdtp_conn_t *A,uint8_t reid, uint8_t leid,
   uint8_t oct0, uint8_t oct1, uint8_t min_bitpool, uint8_t max_bitpool)
{
  uint8_t B[200];
  int L;
  avdtp_conn_t *K;

  printf(">Set Configuration\n");
  avdtp_trans++;
  L= bpack(B, sizeof(B), 
         "bb", avdtp_trans<<4, 0x03, // set configuration
         "bb", reid<<2, leid<<2, 
         "bb", 1, 0, // media transport cap.
         "blb[", 7, // codec cap.
         "bb", 0, 0, // audio, sbc
         "bb", oct0, oct1, // SBC specific info
         "bb", min_bitpool, max_bitpool,
         "]", NULL);
  if (avdtp_wrd(A, B,L)) return NULL;

  printf(">Open Stream\n");
  avdtp_trans++;
  L= bpack(B, sizeof(B),
          "bb", avdtp_trans<<4, 0x06 , // open_stream
          "b", reid<<2,
          NULL);

  if (avdtp_wrd(A, B,L)) return NULL;

  K= avdtp_open(A->addr);
  if (!K) return NULL;

  printf(">Stream Start\n");
  avdtp_trans++;
  L= bpack(B, sizeof(B),
          "bb", avdtp_trans<<4, 0x07, // stream_start
          "b", reid<<2, NULL);

  if (avdtp_wrd(A, B,L)) return NULL;

  return K;
}

#line 294 "/home/dodo/btstack/src/sdp-lib.u"

