
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>



#define SOCKET_SERVER    1
#define SOCKET_CONNECTED 2
#define SOCKET_ERROR     3

typedef struct {
  int fd;
  int state;
  int syserr;
  char errstr[200];
} socket_t;


socket_t *socket_server_local(int port);
socket_t *socket_accept(socket_t *S);
socket_t *socket_connect_local(int port);
void socket_destroy(socket_t *S);
int socket_recv(socket_t *S,uint8_t *buffer,size_t len) ;
int socket_send(socket_t *S,uint8_t *buffer,size_t len) ;
static void socket_close_fd(socket_t *S);
static socket_t* socket_error(socket_t *S,char *operation);

socket_t *socket_server_local(int port)
{
   socket_t *S;
   struct sockaddr_in server_addr;
   int yes;
   socklen_t size_yes;

   S= malloc(sizeof(*S));

   S->fd= socket(PF_INET, SOCK_STREAM, 0);
   if (S->fd<0) 
     return socket_error(S,"socket");

   yes= 1;
   size_yes= sizeof(yes);
   setsockopt(S->fd, SOL_SOCKET, SO_REUSEADDR, &yes, size_yes);

   memset(&server_addr, 0, sizeof(server_addr));
   server_addr.sin_family= AF_INET;
   inet_aton("127.0.0.1", &server_addr.sin_addr);
   server_addr.sin_port = htons((short)port);

   if (bind(S->fd, (struct sockaddr *) &server_addr, sizeof(server_addr)) < 0) 
      return socket_error(S,"bind");

   listen(S->fd,5);

   S->state= SOCKET_SERVER;

   return S;
}


socket_t *socket_accept(socket_t *S)
{
   socket_t *C;
   socklen_t client_len;
   struct sockaddr_in client_addr;

   if (S->state!=SOCKET_SERVER) return 0;
 
   C= malloc(sizeof(*C));

   client_len = sizeof(client_addr);
   C->fd= accept(S->fd, (struct sockaddr *) &client_addr, &client_len);
   if (C->fd<0) 
     return socket_error(C,"accept");

   C->state= SOCKET_CONNECTED;
   return C;
}


socket_t *socket_connect_local(int port)
{
   socket_t *S;
   struct sockaddr_in addr;
   socklen_t len;

   S= malloc(sizeof(*S));

   S->fd= socket(PF_INET, SOCK_STREAM, 0);
   if (S->fd<0) 
      return socket_error(S,"socket");

   len= sizeof(addr);
   memset(&addr, 0, sizeof(addr));
   addr.sin_family= AF_INET;
   inet_aton("127.0.0.1",&addr.sin_addr);
   addr.sin_port= htons( (short) port);
   if (connect(S->fd,(struct sockaddr*) &addr, len)) 
      return socket_error(S,"connect");

   S->state= SOCKET_CONNECTED;
   return S;
}


void socket_destroy(socket_t *S)
{
  socket_close_fd(S);
  free(S);
}


int socket_recv(socket_t *S,uint8_t *buffer,size_t len) {
    int r;

    if (S->state!=SOCKET_CONNECTED) 
     goto invalid_state;

again:
    r= recv(S->fd, buffer, len, MSG_NOSIGNAL);
    if (r<=0) {
      if (errno==EINTR) goto again;
      socket_error(S, "recv");
      return -1;
    }
    return r;

invalid_state:
    snprintf(S->errstr,sizeof(S->errstr), 
        "recv: socket is closed or is a server socket");
    socket_close_fd(S);
    S->syserr= EINVAL;
    S->state= SOCKET_ERROR;
    return -1;
}


int socket_send(socket_t *S,uint8_t *buffer,size_t len) {
    int r;

    if (S->state!=SOCKET_CONNECTED) 
     goto invalid_state;

again:
    r= send(S->fd, buffer, len, MSG_NOSIGNAL);
    if (r<=0) {
      if (errno==EINTR) goto again;
      socket_error(S, "send");
      return -1;
    }
    return r;

invalid_state:
    snprintf(S->errstr,sizeof(S->errstr), 
        "send: socket is closed or is a server socket");
    socket_close_fd(S);
    S->syserr= EINVAL;
    S->state= SOCKET_ERROR;
    return -1;
}


static void socket_close_fd(socket_t *S)
{
    if (S->fd>=0) {
      shutdown(S->fd,3);
      close(S->fd);
      S->fd= -1;
    }
}

static socket_t* socket_error(socket_t *S,char *operation)
{
    S->syserr= errno;
    snprintf(S->errstr, sizeof(S->errstr), 
             "%s: %s", operation, strerror(S->syserr));
    S->state= SOCKET_ERROR;

    socket_close_fd(S);
    return S;
}


