#include <pthread.h>
#include <stdlib.h>
#include <unistd.h>
#include "mbox.h"

struct mbox {
  pthread_mutex_t lock;
  int len, head,size;
  void **data;
  int fd[2];
};

mbox_t *mbox_new()
{
   mbox_t *M;
   M= malloc(sizeof(*M));
   M->size= 64;
   M->data= malloc(M->size*sizeof(void*));
   M->head=0; M->len= 0;
   if (pthread_mutex_init(&M->lock, 0)) goto out_data;
   if (pipe(M->fd)) goto out_lock;
   return M;
out_lock:
   pthread_mutex_destroy(&M->lock);
out_data:
   free(M->data);
   free(M);
   return 0;
}

int mbox_fd(mbox_t *m)
{
   return m->fd[0];
}

void mbox_put(mbox_t *m, void *data)
{
   pthread_mutex_lock(&m->lock);
   if (m->len==m->size)
   {
      void **ND; int NS,i;
      NS= m->size*2;
      ND= malloc(sizeof(void*)*NS);
      for(i=0;i<m->len;i++)
        ND[i]= m->data[(m->head+i)%m->size];
      free(m->data);
      m->head= 0;
      m->size= NS;
      m->data= ND;
   }
   m->data[(m->head+m->len)%m->size]= data;
   m->len++;
   pthread_mutex_unlock(&m->lock); 
   write(m->fd[1], data, 1);
}

void* mbox_get(mbox_t *m)
{
   char c;
   void *R;
again:
   read(m->fd[0], &c, 1);
   pthread_mutex_lock(&m->lock);
   if (m->len==0) {
      // the message was cleared, but we couldn't clear the byte in 
      // the pipe of course.
      pthread_mutex_unlock(&m->lock);
      goto again;
   }
   R= m->data[m->head];
   m->head= (m->head+1)%m->size;
   m->len--;
   pthread_mutex_unlock(&m->lock);
   return R;
}

void mbox_clear(mbox_t *m)
{
   pthread_mutex_lock(&m->lock);
   m->len= 0; m->head= 0;
   pthread_mutex_unlock(&m->lock);
}

int mbox_clear_to_nb(mbox_t *m, int tag)
{
   int i,f=0;
   pthread_mutex_lock(&m->lock);
   for(i=0;i<m->len;i++)
   {
      int *K= m->data[(i+m->head)%m->size];
      if (*K==tag) { f= 1; break; }
   }

   if (f)  {
     m->head= (m->head+i+1)%m->size;
     m->len-= (i+1);
   }

   pthread_mutex_unlock(&m->lock);
   return f; 
}
