
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <limits.h>
#include <fx.h>
#include <stdint.h>
#include <FXPNGIcon.h>

static int app_created;

#include "gui.h"


#include "actions.h"
#include "mbox.h"
#include "msg_dpy.h"

extern mbox_t* dpy_mbox;


 

static void button_set_text(gui_button_t *B,const char *text)
{
  FXButton *W= (FXButton*) B->widget;
  W->setText(text);
}

static gui_button_class_t
 gui_button_class= {
  button_set_text,
};




class FYListItem : public FXListItem 
{
  FXDECLARE(FYListItem)
protected:
  FYListItem() : FXListItem() { }
public:
  FYListItem(const FXString& text,FXIcon* ic, void *pdata)
   : FXListItem(text,ic,NULL)
  { data_p= pdata; }
  FYListItem(const FXString& text,FXIcon* ic, int idata)
   : FXListItem(text,ic,NULL)
  { data_i= idata; }
  FYListItem(const FXString& text,FXIcon* ic, double fdata)
   : FXListItem(text,ic,NULL)
  { data_f= fdata; }
  ~FYListItem() { }
// These could be in a union, but who cares about space
  int data_i;
  double data_f;
  void *data_p;
};


FXIMPLEMENT(FYListItem,FXListItem,NULL,0)




static void clear_selection
  (FXList *L)
{
  int i;
  int N= L->getNumItems();
  for(i=0;i<N;i++)
    L->deselectItem(i);
}

static FYListItem* get_selected(FXList *L)
{
   int i, N= L->getNumItems();
   for(i=0;i<N;i++)
     if (L->isItemSelected(i))
       return (FYListItem*) L->getItem(i);
   return 0;
}




static inline FXList* list(gui_intlist_t *IL)
{
  return (FXList*) (IL->widget);
}

static int intlist_get_selected
   (gui_intlist_t* IL, int *ret)
{
  FYListItem *item= get_selected(list(IL));
  if (item) { *ret= item->data_i; return 1; }
  return 0;
}

static int intlist_find_item(gui_intlist_t *IL, int key)
{
  int i;
  FXList* L= list(IL);
  int N= L->getNumItems();
  for(i=0;i<N;i++)
    {
        FYListItem *it= (FYListItem*) L->getItem(i);
        if (it->data_i==key) return i;
    }
  return -1;
}

static int intlist_select_item
  (gui_intlist_t *IL, int key,int sel)
{
  int i= intlist_find_item(IL, key); 
  if (i==-1) return 0;
  if (sel) list(IL)->selectItem(i);
  else list(IL)->deselectItem(i); 
  return 1;
}


static void intlist_append_item
  (gui_intlist_t *IL, const char *text, int key)
{
  list(IL)->appendItem(new FYListItem(text, NULL,key));
}

static void intlist_prepend_item
  (gui_intlist_t *IL, const char *text, int key)
{
  list(IL)->prependItem(new FYListItem(text, NULL,key));
}

static void intlist_insert_item
  (gui_intlist_t *IL, int poskey, const char *text, int key)
{
  int i= intlist_find_item(IL, poskey);
  if (i!=-1)
    list(IL)->insertItem(i, new FYListItem(text, NULL, key));
  else
    list(IL)->appendItem(new FYListItem(text, NULL, key));
}

static void intlist_clear_selection(gui_intlist_t *IL)
{
  clear_selection(list(IL));
}

static void intlist_remove_item(gui_intlist_t *IL, int key)
{
  int i= intlist_find_item(IL, key);
  if (i==-1) return ;
  list(IL)->removeItem(i);
}

static gui_intlist_class_t gui_intlist_class=
{
   intlist_get_selected,
   intlist_select_item,
   intlist_clear_selection,
   intlist_append_item,
   intlist_prepend_item,
   intlist_insert_item,
   intlist_remove_item,
};



static int item_type(FXFileList *L, int item)
{
  if (L->isItemDirectory(item)) return 'D';
  return 'F';
}

static int filelist_get_current
  (gui_file_list_t* DL, const char **Rname, int *Rtype)
{
  FXFileList *L= (FXFileList*) DL->widget;
  FXString cur;
  int item;

  item= L->getCurrentItem();
  if (item<0) return 0;
  cur= L->getCurrentFile();
  if (Rname) *Rname= strdup(cur.text());
  if (Rtype) *Rtype= item_type(L,item);
  return 1;
}

static void filelist_set_current_file
  (gui_file_list_t *FL, const char *fn)
{
  FXFileList *L= (FXFileList*) FL->widget;
  L->setCurrentFile(fn);
}

static void filelist_set_current_directory
  (gui_file_list_t *DL, const char *fn)
{
  FXFileList *L= (FXFileList*) DL->widget;
  L->setDirectory(fn);
}

static void filelist_get_current_directory
  (gui_file_list_t* DL, const char **Rname)
{
  FXFileList *L= (FXFileList*) DL->widget;
  FXString cur;

  cur= L->getDirectory();
  if (Rname) *Rname= strdup(cur.text());
}

static gui_file_list_class_t gui_file_list_class= {
   filelist_get_current,
   filelist_set_current_file,
   filelist_set_current_directory,
   filelist_get_current_directory,
};


 
 
  static FXIcon *VFLL1;
  static unsigned char YQRZ_2 []= {
0x89,0x50,0x4e,0x47,0x0d,0x0a,0x1a,0x0a,0x00,0x00,
0x00,0x0d,0x49,0x48,0x44,0x52,0x00,0x00,0x00,0x20,
0x00,0x00,0x00,0x20,0x08,0x06,0x00,0x00,0x00,0x73,
0x7a,0x7a,0xf4,0x00,0x00,0x00,0x19,0x74,0x45,0x58,
0x74,0x53,0x6f,0x66,0x74,0x77,0x61,0x72,0x65,0x00,
0x41,0x64,0x6f,0x62,0x65,0x20,0x49,0x6d,0x61,0x67,
0x65,0x52,0x65,0x61,0x64,0x79,0x71,0xc9,0x65,0x3c,
0x00,0x00,0x03,0xad,0x69,0x54,0x58,0x74,0x58,0x4d,
0x4c,0x3a,0x63,0x6f,0x6d,0x2e,0x61,0x64,0x6f,0x62,
0x65,0x2e,0x78,0x6d,0x70,0x00,0x00,0x00,0x00,0x00,
0x3c,0x3f,0x78,0x70,0x61,0x63,0x6b,0x65,0x74,0x20,
0x62,0x65,0x67,0x69,0x6e,0x3d,0x22,0xef,0xbb,0xbf,
0x22,0x20,0x69,0x64,0x3d,0x22,0x57,0x35,0x4d,0x30,
0x4d,0x70,0x43,0x65,0x68,0x69,0x48,0x7a,0x72,0x65,
0x53,0x7a,0x4e,0x54,0x63,0x7a,0x6b,0x63,0x39,0x64,
0x22,0x3f,0x3e,0x20,0x3c,0x78,0x3a,0x78,0x6d,0x70,
0x6d,0x65,0x74,0x61,0x20,0x78,0x6d,0x6c,0x6e,0x73,
0x3a,0x78,0x3d,0x22,0x61,0x64,0x6f,0x62,0x65,0x3a,
0x6e,0x73,0x3a,0x6d,0x65,0x74,0x61,0x2f,0x22,0x20,
0x78,0x3a,0x78,0x6d,0x70,0x74,0x6b,0x3d,0x22,0x41,
0x64,0x6f,0x62,0x65,0x20,0x58,0x4d,0x50,0x20,0x43,
0x6f,0x72,0x65,0x20,0x35,0x2e,0x30,0x2d,0x63,0x30,
0x36,0x30,0x20,0x36,0x31,0x2e,0x31,0x33,0x34,0x37,
0x37,0x37,0x2c,0x20,0x32,0x30,0x31,0x30,0x2f,0x30,
0x32,0x2f,0x31,0x32,0x2d,0x31,0x37,0x3a,0x33,0x32,
0x3a,0x30,0x30,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x22,0x3e,0x20,0x3c,0x72,0x64,0x66,0x3a,0x52,
0x44,0x46,0x20,0x78,0x6d,0x6c,0x6e,0x73,0x3a,0x72,
0x64,0x66,0x3d,0x22,0x68,0x74,0x74,0x70,0x3a,0x2f,
0x2f,0x77,0x77,0x77,0x2e,0x77,0x33,0x2e,0x6f,0x72,
0x67,0x2f,0x31,0x39,0x39,0x39,0x2f,0x30,0x32,0x2f,
0x32,0x32,0x2d,0x72,0x64,0x66,0x2d,0x73,0x79,0x6e,
0x74,0x61,0x78,0x2d,0x6e,0x73,0x23,0x22,0x3e,0x20,
0x3c,0x72,0x64,0x66,0x3a,0x44,0x65,0x73,0x63,0x72,
0x69,0x70,0x74,0x69,0x6f,0x6e,0x20,0x72,0x64,0x66,
0x3a,0x61,0x62,0x6f,0x75,0x74,0x3d,0x22,0x22,0x20,
0x78,0x6d,0x6c,0x6e,0x73,0x3a,0x78,0x6d,0x70,0x52,
0x69,0x67,0x68,0x74,0x73,0x3d,0x22,0x68,0x74,0x74,
0x70,0x3a,0x2f,0x2f,0x6e,0x73,0x2e,0x61,0x64,0x6f,
0x62,0x65,0x2e,0x63,0x6f,0x6d,0x2f,0x78,0x61,0x70,
0x2f,0x31,0x2e,0x30,0x2f,0x72,0x69,0x67,0x68,0x74,
0x73,0x2f,0x22,0x20,0x78,0x6d,0x6c,0x6e,0x73,0x3a,
0x78,0x6d,0x70,0x4d,0x4d,0x3d,0x22,0x68,0x74,0x74,
0x70,0x3a,0x2f,0x2f,0x6e,0x73,0x2e,0x61,0x64,0x6f,
0x62,0x65,0x2e,0x63,0x6f,0x6d,0x2f,0x78,0x61,0x70,
0x2f,0x31,0x2e,0x30,0x2f,0x6d,0x6d,0x2f,0x22,0x20,
0x78,0x6d,0x6c,0x6e,0x73,0x3a,0x73,0x74,0x52,0x65,
0x66,0x3d,0x22,0x68,0x74,0x74,0x70,0x3a,0x2f,0x2f,
0x6e,0x73,0x2e,0x61,0x64,0x6f,0x62,0x65,0x2e,0x63,
0x6f,0x6d,0x2f,0x78,0x61,0x70,0x2f,0x31,0x2e,0x30,
0x2f,0x73,0x54,0x79,0x70,0x65,0x2f,0x52,0x65,0x73,
0x6f,0x75,0x72,0x63,0x65,0x52,0x65,0x66,0x23,0x22,
0x20,0x78,0x6d,0x6c,0x6e,0x73,0x3a,0x78,0x6d,0x70,
0x3d,0x22,0x68,0x74,0x74,0x70,0x3a,0x2f,0x2f,0x6e,
0x73,0x2e,0x61,0x64,0x6f,0x62,0x65,0x2e,0x63,0x6f,
0x6d,0x2f,0x78,0x61,0x70,0x2f,0x31,0x2e,0x30,0x2f,
0x22,0x20,0x78,0x6d,0x70,0x52,0x69,0x67,0x68,0x74,
0x73,0x3a,0x4d,0x61,0x72,0x6b,0x65,0x64,0x3d,0x22,
0x46,0x61,0x6c,0x73,0x65,0x22,0x20,0x78,0x6d,0x70,
0x4d,0x4d,0x3a,0x4f,0x72,0x69,0x67,0x69,0x6e,0x61,
0x6c,0x44,0x6f,0x63,0x75,0x6d,0x65,0x6e,0x74,0x49,
0x44,0x3d,0x22,0x75,0x75,0x69,0x64,0x3a,0x41,0x45,
0x33,0x46,0x34,0x36,0x30,0x38,0x38,0x46,0x38,0x39,
0x44,0x45,0x31,0x31,0x41,0x38,0x36,0x30,0x39,0x39,
0x46,0x31,0x32,0x31,0x41,0x32,0x30,0x34,0x32,0x39,
0x22,0x20,0x78,0x6d,0x70,0x4d,0x4d,0x3a,0x44,0x6f,
0x63,0x75,0x6d,0x65,0x6e,0x74,0x49,0x44,0x3d,0x22,
0x78,0x6d,0x70,0x2e,0x64,0x69,0x64,0x3a,0x39,0x31,
0x35,0x33,0x37,0x43,0x37,0x46,0x42,0x35,0x45,0x46,
0x31,0x31,0x45,0x30,0x42,0x41,0x41,0x33,0x42,0x42,
0x45,0x34,0x44,0x37,0x42,0x31,0x33,0x35,0x31,0x44,
0x22,0x20,0x78,0x6d,0x70,0x4d,0x4d,0x3a,0x49,0x6e,
0x73,0x74,0x61,0x6e,0x63,0x65,0x49,0x44,0x3d,0x22,
0x78,0x6d,0x70,0x2e,0x69,0x69,0x64,0x3a,0x39,0x31,
0x35,0x33,0x37,0x43,0x37,0x45,0x42,0x35,0x45,0x46,
0x31,0x31,0x45,0x30,0x42,0x41,0x41,0x33,0x42,0x42,
0x45,0x34,0x44,0x37,0x42,0x31,0x33,0x35,0x31,0x44,
0x22,0x20,0x78,0x6d,0x70,0x3a,0x43,0x72,0x65,0x61,
0x74,0x6f,0x72,0x54,0x6f,0x6f,0x6c,0x3d,0x22,0x41,
0x64,0x6f,0x62,0x65,0x20,0x50,0x68,0x6f,0x74,0x6f,
0x73,0x68,0x6f,0x70,0x20,0x43,0x53,0x35,0x20,0x57,
0x69,0x6e,0x64,0x6f,0x77,0x73,0x22,0x3e,0x20,0x3c,
0x78,0x6d,0x70,0x4d,0x4d,0x3a,0x44,0x65,0x72,0x69,
0x76,0x65,0x64,0x46,0x72,0x6f,0x6d,0x20,0x73,0x74,
0x52,0x65,0x66,0x3a,0x69,0x6e,0x73,0x74,0x61,0x6e,
0x63,0x65,0x49,0x44,0x3d,0x22,0x78,0x6d,0x70,0x2e,
0x69,0x69,0x64,0x3a,0x33,0x32,0x45,0x42,0x39,0x34,
0x34,0x44,0x45,0x46,0x42,0x35,0x45,0x30,0x31,0x31,
0x39,0x41,0x42,0x34,0x41,0x46,0x44,0x39,0x30,0x33,
0x39,0x46,0x33,0x35,0x45,0x42,0x22,0x20,0x73,0x74,
0x52,0x65,0x66,0x3a,0x64,0x6f,0x63,0x75,0x6d,0x65,
0x6e,0x74,0x49,0x44,0x3d,0x22,0x75,0x75,0x69,0x64,
0x3a,0x41,0x45,0x33,0x46,0x34,0x36,0x30,0x38,0x38,
0x46,0x38,0x39,0x44,0x45,0x31,0x31,0x41,0x38,0x36,
0x30,0x39,0x39,0x46,0x31,0x32,0x31,0x41,0x32,0x30,
0x34,0x32,0x39,0x22,0x2f,0x3e,0x20,0x3c,0x2f,0x72,
0x64,0x66,0x3a,0x44,0x65,0x73,0x63,0x72,0x69,0x70,
0x74,0x69,0x6f,0x6e,0x3e,0x20,0x3c,0x2f,0x72,0x64,
0x66,0x3a,0x52,0x44,0x46,0x3e,0x20,0x3c,0x2f,0x78,
0x3a,0x78,0x6d,0x70,0x6d,0x65,0x74,0x61,0x3e,0x20,
0x3c,0x3f,0x78,0x70,0x61,0x63,0x6b,0x65,0x74,0x20,
0x65,0x6e,0x64,0x3d,0x22,0x72,0x22,0x3f,0x3e,0x4f,
0x8c,0x38,0x3d,0x00,0x00,0x02,0x62,0x49,0x44,0x41,
0x54,0x78,0xda,0xe4,0x97,0x41,0x4b,0x1b,0x41,0x14,
0x80,0xdf,0x4a,0x0e,0xa5,0x22,0x11,0xc1,0x43,0xa5,
0x48,0x41,0x83,0x3f,0xa1,0x87,0x05,0x8b,0x88,0x14,
0xe9,0xc5,0x8b,0xfe,0x81,0x7a,0xf0,0x07,0xd4,0xb6,
0xb1,0x8a,0xa6,0x1a,0x0d,0x88,0xff,0xc0,0xb3,0x78,
0xf0,0x2a,0x22,0xa2,0x2e,0x82,0xa0,0x77,0x4f,0x09,
0x85,0xb6,0x20,0x16,0xaa,0x20,0x22,0xed,0x41,0x37,
0xaf,0xb3,0xeb,0x24,0xee,0xcc,0x7b,0x6f,0x33,0x97,
0xc4,0x43,0x67,0xb3,0xcc,0x64,0xe7,0xcd,0x7c,0x6f,
0xde,0x9b,0x79,0x33,0xe3,0x21,0x22,0x3c,0x65,0x6a,
0x83,0x27,0x4e,0x99,0x5a,0x61,0xed,0x6c,0x1c,0x10,
0x42,0x00,0xac,0xce,0xab,0x7c,0x01,0x31,0x54,0xff,
0xef,0xd5,0x7f,0x95,0xe3,0x7d,0x5c,0x17,0xe7,0x71,
0x39,0xca,0x43,0xa6,0x1c,0xea,0x72,0xa8,0xcb,0xd5,
0xb8,0x0c,0x51,0x0e,0x51,0x19,0x55,0xbf,0x58,0x88,
0xbe,0xaf,0x0e,0xde,0x99,0x16,0xd0,0x70,0x6c,0x22,
0x5c,0x2b,0xa0,0x18,0x09,0xb7,0x3f,0xba,0xe0,0x61,
0xe4,0xd0,0x44,0xb8,0xfe,0x1e,0x97,0xe7,0x89,0x02,
0x4d,0x1e,0x79,0x12,0xae,0xbe,0xc3,0x02,0x55,0xa0,
0x75,0x70,0x48,0xba,0x20,0xf3,0x68,0x01,0x1e,0x3e,
0xd1,0xb7,0x06,0xb9,0xec,0x9b,0x58,0xa6,0x7c,0x7d,
0x00,0x1b,0x95,0x49,0x27,0x78,0xc9,0xff,0x1b,0xb7,
0x39,0xb9,0x58,0x87,0xad,0xf2,0x94,0x09,0x47,0x6e,
0x19,0x0a,0x23,0xef,0xcf,0x0e,0x02,0xea,0x27,0xd7,
0x39,0xe4,0x04,0x5f,0xf1,0xff,0xd4,0xdb,0xbc,0x7e,
0xf1,0x9e,0xc0,0x79,0x0b,0x08,0x66,0x07,0xd5,0x61,
0x32,0x35,0x86,0xdf,0x32,0x6d,0x4c,0x78,0x32,0xf6,
0x65,0xcc,0x8e,0x39,0x9f,0x5b,0x9d,0xa5,0xc0,0x97,
0xfd,0x1b,0x22,0xff,0x21,0x78,0x46,0xe0,0xbc,0x02,
0xd2,0x84,0xe3,0x14,0x60,0xe0,0x45,0xff,0x9a,0xc8,
0x4e,0x07,0xcf,0x59,0xb8,0xac,0x00,0x37,0xdb,0xa3,
0x00,0x45,0x5c,0x60,0xc2,0x97,0xfc,0x2b,0x22,0xf7,
0x31,0xe8,0x10,0xe1,0x82,0x0b,0xa4,0xa5,0x66,0x5b,
0xc0,0x84,0x2f,0xfa,0xbf,0x89,0xcc,0xa7,0x20,0x9b,
0x0a,0x17,0x2c,0x20,0xac,0x73,0xdb,0x02,0x09,0xf8,
0x57,0xff,0x17,0xa9,0xff,0x1c,0x74,0x35,0x84,0x3b,
0xba,0x20,0x64,0x5d,0x50,0x83,0x17,0xfc,0x73,0x52,
0x97,0x0f,0xba,0x9d,0xe0,0x0e,0x2e,0x30,0x97,0x9a,
0x39,0x07,0x74,0x54,0xb3,0x15,0x63,0x96,0x9a,0x08,
0xe7,0x02,0x91,0x04,0xaf,0x5a,0x4f,0x2d,0xa4,0xce,
0x1d,0xf5,0x90,0xba,0xe2,0xd0,0x85,0x13,0x1c,0x79,
0x05,0x84,0x08,0xa7,0xe7,0x41,0xfd,0x4d,0xc4,0xf6,
0xb9,0xa0,0x87,0xd4,0x17,0x87,0xcf,0x1b,0xc2,0x91,
0x3b,0x11,0x49,0x11,0x8e,0x2a,0x60,0x6e,0x2c,0xb3,
0x87,0x2f,0x89,0xcc,0xf2,0xc8,0xcf,0x74,0xb8,0xec,
0x02,0x61,0x57,0x4b,0xbe,0xf6,0xae,0xa6,0x7e,0xb3,
0xfb,0xbd,0x44,0x6e,0xe5,0xed,0x77,0x11,0x8e,0xa2,
0x02,0x2c,0x9c,0x71,0x01,0x33,0xe1,0xbe,0xec,0xbd,
0x22,0xb2,0xa5,0xd1,0x6f,0x2c,0x3c,0xc5,0x05,0xcc,
0x7e,0x6e,0x5b,0x20,0x65,0xb6,0xcf,0xec,0xf6,0x11,
0xf9,0xd2,0xbb,0x0a,0x85,0xf3,0x16,0x10,0x0e,0x13,
0x8c,0x0b,0xd2,0x96,0x5a,0x7e,0x27,0x47,0xda,0x38,
0xbb,0x80,0x3b,0xc9,0x54,0xae,0x0e,0xf4,0xf6,0x5a,
0x85,0xf2,0xe5,0xbe,0xd3,0x3a,0xcf,0x6f,0x0f,0xd4,
0xdb,0x9c,0xfe,0xd8,0x14,0xe1,0x51,0xf2,0x6a,0x87,
0x83,0xfc,0x71,0x3b,0x36,0x3c,0x46,0x39,0x06,0x19,
0xd6,0xe7,0x16,0x7c,0x7d,0x0c,0x3d,0x6b,0x0e,0xb4,
0x0e,0xce,0xde,0x8c,0xf4,0xa5,0xa1,0x55,0x70,0xee,
0x54,0xfc,0x70,0x63,0x69,0xd1,0xc8,0x0b,0xc2,0x2a,
0x40,0x2f,0xb6,0x44,0x73,0x47,0xee,0xb1,0x93,0xf0,
0xbf,0xbd,0x1d,0xff,0x13,0x60,0x00,0xb6,0xcc,0x2e,
0x5a,0x3f,0xe9,0x6f,0x22,0x00,0x00,0x00,0x00,0x49,
0x45,0x4e,0x44,0xae,0x42,0x60,0x82,

  };

  static FXIcon *VFLL2;
  static unsigned char YQRZ_8 []= {
0x89,0x50,0x4e,0x47,0x0d,0x0a,0x1a,0x0a,0x00,0x00,
0x00,0x0d,0x49,0x48,0x44,0x52,0x00,0x00,0x00,0x20,
0x00,0x00,0x00,0x20,0x08,0x06,0x00,0x00,0x00,0x73,
0x7a,0x7a,0xf4,0x00,0x00,0x00,0x19,0x74,0x45,0x58,
0x74,0x53,0x6f,0x66,0x74,0x77,0x61,0x72,0x65,0x00,
0x41,0x64,0x6f,0x62,0x65,0x20,0x49,0x6d,0x61,0x67,
0x65,0x52,0x65,0x61,0x64,0x79,0x71,0xc9,0x65,0x3c,
0x00,0x00,0x03,0xad,0x69,0x54,0x58,0x74,0x58,0x4d,
0x4c,0x3a,0x63,0x6f,0x6d,0x2e,0x61,0x64,0x6f,0x62,
0x65,0x2e,0x78,0x6d,0x70,0x00,0x00,0x00,0x00,0x00,
0x3c,0x3f,0x78,0x70,0x61,0x63,0x6b,0x65,0x74,0x20,
0x62,0x65,0x67,0x69,0x6e,0x3d,0x22,0xef,0xbb,0xbf,
0x22,0x20,0x69,0x64,0x3d,0x22,0x57,0x35,0x4d,0x30,
0x4d,0x70,0x43,0x65,0x68,0x69,0x48,0x7a,0x72,0x65,
0x53,0x7a,0x4e,0x54,0x63,0x7a,0x6b,0x63,0x39,0x64,
0x22,0x3f,0x3e,0x20,0x3c,0x78,0x3a,0x78,0x6d,0x70,
0x6d,0x65,0x74,0x61,0x20,0x78,0x6d,0x6c,0x6e,0x73,
0x3a,0x78,0x3d,0x22,0x61,0x64,0x6f,0x62,0x65,0x3a,
0x6e,0x73,0x3a,0x6d,0x65,0x74,0x61,0x2f,0x22,0x20,
0x78,0x3a,0x78,0x6d,0x70,0x74,0x6b,0x3d,0x22,0x41,
0x64,0x6f,0x62,0x65,0x20,0x58,0x4d,0x50,0x20,0x43,
0x6f,0x72,0x65,0x20,0x35,0x2e,0x30,0x2d,0x63,0x30,
0x36,0x30,0x20,0x36,0x31,0x2e,0x31,0x33,0x34,0x37,
0x37,0x37,0x2c,0x20,0x32,0x30,0x31,0x30,0x2f,0x30,
0x32,0x2f,0x31,0x32,0x2d,0x31,0x37,0x3a,0x33,0x32,
0x3a,0x30,0x30,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x22,0x3e,0x20,0x3c,0x72,0x64,0x66,0x3a,0x52,
0x44,0x46,0x20,0x78,0x6d,0x6c,0x6e,0x73,0x3a,0x72,
0x64,0x66,0x3d,0x22,0x68,0x74,0x74,0x70,0x3a,0x2f,
0x2f,0x77,0x77,0x77,0x2e,0x77,0x33,0x2e,0x6f,0x72,
0x67,0x2f,0x31,0x39,0x39,0x39,0x2f,0x30,0x32,0x2f,
0x32,0x32,0x2d,0x72,0x64,0x66,0x2d,0x73,0x79,0x6e,
0x74,0x61,0x78,0x2d,0x6e,0x73,0x23,0x22,0x3e,0x20,
0x3c,0x72,0x64,0x66,0x3a,0x44,0x65,0x73,0x63,0x72,
0x69,0x70,0x74,0x69,0x6f,0x6e,0x20,0x72,0x64,0x66,
0x3a,0x61,0x62,0x6f,0x75,0x74,0x3d,0x22,0x22,0x20,
0x78,0x6d,0x6c,0x6e,0x73,0x3a,0x78,0x6d,0x70,0x52,
0x69,0x67,0x68,0x74,0x73,0x3d,0x22,0x68,0x74,0x74,
0x70,0x3a,0x2f,0x2f,0x6e,0x73,0x2e,0x61,0x64,0x6f,
0x62,0x65,0x2e,0x63,0x6f,0x6d,0x2f,0x78,0x61,0x70,
0x2f,0x31,0x2e,0x30,0x2f,0x72,0x69,0x67,0x68,0x74,
0x73,0x2f,0x22,0x20,0x78,0x6d,0x6c,0x6e,0x73,0x3a,
0x78,0x6d,0x70,0x4d,0x4d,0x3d,0x22,0x68,0x74,0x74,
0x70,0x3a,0x2f,0x2f,0x6e,0x73,0x2e,0x61,0x64,0x6f,
0x62,0x65,0x2e,0x63,0x6f,0x6d,0x2f,0x78,0x61,0x70,
0x2f,0x31,0x2e,0x30,0x2f,0x6d,0x6d,0x2f,0x22,0x20,
0x78,0x6d,0x6c,0x6e,0x73,0x3a,0x73,0x74,0x52,0x65,
0x66,0x3d,0x22,0x68,0x74,0x74,0x70,0x3a,0x2f,0x2f,
0x6e,0x73,0x2e,0x61,0x64,0x6f,0x62,0x65,0x2e,0x63,
0x6f,0x6d,0x2f,0x78,0x61,0x70,0x2f,0x31,0x2e,0x30,
0x2f,0x73,0x54,0x79,0x70,0x65,0x2f,0x52,0x65,0x73,
0x6f,0x75,0x72,0x63,0x65,0x52,0x65,0x66,0x23,0x22,
0x20,0x78,0x6d,0x6c,0x6e,0x73,0x3a,0x78,0x6d,0x70,
0x3d,0x22,0x68,0x74,0x74,0x70,0x3a,0x2f,0x2f,0x6e,
0x73,0x2e,0x61,0x64,0x6f,0x62,0x65,0x2e,0x63,0x6f,
0x6d,0x2f,0x78,0x61,0x70,0x2f,0x31,0x2e,0x30,0x2f,
0x22,0x20,0x78,0x6d,0x70,0x52,0x69,0x67,0x68,0x74,
0x73,0x3a,0x4d,0x61,0x72,0x6b,0x65,0x64,0x3d,0x22,
0x46,0x61,0x6c,0x73,0x65,0x22,0x20,0x78,0x6d,0x70,
0x4d,0x4d,0x3a,0x4f,0x72,0x69,0x67,0x69,0x6e,0x61,
0x6c,0x44,0x6f,0x63,0x75,0x6d,0x65,0x6e,0x74,0x49,
0x44,0x3d,0x22,0x75,0x75,0x69,0x64,0x3a,0x41,0x45,
0x33,0x46,0x34,0x36,0x30,0x38,0x38,0x46,0x38,0x39,
0x44,0x45,0x31,0x31,0x41,0x38,0x36,0x30,0x39,0x39,
0x46,0x31,0x32,0x31,0x41,0x32,0x30,0x34,0x32,0x39,
0x22,0x20,0x78,0x6d,0x70,0x4d,0x4d,0x3a,0x44,0x6f,
0x63,0x75,0x6d,0x65,0x6e,0x74,0x49,0x44,0x3d,0x22,
0x78,0x6d,0x70,0x2e,0x64,0x69,0x64,0x3a,0x39,0x30,
0x46,0x42,0x34,0x35,0x37,0x37,0x42,0x35,0x45,0x46,
0x31,0x31,0x45,0x30,0x42,0x41,0x41,0x33,0x42,0x42,
0x45,0x34,0x44,0x37,0x42,0x31,0x33,0x35,0x31,0x44,
0x22,0x20,0x78,0x6d,0x70,0x4d,0x4d,0x3a,0x49,0x6e,
0x73,0x74,0x61,0x6e,0x63,0x65,0x49,0x44,0x3d,0x22,
0x78,0x6d,0x70,0x2e,0x69,0x69,0x64,0x3a,0x39,0x30,
0x46,0x42,0x34,0x35,0x37,0x36,0x42,0x35,0x45,0x46,
0x31,0x31,0x45,0x30,0x42,0x41,0x41,0x33,0x42,0x42,
0x45,0x34,0x44,0x37,0x42,0x31,0x33,0x35,0x31,0x44,
0x22,0x20,0x78,0x6d,0x70,0x3a,0x43,0x72,0x65,0x61,
0x74,0x6f,0x72,0x54,0x6f,0x6f,0x6c,0x3d,0x22,0x41,
0x64,0x6f,0x62,0x65,0x20,0x50,0x68,0x6f,0x74,0x6f,
0x73,0x68,0x6f,0x70,0x20,0x43,0x53,0x35,0x20,0x57,
0x69,0x6e,0x64,0x6f,0x77,0x73,0x22,0x3e,0x20,0x3c,
0x78,0x6d,0x70,0x4d,0x4d,0x3a,0x44,0x65,0x72,0x69,
0x76,0x65,0x64,0x46,0x72,0x6f,0x6d,0x20,0x73,0x74,
0x52,0x65,0x66,0x3a,0x69,0x6e,0x73,0x74,0x61,0x6e,
0x63,0x65,0x49,0x44,0x3d,0x22,0x78,0x6d,0x70,0x2e,
0x69,0x69,0x64,0x3a,0x33,0x32,0x45,0x42,0x39,0x34,
0x34,0x44,0x45,0x46,0x42,0x35,0x45,0x30,0x31,0x31,
0x39,0x41,0x42,0x34,0x41,0x46,0x44,0x39,0x30,0x33,
0x39,0x46,0x33,0x35,0x45,0x42,0x22,0x20,0x73,0x74,
0x52,0x65,0x66,0x3a,0x64,0x6f,0x63,0x75,0x6d,0x65,
0x6e,0x74,0x49,0x44,0x3d,0x22,0x75,0x75,0x69,0x64,
0x3a,0x41,0x45,0x33,0x46,0x34,0x36,0x30,0x38,0x38,
0x46,0x38,0x39,0x44,0x45,0x31,0x31,0x41,0x38,0x36,
0x30,0x39,0x39,0x46,0x31,0x32,0x31,0x41,0x32,0x30,
0x34,0x32,0x39,0x22,0x2f,0x3e,0x20,0x3c,0x2f,0x72,
0x64,0x66,0x3a,0x44,0x65,0x73,0x63,0x72,0x69,0x70,
0x74,0x69,0x6f,0x6e,0x3e,0x20,0x3c,0x2f,0x72,0x64,
0x66,0x3a,0x52,0x44,0x46,0x3e,0x20,0x3c,0x2f,0x78,
0x3a,0x78,0x6d,0x70,0x6d,0x65,0x74,0x61,0x3e,0x20,
0x3c,0x3f,0x78,0x70,0x61,0x63,0x6b,0x65,0x74,0x20,
0x65,0x6e,0x64,0x3d,0x22,0x72,0x22,0x3f,0x3e,0x4c,
0x94,0x28,0x37,0x00,0x00,0x01,0xc5,0x49,0x44,0x41,
0x54,0x78,0xda,0xe4,0x97,0x4b,0x4e,0xc3,0x30,0x10,
0x86,0x27,0xa8,0x07,0xe8,0xb2,0x12,0x12,0x42,0xe2,
0xd1,0x23,0x80,0x40,0x3c,0x54,0xce,0x80,0xc4,0x19,
0x38,0x01,0xaf,0xd2,0x16,0xf1,0x92,0x10,0x7b,0x4e,
0xc0,0x25,0x2c,0x0e,0xc0,0x12,0x89,0x15,0x17,0x60,
0x41,0x55,0xba,0xa7,0xce,0xd4,0x71,0x03,0x4d,0x9b,
0xb1,0xf3,0xa7,0x90,0x76,0x81,0x2b,0xcb,0x13,0x47,
0xed,0xf7,0x7b,0xc6,0x9e,0xa9,0x03,0x66,0xa6,0x59,
0xb6,0x39,0x9a,0x71,0x2b,0x7d,0x1b,0xf7,0xaf,0xfb,
0xc4,0xa4,0x89,0x38,0x6c,0x98,0xb1,0xc9,0xac,0xcd,
0x73,0xcf,0x3c,0x9b,0x91,0x7b,0xf6,0x9d,0x1d,0xad,
0x1d,0x8d,0x5a,0xb0,0x75,0x6c,0xeb,0xd8,0x0e,0xad,
0x4d,0xd1,0x48,0x91,0xcd,0xe6,0x77,0xb9,0x15,0xcd,
0xdf,0x6d,0x7f,0x8d,0x7a,0x20,0x86,0x73,0x81,0xf0,
0x58,0x80,0x61,0x24,0xc2,0x3e,0x0c,0xc1,0x60,0xe5,
0x54,0x20,0x3c,0x9e,0xb7,0x76,0x23,0x25,0xa0,0xe0,
0x95,0x27,0xe1,0x66,0x9e,0x9a,0x69,0x01,0xd3,0x83,
0x53,0x32,0x04,0xa5,0xa1,0x07,0x30,0xf8,0xc1,0xf2,
0x03,0x55,0xcb,0x7b,0xf6,0x3b,0xf5,0xe7,0xca,0x64,
0x70,0x96,0x8e,0x21,0xb8,0xf2,0xd5,0x72,0x8d,0x38,
0xfe,0x5c,0xac,0xbd,0x4f,0x04,0x17,0x37,0x21,0xea,
0x76,0x32,0x80,0x64,0xbf,0x5c,0xff,0xc8,0x0d,0x67,
0xc9,0x03,0x70,0xcc,0x29,0x4c,0xf5,0xab,0x8d,0x4e,
0x2e,0xb8,0x2c,0x00,0xdd,0x70,0x82,0x00,0x2b,0x62,
0xb3,0x0b,0xc3,0xdd,0x02,0xd0,0xdd,0xee,0xe8,0xd7,
0x5b,0x9f,0x10,0xdc,0x11,0x02,0x14,0x1e,0x7a,0xfb,
0xcd,0x4e,0x37,0x13,0xee,0xf0,0x00,0x78,0xce,0x49,
0x67,0xf6,0xdb,0xdd,0x8e,0x17,0x0e,0x86,0x40,0x4e,
0x32,0x88,0x00,0x2b,0xa2,0xd6,0x76,0xc2,0x93,0x02,
0x4a,0x72,0x08,0x7c,0x19,0x4e,0xc3,0xa5,0xd6,0x09,
0x17,0x05,0x80,0xe9,0x35,0xb4,0xe7,0x3f,0xbb,0x1d,
0xa9,0x8a,0x13,0x2e,0x7b,0x00,0xcd,0xed,0x80,0x07,
0x8e,0xd5,0xbc,0x17,0xce,0x24,0x86,0x00,0x2c,0x2c,
0x19,0x02,0x4e,0xd4,0x42,0x36,0xdc,0x1d,0x02,0xa0,
0xb0,0x78,0x42,0x70,0xaa,0x16,0x21,0x38,0x3b,0x05,
0x20,0x85,0xc5,0xe1,0x81,0x33,0xb5,0x04,0xc3,0x3d,
0x21,0x40,0xaa,0x5a,0xda,0x03,0x75,0xb5,0x92,0x0f,
0x2e,0x7b,0x00,0x2c,0xa9,0x63,0x02,0xce,0x55,0x35,
0x37,0xdc,0x99,0x09,0x91,0x92,0xfa,0xd6,0x7e,0xfa,
0x29,0xc5,0xbf,0x85,0x8f,0x78,0x00,0xad,0xe7,0x8f,
0x2f,0x87,0xfe,0x0c,0x97,0x03,0x3e,0x56,0x8c,0xf2,
0xd5,0xf3,0xbf,0x80,0x8f,0x85,0xc0,0x5e,0x1a,0xa6,
0x05,0x97,0xfe,0x15,0x0f,0x6e,0x2c,0x53,0x5a,0x79,
0x4b,0x10,0x60,0xe1,0x81,0xf5,0x44,0xb1,0x2b,0x0f,
0x92,0x13,0xc1,0xbf,0xbf,0x1d,0xf7,0x05,0x18,0x00,
0x35,0x2b,0x45,0x3b,0x86,0x63,0x1f,0x25,0x00,0x00,
0x00,0x00,0x49,0x45,0x4e,0x44,0xae,0x42,0x60,0x82,

  };

  static FXIcon *VFLL3;
  static unsigned char YQRZ_14 []= {
0x89,0x50,0x4e,0x47,0x0d,0x0a,0x1a,0x0a,0x00,0x00,
0x00,0x0d,0x49,0x48,0x44,0x52,0x00,0x00,0x00,0x20,
0x00,0x00,0x00,0x20,0x08,0x06,0x00,0x00,0x00,0x73,
0x7a,0x7a,0xf4,0x00,0x00,0x00,0x19,0x74,0x45,0x58,
0x74,0x53,0x6f,0x66,0x74,0x77,0x61,0x72,0x65,0x00,
0x41,0x64,0x6f,0x62,0x65,0x20,0x49,0x6d,0x61,0x67,
0x65,0x52,0x65,0x61,0x64,0x79,0x71,0xc9,0x65,0x3c,
0x00,0x00,0x03,0xad,0x69,0x54,0x58,0x74,0x58,0x4d,
0x4c,0x3a,0x63,0x6f,0x6d,0x2e,0x61,0x64,0x6f,0x62,
0x65,0x2e,0x78,0x6d,0x70,0x00,0x00,0x00,0x00,0x00,
0x3c,0x3f,0x78,0x70,0x61,0x63,0x6b,0x65,0x74,0x20,
0x62,0x65,0x67,0x69,0x6e,0x3d,0x22,0xef,0xbb,0xbf,
0x22,0x20,0x69,0x64,0x3d,0x22,0x57,0x35,0x4d,0x30,
0x4d,0x70,0x43,0x65,0x68,0x69,0x48,0x7a,0x72,0x65,
0x53,0x7a,0x4e,0x54,0x63,0x7a,0x6b,0x63,0x39,0x64,
0x22,0x3f,0x3e,0x20,0x3c,0x78,0x3a,0x78,0x6d,0x70,
0x6d,0x65,0x74,0x61,0x20,0x78,0x6d,0x6c,0x6e,0x73,
0x3a,0x78,0x3d,0x22,0x61,0x64,0x6f,0x62,0x65,0x3a,
0x6e,0x73,0x3a,0x6d,0x65,0x74,0x61,0x2f,0x22,0x20,
0x78,0x3a,0x78,0x6d,0x70,0x74,0x6b,0x3d,0x22,0x41,
0x64,0x6f,0x62,0x65,0x20,0x58,0x4d,0x50,0x20,0x43,
0x6f,0x72,0x65,0x20,0x35,0x2e,0x30,0x2d,0x63,0x30,
0x36,0x30,0x20,0x36,0x31,0x2e,0x31,0x33,0x34,0x37,
0x37,0x37,0x2c,0x20,0x32,0x30,0x31,0x30,0x2f,0x30,
0x32,0x2f,0x31,0x32,0x2d,0x31,0x37,0x3a,0x33,0x32,
0x3a,0x30,0x30,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x22,0x3e,0x20,0x3c,0x72,0x64,0x66,0x3a,0x52,
0x44,0x46,0x20,0x78,0x6d,0x6c,0x6e,0x73,0x3a,0x72,
0x64,0x66,0x3d,0x22,0x68,0x74,0x74,0x70,0x3a,0x2f,
0x2f,0x77,0x77,0x77,0x2e,0x77,0x33,0x2e,0x6f,0x72,
0x67,0x2f,0x31,0x39,0x39,0x39,0x2f,0x30,0x32,0x2f,
0x32,0x32,0x2d,0x72,0x64,0x66,0x2d,0x73,0x79,0x6e,
0x74,0x61,0x78,0x2d,0x6e,0x73,0x23,0x22,0x3e,0x20,
0x3c,0x72,0x64,0x66,0x3a,0x44,0x65,0x73,0x63,0x72,
0x69,0x70,0x74,0x69,0x6f,0x6e,0x20,0x72,0x64,0x66,
0x3a,0x61,0x62,0x6f,0x75,0x74,0x3d,0x22,0x22,0x20,
0x78,0x6d,0x6c,0x6e,0x73,0x3a,0x78,0x6d,0x70,0x52,
0x69,0x67,0x68,0x74,0x73,0x3d,0x22,0x68,0x74,0x74,
0x70,0x3a,0x2f,0x2f,0x6e,0x73,0x2e,0x61,0x64,0x6f,
0x62,0x65,0x2e,0x63,0x6f,0x6d,0x2f,0x78,0x61,0x70,
0x2f,0x31,0x2e,0x30,0x2f,0x72,0x69,0x67,0x68,0x74,
0x73,0x2f,0x22,0x20,0x78,0x6d,0x6c,0x6e,0x73,0x3a,
0x78,0x6d,0x70,0x4d,0x4d,0x3d,0x22,0x68,0x74,0x74,
0x70,0x3a,0x2f,0x2f,0x6e,0x73,0x2e,0x61,0x64,0x6f,
0x62,0x65,0x2e,0x63,0x6f,0x6d,0x2f,0x78,0x61,0x70,
0x2f,0x31,0x2e,0x30,0x2f,0x6d,0x6d,0x2f,0x22,0x20,
0x78,0x6d,0x6c,0x6e,0x73,0x3a,0x73,0x74,0x52,0x65,
0x66,0x3d,0x22,0x68,0x74,0x74,0x70,0x3a,0x2f,0x2f,
0x6e,0x73,0x2e,0x61,0x64,0x6f,0x62,0x65,0x2e,0x63,
0x6f,0x6d,0x2f,0x78,0x61,0x70,0x2f,0x31,0x2e,0x30,
0x2f,0x73,0x54,0x79,0x70,0x65,0x2f,0x52,0x65,0x73,
0x6f,0x75,0x72,0x63,0x65,0x52,0x65,0x66,0x23,0x22,
0x20,0x78,0x6d,0x6c,0x6e,0x73,0x3a,0x78,0x6d,0x70,
0x3d,0x22,0x68,0x74,0x74,0x70,0x3a,0x2f,0x2f,0x6e,
0x73,0x2e,0x61,0x64,0x6f,0x62,0x65,0x2e,0x63,0x6f,
0x6d,0x2f,0x78,0x61,0x70,0x2f,0x31,0x2e,0x30,0x2f,
0x22,0x20,0x78,0x6d,0x70,0x52,0x69,0x67,0x68,0x74,
0x73,0x3a,0x4d,0x61,0x72,0x6b,0x65,0x64,0x3d,0x22,
0x46,0x61,0x6c,0x73,0x65,0x22,0x20,0x78,0x6d,0x70,
0x4d,0x4d,0x3a,0x4f,0x72,0x69,0x67,0x69,0x6e,0x61,
0x6c,0x44,0x6f,0x63,0x75,0x6d,0x65,0x6e,0x74,0x49,
0x44,0x3d,0x22,0x75,0x75,0x69,0x64,0x3a,0x41,0x45,
0x33,0x46,0x34,0x36,0x30,0x38,0x38,0x46,0x38,0x39,
0x44,0x45,0x31,0x31,0x41,0x38,0x36,0x30,0x39,0x39,
0x46,0x31,0x32,0x31,0x41,0x32,0x30,0x34,0x32,0x39,
0x22,0x20,0x78,0x6d,0x70,0x4d,0x4d,0x3a,0x44,0x6f,
0x63,0x75,0x6d,0x65,0x6e,0x74,0x49,0x44,0x3d,0x22,
0x78,0x6d,0x70,0x2e,0x64,0x69,0x64,0x3a,0x39,0x31,
0x31,0x46,0x30,0x38,0x42,0x41,0x42,0x35,0x45,0x46,
0x31,0x31,0x45,0x30,0x42,0x41,0x41,0x33,0x42,0x42,
0x45,0x34,0x44,0x37,0x42,0x31,0x33,0x35,0x31,0x44,
0x22,0x20,0x78,0x6d,0x70,0x4d,0x4d,0x3a,0x49,0x6e,
0x73,0x74,0x61,0x6e,0x63,0x65,0x49,0x44,0x3d,0x22,
0x78,0x6d,0x70,0x2e,0x69,0x69,0x64,0x3a,0x39,0x30,
0x46,0x42,0x34,0x35,0x37,0x45,0x42,0x35,0x45,0x46,
0x31,0x31,0x45,0x30,0x42,0x41,0x41,0x33,0x42,0x42,
0x45,0x34,0x44,0x37,0x42,0x31,0x33,0x35,0x31,0x44,
0x22,0x20,0x78,0x6d,0x70,0x3a,0x43,0x72,0x65,0x61,
0x74,0x6f,0x72,0x54,0x6f,0x6f,0x6c,0x3d,0x22,0x41,
0x64,0x6f,0x62,0x65,0x20,0x50,0x68,0x6f,0x74,0x6f,
0x73,0x68,0x6f,0x70,0x20,0x43,0x53,0x35,0x20,0x57,
0x69,0x6e,0x64,0x6f,0x77,0x73,0x22,0x3e,0x20,0x3c,
0x78,0x6d,0x70,0x4d,0x4d,0x3a,0x44,0x65,0x72,0x69,
0x76,0x65,0x64,0x46,0x72,0x6f,0x6d,0x20,0x73,0x74,
0x52,0x65,0x66,0x3a,0x69,0x6e,0x73,0x74,0x61,0x6e,
0x63,0x65,0x49,0x44,0x3d,0x22,0x78,0x6d,0x70,0x2e,
0x69,0x69,0x64,0x3a,0x33,0x32,0x45,0x42,0x39,0x34,
0x34,0x44,0x45,0x46,0x42,0x35,0x45,0x30,0x31,0x31,
0x39,0x41,0x42,0x34,0x41,0x46,0x44,0x39,0x30,0x33,
0x39,0x46,0x33,0x35,0x45,0x42,0x22,0x20,0x73,0x74,
0x52,0x65,0x66,0x3a,0x64,0x6f,0x63,0x75,0x6d,0x65,
0x6e,0x74,0x49,0x44,0x3d,0x22,0x75,0x75,0x69,0x64,
0x3a,0x41,0x45,0x33,0x46,0x34,0x36,0x30,0x38,0x38,
0x46,0x38,0x39,0x44,0x45,0x31,0x31,0x41,0x38,0x36,
0x30,0x39,0x39,0x46,0x31,0x32,0x31,0x41,0x32,0x30,
0x34,0x32,0x39,0x22,0x2f,0x3e,0x20,0x3c,0x2f,0x72,
0x64,0x66,0x3a,0x44,0x65,0x73,0x63,0x72,0x69,0x70,
0x74,0x69,0x6f,0x6e,0x3e,0x20,0x3c,0x2f,0x72,0x64,
0x66,0x3a,0x52,0x44,0x46,0x3e,0x20,0x3c,0x2f,0x78,
0x3a,0x78,0x6d,0x70,0x6d,0x65,0x74,0x61,0x3e,0x20,
0x3c,0x3f,0x78,0x70,0x61,0x63,0x6b,0x65,0x74,0x20,
0x65,0x6e,0x64,0x3d,0x22,0x72,0x22,0x3f,0x3e,0xb0,
0xd6,0x56,0x84,0x00,0x00,0x01,0xfd,0x49,0x44,0x41,
0x54,0x78,0xda,0xe4,0x57,0xbd,0x4e,0xc3,0x30,0x10,
0xbe,0x54,0xdd,0x58,0x5a,0x5e,0x01,0xd1,0xf6,0x29,
0x10,0x88,0x01,0xc4,0x84,0x84,0x10,0x13,0xe2,0x39,
0x98,0xda,0xce,0x88,0x85,0x99,0x81,0x05,0xde,0x80,
0x8d,0xb2,0xf1,0x06,0x88,0xd2,0xbe,0x01,0x08,0x10,
0x3b,0x4d,0x0e,0xc7,0x75,0x5a,0xff,0xdc,0x9d,0xc3,
0x90,0x74,0xc0,0x8b,0x2f,0xb9,0xa6,0xdf,0xdd,0xf7,
0xf9,0xce,0x76,0x82,0x88,0xb0,0xca,0xd1,0x80,0x15,
0x8f,0x66,0x61,0x5c,0x3e,0x1f,0x03,0x42,0x0a,0x80,
0x59,0x5f,0xcd,0x03,0xc4,0x54,0x3d,0xcf,0xd4,0xb3,
0x9a,0x71,0xa6,0x7d,0x7a,0xd6,0x76,0x3e,0xa7,0x84,
0x9d,0x1a,0x3b,0x35,0x76,0xa6,0x6d,0xc8,0x67,0xc8,
0x6d,0x54,0xff,0x8b,0xc3,0xfc,0xfd,0xc5,0xd6,0x8f,
0xcb,0x80,0x01,0xc7,0x0a,0xc1,0x4d,0x00,0x0a,0xc3,
0x92,0x7d,0x29,0xc1,0x3c,0x73,0xa8,0x10,0xdc,0xbc,
0xd7,0x76,0x3f,0x08,0xa0,0xe2,0xcc,0x6d,0x70,0xf5,
0x1e,0x06,0x61,0x00,0xf5,0x81,0x83,0x2d,0x41,0x73,
0xc9,0x80,0x0c,0x7e,0xb2,0x71,0x05,0x9d,0xd6,0x8e,
0xfe,0xed,0xe4,0xfb,0x01,0x6e,0xa7,0x67,0x0e,0xf8,
0x69,0xf7,0x0e,0x7a,0xed,0x7d,0xed,0x1f,0x7f,0xdd,
0xc3,0xcd,0xcb,0x11,0x0f,0x8e,0x44,0x15,0xc4,0x32,
0xdf,0x6c,0x6d,0x03,0x9a,0x2f,0x3b,0xad,0xdd,0x20,
0xf3,0x6e,0x7b,0x6f,0xe1,0xef,0xad,0x1f,0x88,0xe0,
0x34,0x03,0x11,0xda,0x41,0x51,0x6a,0x8f,0x80,0xf6,
0xc0,0xcf,0x83,0x23,0xc5,0x40,0x54,0x73,0x1f,0x80,
0xd0,0xdc,0x0d,0x80,0x07,0xa7,0x03,0x88,0x2e,0x38,
0x2a,0x43,0x6b,0xc1,0x51,0x01,0x30,0xe0,0x7c,0x00,
0xd2,0x6a,0xcf,0x1b,0x95,0x27,0x81,0xb3,0xda,0x7d,
0xbf,0x00,0xce,0x48,0x10,0x29,0x35,0x2f,0xc3,0xa0,
0xd4,0x08,0x06,0x38,0x70,0x86,0x81,0x78,0x9d,0xbb,
0x19,0x7a,0x75,0xee,0xf9,0x25,0xf0,0x92,0x12,0x78,
0x4d,0x26,0x90,0xc0,0x6b,0x32,0xa4,0x1f,0xfe,0x10,
0x40,0xac,0xc3,0x91,0x00,0xb6,0x04,0xbe,0x5f,0x00,
0x97,0x25,0xa0,0xdb,0x6b,0x16,0xa9,0xf3,0xd0,0xcf,
0x83,0x0b,0x12,0x48,0xbd,0x3d,0x15,0xeb,0xbc,0x0c,
0x03,0x05,0x38,0x02,0xdb,0x88,0x84,0x8d,0x85,0x0a,
0xc0,0x5e,0xed,0x91,0x00,0x1c,0x70,0x5e,0x02,0x61,
0x57,0x0b,0x3a,0xa1,0x57,0x6a,0x82,0x04,0x3e,0x38,
0x5f,0x86,0xd2,0x96,0x4a,0x30,0xe0,0x94,0x1a,0xc3,
0x00,0x05,0x2e,0x48,0x20,0xec,0xe7,0x41,0x23,0xf2,
0x4a,0x8d,0x60,0x80,0x05,0xa7,0x19,0x88,0x1c,0x26,
0xc8,0xcd,0x06,0xd8,0x00,0x24,0x70,0x5b,0x82,0x06,
0xdb,0xdb,0xbd,0x3a,0x9f,0x7e,0x3e,0x9a,0x2d,0x37,
0x83,0xc9,0xc7,0x28,0x58,0x64,0xaf,0xef,0xa3,0x85,
0x7f,0xfc,0x36,0x2a,0x05,0x9e,0x8f,0xa4,0x38,0x1c,
0x9c,0x3f,0xad,0x61,0xf4,0x18,0x85,0x25,0x3a,0x5c,
0xc9,0xcc,0xaf,0x0f,0x31,0xf1,0x0e,0xa5,0xf5,0x81,
0x93,0x37,0x23,0x73,0x69,0xa8,0x0b,0x9c,0x3a,0x15,
0xcf,0x6f,0x2c,0x35,0x65,0x3e,0x24,0x02,0xd0,0xe0,
0x89,0x66,0xa2,0xda,0xcc,0x13,0x72,0x11,0xfe,0xdb,
0xdb,0xf1,0xaf,0x00,0x03,0x00,0xb5,0xe2,0xf9,0xc8,
0x45,0x99,0x4f,0x8d,0x00,0x00,0x00,0x00,0x49,0x45,
0x4e,0x44,0xae,0x42,0x60,0x82,

  };

  static FXIcon *VFLL4;
  static unsigned char YQRZ_20 []= {
0x89,0x50,0x4e,0x47,0x0d,0x0a,0x1a,0x0a,0x00,0x00,
0x00,0x0d,0x49,0x48,0x44,0x52,0x00,0x00,0x00,0x20,
0x00,0x00,0x00,0x20,0x08,0x06,0x00,0x00,0x00,0x73,
0x7a,0x7a,0xf4,0x00,0x00,0x00,0x19,0x74,0x45,0x58,
0x74,0x53,0x6f,0x66,0x74,0x77,0x61,0x72,0x65,0x00,
0x41,0x64,0x6f,0x62,0x65,0x20,0x49,0x6d,0x61,0x67,
0x65,0x52,0x65,0x61,0x64,0x79,0x71,0xc9,0x65,0x3c,
0x00,0x00,0x03,0xad,0x69,0x54,0x58,0x74,0x58,0x4d,
0x4c,0x3a,0x63,0x6f,0x6d,0x2e,0x61,0x64,0x6f,0x62,
0x65,0x2e,0x78,0x6d,0x70,0x00,0x00,0x00,0x00,0x00,
0x3c,0x3f,0x78,0x70,0x61,0x63,0x6b,0x65,0x74,0x20,
0x62,0x65,0x67,0x69,0x6e,0x3d,0x22,0xef,0xbb,0xbf,
0x22,0x20,0x69,0x64,0x3d,0x22,0x57,0x35,0x4d,0x30,
0x4d,0x70,0x43,0x65,0x68,0x69,0x48,0x7a,0x72,0x65,
0x53,0x7a,0x4e,0x54,0x63,0x7a,0x6b,0x63,0x39,0x64,
0x22,0x3f,0x3e,0x20,0x3c,0x78,0x3a,0x78,0x6d,0x70,
0x6d,0x65,0x74,0x61,0x20,0x78,0x6d,0x6c,0x6e,0x73,
0x3a,0x78,0x3d,0x22,0x61,0x64,0x6f,0x62,0x65,0x3a,
0x6e,0x73,0x3a,0x6d,0x65,0x74,0x61,0x2f,0x22,0x20,
0x78,0x3a,0x78,0x6d,0x70,0x74,0x6b,0x3d,0x22,0x41,
0x64,0x6f,0x62,0x65,0x20,0x58,0x4d,0x50,0x20,0x43,
0x6f,0x72,0x65,0x20,0x35,0x2e,0x30,0x2d,0x63,0x30,
0x36,0x30,0x20,0x36,0x31,0x2e,0x31,0x33,0x34,0x37,
0x37,0x37,0x2c,0x20,0x32,0x30,0x31,0x30,0x2f,0x30,
0x32,0x2f,0x31,0x32,0x2d,0x31,0x37,0x3a,0x33,0x32,
0x3a,0x30,0x30,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x22,0x3e,0x20,0x3c,0x72,0x64,0x66,0x3a,0x52,
0x44,0x46,0x20,0x78,0x6d,0x6c,0x6e,0x73,0x3a,0x72,
0x64,0x66,0x3d,0x22,0x68,0x74,0x74,0x70,0x3a,0x2f,
0x2f,0x77,0x77,0x77,0x2e,0x77,0x33,0x2e,0x6f,0x72,
0x67,0x2f,0x31,0x39,0x39,0x39,0x2f,0x30,0x32,0x2f,
0x32,0x32,0x2d,0x72,0x64,0x66,0x2d,0x73,0x79,0x6e,
0x74,0x61,0x78,0x2d,0x6e,0x73,0x23,0x22,0x3e,0x20,
0x3c,0x72,0x64,0x66,0x3a,0x44,0x65,0x73,0x63,0x72,
0x69,0x70,0x74,0x69,0x6f,0x6e,0x20,0x72,0x64,0x66,
0x3a,0x61,0x62,0x6f,0x75,0x74,0x3d,0x22,0x22,0x20,
0x78,0x6d,0x6c,0x6e,0x73,0x3a,0x78,0x6d,0x70,0x52,
0x69,0x67,0x68,0x74,0x73,0x3d,0x22,0x68,0x74,0x74,
0x70,0x3a,0x2f,0x2f,0x6e,0x73,0x2e,0x61,0x64,0x6f,
0x62,0x65,0x2e,0x63,0x6f,0x6d,0x2f,0x78,0x61,0x70,
0x2f,0x31,0x2e,0x30,0x2f,0x72,0x69,0x67,0x68,0x74,
0x73,0x2f,0x22,0x20,0x78,0x6d,0x6c,0x6e,0x73,0x3a,
0x78,0x6d,0x70,0x4d,0x4d,0x3d,0x22,0x68,0x74,0x74,
0x70,0x3a,0x2f,0x2f,0x6e,0x73,0x2e,0x61,0x64,0x6f,
0x62,0x65,0x2e,0x63,0x6f,0x6d,0x2f,0x78,0x61,0x70,
0x2f,0x31,0x2e,0x30,0x2f,0x6d,0x6d,0x2f,0x22,0x20,
0x78,0x6d,0x6c,0x6e,0x73,0x3a,0x73,0x74,0x52,0x65,
0x66,0x3d,0x22,0x68,0x74,0x74,0x70,0x3a,0x2f,0x2f,
0x6e,0x73,0x2e,0x61,0x64,0x6f,0x62,0x65,0x2e,0x63,
0x6f,0x6d,0x2f,0x78,0x61,0x70,0x2f,0x31,0x2e,0x30,
0x2f,0x73,0x54,0x79,0x70,0x65,0x2f,0x52,0x65,0x73,
0x6f,0x75,0x72,0x63,0x65,0x52,0x65,0x66,0x23,0x22,
0x20,0x78,0x6d,0x6c,0x6e,0x73,0x3a,0x78,0x6d,0x70,
0x3d,0x22,0x68,0x74,0x74,0x70,0x3a,0x2f,0x2f,0x6e,
0x73,0x2e,0x61,0x64,0x6f,0x62,0x65,0x2e,0x63,0x6f,
0x6d,0x2f,0x78,0x61,0x70,0x2f,0x31,0x2e,0x30,0x2f,
0x22,0x20,0x78,0x6d,0x70,0x52,0x69,0x67,0x68,0x74,
0x73,0x3a,0x4d,0x61,0x72,0x6b,0x65,0x64,0x3d,0x22,
0x46,0x61,0x6c,0x73,0x65,0x22,0x20,0x78,0x6d,0x70,
0x4d,0x4d,0x3a,0x4f,0x72,0x69,0x67,0x69,0x6e,0x61,
0x6c,0x44,0x6f,0x63,0x75,0x6d,0x65,0x6e,0x74,0x49,
0x44,0x3d,0x22,0x75,0x75,0x69,0x64,0x3a,0x41,0x45,
0x33,0x46,0x34,0x36,0x30,0x38,0x38,0x46,0x38,0x39,
0x44,0x45,0x31,0x31,0x41,0x38,0x36,0x30,0x39,0x39,
0x46,0x31,0x32,0x31,0x41,0x32,0x30,0x34,0x32,0x39,
0x22,0x20,0x78,0x6d,0x70,0x4d,0x4d,0x3a,0x44,0x6f,
0x63,0x75,0x6d,0x65,0x6e,0x74,0x49,0x44,0x3d,0x22,
0x78,0x6d,0x70,0x2e,0x64,0x69,0x64,0x3a,0x39,0x31,
0x35,0x33,0x37,0x43,0x37,0x42,0x42,0x35,0x45,0x46,
0x31,0x31,0x45,0x30,0x42,0x41,0x41,0x33,0x42,0x42,
0x45,0x34,0x44,0x37,0x42,0x31,0x33,0x35,0x31,0x44,
0x22,0x20,0x78,0x6d,0x70,0x4d,0x4d,0x3a,0x49,0x6e,
0x73,0x74,0x61,0x6e,0x63,0x65,0x49,0x44,0x3d,0x22,
0x78,0x6d,0x70,0x2e,0x69,0x69,0x64,0x3a,0x39,0x31,
0x35,0x33,0x37,0x43,0x37,0x41,0x42,0x35,0x45,0x46,
0x31,0x31,0x45,0x30,0x42,0x41,0x41,0x33,0x42,0x42,
0x45,0x34,0x44,0x37,0x42,0x31,0x33,0x35,0x31,0x44,
0x22,0x20,0x78,0x6d,0x70,0x3a,0x43,0x72,0x65,0x61,
0x74,0x6f,0x72,0x54,0x6f,0x6f,0x6c,0x3d,0x22,0x41,
0x64,0x6f,0x62,0x65,0x20,0x50,0x68,0x6f,0x74,0x6f,
0x73,0x68,0x6f,0x70,0x20,0x43,0x53,0x35,0x20,0x57,
0x69,0x6e,0x64,0x6f,0x77,0x73,0x22,0x3e,0x20,0x3c,
0x78,0x6d,0x70,0x4d,0x4d,0x3a,0x44,0x65,0x72,0x69,
0x76,0x65,0x64,0x46,0x72,0x6f,0x6d,0x20,0x73,0x74,
0x52,0x65,0x66,0x3a,0x69,0x6e,0x73,0x74,0x61,0x6e,
0x63,0x65,0x49,0x44,0x3d,0x22,0x78,0x6d,0x70,0x2e,
0x69,0x69,0x64,0x3a,0x33,0x32,0x45,0x42,0x39,0x34,
0x34,0x44,0x45,0x46,0x42,0x35,0x45,0x30,0x31,0x31,
0x39,0x41,0x42,0x34,0x41,0x46,0x44,0x39,0x30,0x33,
0x39,0x46,0x33,0x35,0x45,0x42,0x22,0x20,0x73,0x74,
0x52,0x65,0x66,0x3a,0x64,0x6f,0x63,0x75,0x6d,0x65,
0x6e,0x74,0x49,0x44,0x3d,0x22,0x75,0x75,0x69,0x64,
0x3a,0x41,0x45,0x33,0x46,0x34,0x36,0x30,0x38,0x38,
0x46,0x38,0x39,0x44,0x45,0x31,0x31,0x41,0x38,0x36,
0x30,0x39,0x39,0x46,0x31,0x32,0x31,0x41,0x32,0x30,
0x34,0x32,0x39,0x22,0x2f,0x3e,0x20,0x3c,0x2f,0x72,
0x64,0x66,0x3a,0x44,0x65,0x73,0x63,0x72,0x69,0x70,
0x74,0x69,0x6f,0x6e,0x3e,0x20,0x3c,0x2f,0x72,0x64,
0x66,0x3a,0x52,0x44,0x46,0x3e,0x20,0x3c,0x2f,0x78,
0x3a,0x78,0x6d,0x70,0x6d,0x65,0x74,0x61,0x3e,0x20,
0x3c,0x3f,0x78,0x70,0x61,0x63,0x6b,0x65,0x74,0x20,
0x65,0x6e,0x64,0x3d,0x22,0x72,0x22,0x3f,0x3e,0x82,
0xc1,0xaf,0xf6,0x00,0x00,0x02,0x59,0x49,0x44,0x41,
0x54,0x78,0xda,0xe4,0x97,0xbf,0x6b,0xdb,0x40,0x14,
0xc7,0x9f,0x8a,0xb7,0x6e,0x21,0x93,0xa9,0x69,0x09,
0x94,0xc4,0x90,0xa1,0x85,0xb6,0xb4,0xa8,0x78,0x4f,
0xa7,0xb4,0xd0,0x35,0x7f,0x46,0xc0,0x89,0x6d,0xfd,
0x70,0xb7,0xd2,0xff,0xa0,0x90,0x39,0xdd,0x3b,0x1a,
0xb4,0x84,0x96,0x36,0x90,0xa1,0xa9,0x5d,0x12,0x02,
0x1e,0x42,0x82,0x33,0xe4,0x1f,0x88,0xe5,0xd7,0xd3,
0x59,0xb6,0xa5,0xbb,0xef,0x49,0xca,0x60,0x67,0xc8,
0x19,0xf9,0x4e,0x3a,0x9d,0x3e,0xef,0xbd,0xef,0xbb,
0x93,0xce,0x62,0x66,0xba,0xcb,0xf2,0x80,0xee,0xb8,
0x94,0x26,0x8d,0x2f,0xc7,0x1f,0x89,0x29,0x24,0xe2,
0x91,0x23,0x6a,0x97,0x39,0x14,0xe7,0x43,0x71,0x2e,
0x6a,0x1e,0xca,0x3e,0x59,0xcb,0x76,0x54,0x87,0xa0,
0x1d,0xc6,0xed,0x30,0x6e,0x8f,0x64,0x9b,0xa2,0x9a,
0xa2,0x36,0x8b,0xe7,0xb2,0x17,0x5d,0xff,0x5c,0xbb,
0x49,0x47,0x20,0x86,0xf3,0x1c,0xe1,0xb1,0x01,0x82,
0x91,0x90,0x7d,0x26,0xc1,0xd8,0x73,0x9a,0x23,0x3c,
0xbe,0x2e,0xdb,0x8e,0x66,0xc0,0x9c,0x3d,0x4f,0xc2,
0xc5,0x75,0x72,0x75,0x03,0x16,0x07,0xa7,0xa4,0x04,
0xa5,0x59,0x04,0x74,0xf8,0x46,0x65,0x9b,0x9e,0x2f,
0x7f,0x90,0xfd,0xfe,0x61,0xb5,0x10,0x7c,0x6b,0x6d,
0x9f,0xaa,0x4b,0xef,0xe4,0x98,0xee,0xf5,0x77,0xda,
0x3b,0xde,0xd4,0xe1,0x8c,0xa6,0x21,0xf0,0xfc,0xd9,
0xf2,0x7b,0xe2,0xf8,0xd7,0x7c,0xd1,0x2d,0xe4,0xf9,
0xda,0xd2,0xc6,0x74,0x4c,0x64,0x08,0x82,0xc3,0x24,
0x44,0x61,0x27,0x11,0xc6,0xe4,0xe1,0xbc,0x3c,0xcd,
0x0d,0xbb,0x3a,0x06,0xc1,0x19,0x45,0x00,0x69,0xde,
0x3e,0x5c,0x1f,0x6b,0x99,0x38,0xdc,0x57,0xfd,0x6c,
0xcd,0x95,0xfb,0x11,0x1c,0x1b,0x60,0x48,0x38,0xff,
0xf7,0xaa,0xf6,0x50,0xef,0xf5,0xb9,0x39,0xe1,0x54,
0x03,0x00,0xdc,0x6c,0x80,0x21,0xdb,0xbd,0x5f,0x2b,
0x63,0x8f,0x13,0x87,0xff,0xe6,0x02,0x67,0xbb,0x72,
0x1f,0x82,0x1b,0x24,0xc8,0x9e,0x6a,0xee,0xcf,0xc7,
0x9a,0x77,0xbe,0x3d,0xd0,0xa7,0x1a,0x88,0x80,0x0a,
0x37,0x44,0x20,0x7f,0x9e,0x3b,0x3f,0x1e,0x69,0x1e,
0xb6,0xdf,0x0e,0x94,0x79,0x9e,0xee,0x47,0xf0,0x82,
0x12,0xe0,0x6c,0x77,0x0e,0xca,0x1a,0xe4,0x53,0x6d,
0x30,0xcb,0xf6,0x5b,0x1a,0x50,0xc2,0x12,0xe4,0xac,
0x70,0x91,0xb6,0x4a,0x99,0x64,0xbb,0xda,0x07,0xe1,
0xd0,0x80,0x82,0xf0,0x96,0xdd,0x17,0xff,0xa3,0x14,
0xa4,0x11,0x94,0xa7,0xd9,0xae,0xf6,0x21,0x38,0x8e,
0x40,0x01,0x78,0xd3,0x3e,0xd3,0x3c,0x6c,0x06,0x95,
0xf4,0x54,0xcb,0x88,0xc0,0x04,0xce,0x04,0x25,0xc8,
0x86,0x37,0xec,0x13,0xed,0xe1,0xad,0xe0,0x89,0x3e,
0xcf,0x0d,0x06,0xa4,0xe0,0x66,0x09,0x30,0x7c,0xd7,
0xfe,0x17,0x2f,0xb3,0xb3,0xe2,0x04,0x2b,0x78,0x91,
0x01,0x12,0xa8,0x70,0x36,0x1a,0x00,0xe0,0x3b,0xf6,
0x5f,0xcd,0x2b,0x37,0x78,0x6a,0x5c,0xe1,0xb4,0x04,
0x05,0xf0,0x0c,0x09,0xd2,0xf0,0xba,0xfd,0x47,0xf3,
0xc8,0x0b,0x56,0xcd,0x70,0x43,0x04,0x98,0x0a,0x49,
0x80,0x5f,0x2c,0xc9,0xe2,0x07,0xd5,0x6c,0x38,0x32,
0x80,0xb2,0x25,0x48,0xad,0x84,0xea,0xda,0x7e,0x74,
0xf9,0x6d,0xfa,0x5a,0x2d,0x04,0x17,0x47,0xef,0xaa,
0x33,0x1d,0xd3,0x1b,0x74,0x32,0xe1,0x51,0xb1,0x26,
0x1f,0x07,0xf5,0x83,0x87,0x9c,0xfb,0x19,0x95,0x03,
0x37,0x25,0x1c,0x82,0x7f,0xdd,0x64,0x4b,0x79,0x19,
0x2d,0x0e,0x0e,0x77,0x46,0xf1,0xa6,0x61,0x51,0x70,
0xf4,0x55,0x3c,0xde,0xb1,0x2c,0xc8,0x73,0x0f,0x18,
0x20,0xe1,0x96,0x8c,0xc4,0x7c,0x3d,0xb7,0x60,0x12,
0xde,0xdb,0xdd,0xf1,0x7f,0x01,0x06,0x00,0x2c,0x50,
0x2a,0x65,0xd5,0x18,0xfc,0xe7,0x00,0x00,0x00,0x00,
0x49,0x45,0x4e,0x44,0xae,0x42,0x60,0x82,

  };

  static FXIcon *VFLL5;
  static unsigned char YQRZ_26 []= {
0x89,0x50,0x4e,0x47,0x0d,0x0a,0x1a,0x0a,0x00,0x00,
0x00,0x0d,0x49,0x48,0x44,0x52,0x00,0x00,0x00,0x20,
0x00,0x00,0x00,0x20,0x08,0x06,0x00,0x00,0x00,0x73,
0x7a,0x7a,0xf4,0x00,0x00,0x00,0x19,0x74,0x45,0x58,
0x74,0x53,0x6f,0x66,0x74,0x77,0x61,0x72,0x65,0x00,
0x41,0x64,0x6f,0x62,0x65,0x20,0x49,0x6d,0x61,0x67,
0x65,0x52,0x65,0x61,0x64,0x79,0x71,0xc9,0x65,0x3c,
0x00,0x00,0x03,0xad,0x69,0x54,0x58,0x74,0x58,0x4d,
0x4c,0x3a,0x63,0x6f,0x6d,0x2e,0x61,0x64,0x6f,0x62,
0x65,0x2e,0x78,0x6d,0x70,0x00,0x00,0x00,0x00,0x00,
0x3c,0x3f,0x78,0x70,0x61,0x63,0x6b,0x65,0x74,0x20,
0x62,0x65,0x67,0x69,0x6e,0x3d,0x22,0xef,0xbb,0xbf,
0x22,0x20,0x69,0x64,0x3d,0x22,0x57,0x35,0x4d,0x30,
0x4d,0x70,0x43,0x65,0x68,0x69,0x48,0x7a,0x72,0x65,
0x53,0x7a,0x4e,0x54,0x63,0x7a,0x6b,0x63,0x39,0x64,
0x22,0x3f,0x3e,0x20,0x3c,0x78,0x3a,0x78,0x6d,0x70,
0x6d,0x65,0x74,0x61,0x20,0x78,0x6d,0x6c,0x6e,0x73,
0x3a,0x78,0x3d,0x22,0x61,0x64,0x6f,0x62,0x65,0x3a,
0x6e,0x73,0x3a,0x6d,0x65,0x74,0x61,0x2f,0x22,0x20,
0x78,0x3a,0x78,0x6d,0x70,0x74,0x6b,0x3d,0x22,0x41,
0x64,0x6f,0x62,0x65,0x20,0x58,0x4d,0x50,0x20,0x43,
0x6f,0x72,0x65,0x20,0x35,0x2e,0x30,0x2d,0x63,0x30,
0x36,0x30,0x20,0x36,0x31,0x2e,0x31,0x33,0x34,0x37,
0x37,0x37,0x2c,0x20,0x32,0x30,0x31,0x30,0x2f,0x30,
0x32,0x2f,0x31,0x32,0x2d,0x31,0x37,0x3a,0x33,0x32,
0x3a,0x30,0x30,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x22,0x3e,0x20,0x3c,0x72,0x64,0x66,0x3a,0x52,
0x44,0x46,0x20,0x78,0x6d,0x6c,0x6e,0x73,0x3a,0x72,
0x64,0x66,0x3d,0x22,0x68,0x74,0x74,0x70,0x3a,0x2f,
0x2f,0x77,0x77,0x77,0x2e,0x77,0x33,0x2e,0x6f,0x72,
0x67,0x2f,0x31,0x39,0x39,0x39,0x2f,0x30,0x32,0x2f,
0x32,0x32,0x2d,0x72,0x64,0x66,0x2d,0x73,0x79,0x6e,
0x74,0x61,0x78,0x2d,0x6e,0x73,0x23,0x22,0x3e,0x20,
0x3c,0x72,0x64,0x66,0x3a,0x44,0x65,0x73,0x63,0x72,
0x69,0x70,0x74,0x69,0x6f,0x6e,0x20,0x72,0x64,0x66,
0x3a,0x61,0x62,0x6f,0x75,0x74,0x3d,0x22,0x22,0x20,
0x78,0x6d,0x6c,0x6e,0x73,0x3a,0x78,0x6d,0x70,0x52,
0x69,0x67,0x68,0x74,0x73,0x3d,0x22,0x68,0x74,0x74,
0x70,0x3a,0x2f,0x2f,0x6e,0x73,0x2e,0x61,0x64,0x6f,
0x62,0x65,0x2e,0x63,0x6f,0x6d,0x2f,0x78,0x61,0x70,
0x2f,0x31,0x2e,0x30,0x2f,0x72,0x69,0x67,0x68,0x74,
0x73,0x2f,0x22,0x20,0x78,0x6d,0x6c,0x6e,0x73,0x3a,
0x78,0x6d,0x70,0x4d,0x4d,0x3d,0x22,0x68,0x74,0x74,
0x70,0x3a,0x2f,0x2f,0x6e,0x73,0x2e,0x61,0x64,0x6f,
0x62,0x65,0x2e,0x63,0x6f,0x6d,0x2f,0x78,0x61,0x70,
0x2f,0x31,0x2e,0x30,0x2f,0x6d,0x6d,0x2f,0x22,0x20,
0x78,0x6d,0x6c,0x6e,0x73,0x3a,0x73,0x74,0x52,0x65,
0x66,0x3d,0x22,0x68,0x74,0x74,0x70,0x3a,0x2f,0x2f,
0x6e,0x73,0x2e,0x61,0x64,0x6f,0x62,0x65,0x2e,0x63,
0x6f,0x6d,0x2f,0x78,0x61,0x70,0x2f,0x31,0x2e,0x30,
0x2f,0x73,0x54,0x79,0x70,0x65,0x2f,0x52,0x65,0x73,
0x6f,0x75,0x72,0x63,0x65,0x52,0x65,0x66,0x23,0x22,
0x20,0x78,0x6d,0x6c,0x6e,0x73,0x3a,0x78,0x6d,0x70,
0x3d,0x22,0x68,0x74,0x74,0x70,0x3a,0x2f,0x2f,0x6e,
0x73,0x2e,0x61,0x64,0x6f,0x62,0x65,0x2e,0x63,0x6f,
0x6d,0x2f,0x78,0x61,0x70,0x2f,0x31,0x2e,0x30,0x2f,
0x22,0x20,0x78,0x6d,0x70,0x52,0x69,0x67,0x68,0x74,
0x73,0x3a,0x4d,0x61,0x72,0x6b,0x65,0x64,0x3d,0x22,
0x46,0x61,0x6c,0x73,0x65,0x22,0x20,0x78,0x6d,0x70,
0x4d,0x4d,0x3a,0x4f,0x72,0x69,0x67,0x69,0x6e,0x61,
0x6c,0x44,0x6f,0x63,0x75,0x6d,0x65,0x6e,0x74,0x49,
0x44,0x3d,0x22,0x75,0x75,0x69,0x64,0x3a,0x41,0x45,
0x33,0x46,0x34,0x36,0x30,0x38,0x38,0x46,0x38,0x39,
0x44,0x45,0x31,0x31,0x41,0x38,0x36,0x30,0x39,0x39,
0x46,0x31,0x32,0x31,0x41,0x32,0x30,0x34,0x32,0x39,
0x22,0x20,0x78,0x6d,0x70,0x4d,0x4d,0x3a,0x44,0x6f,
0x63,0x75,0x6d,0x65,0x6e,0x74,0x49,0x44,0x3d,0x22,
0x78,0x6d,0x70,0x2e,0x64,0x69,0x64,0x3a,0x39,0x35,
0x30,0x31,0x33,0x44,0x35,0x45,0x42,0x35,0x45,0x46,
0x31,0x31,0x45,0x30,0x42,0x41,0x41,0x33,0x42,0x42,
0x45,0x34,0x44,0x37,0x42,0x31,0x33,0x35,0x31,0x44,
0x22,0x20,0x78,0x6d,0x70,0x4d,0x4d,0x3a,0x49,0x6e,
0x73,0x74,0x61,0x6e,0x63,0x65,0x49,0x44,0x3d,0x22,
0x78,0x6d,0x70,0x2e,0x69,0x69,0x64,0x3a,0x39,0x35,
0x30,0x31,0x33,0x44,0x35,0x44,0x42,0x35,0x45,0x46,
0x31,0x31,0x45,0x30,0x42,0x41,0x41,0x33,0x42,0x42,
0x45,0x34,0x44,0x37,0x42,0x31,0x33,0x35,0x31,0x44,
0x22,0x20,0x78,0x6d,0x70,0x3a,0x43,0x72,0x65,0x61,
0x74,0x6f,0x72,0x54,0x6f,0x6f,0x6c,0x3d,0x22,0x41,
0x64,0x6f,0x62,0x65,0x20,0x50,0x68,0x6f,0x74,0x6f,
0x73,0x68,0x6f,0x70,0x20,0x43,0x53,0x35,0x20,0x57,
0x69,0x6e,0x64,0x6f,0x77,0x73,0x22,0x3e,0x20,0x3c,
0x78,0x6d,0x70,0x4d,0x4d,0x3a,0x44,0x65,0x72,0x69,
0x76,0x65,0x64,0x46,0x72,0x6f,0x6d,0x20,0x73,0x74,
0x52,0x65,0x66,0x3a,0x69,0x6e,0x73,0x74,0x61,0x6e,
0x63,0x65,0x49,0x44,0x3d,0x22,0x78,0x6d,0x70,0x2e,
0x69,0x69,0x64,0x3a,0x33,0x32,0x45,0x42,0x39,0x34,
0x34,0x44,0x45,0x46,0x42,0x35,0x45,0x30,0x31,0x31,
0x39,0x41,0x42,0x34,0x41,0x46,0x44,0x39,0x30,0x33,
0x39,0x46,0x33,0x35,0x45,0x42,0x22,0x20,0x73,0x74,
0x52,0x65,0x66,0x3a,0x64,0x6f,0x63,0x75,0x6d,0x65,
0x6e,0x74,0x49,0x44,0x3d,0x22,0x75,0x75,0x69,0x64,
0x3a,0x41,0x45,0x33,0x46,0x34,0x36,0x30,0x38,0x38,
0x46,0x38,0x39,0x44,0x45,0x31,0x31,0x41,0x38,0x36,
0x30,0x39,0x39,0x46,0x31,0x32,0x31,0x41,0x32,0x30,
0x34,0x32,0x39,0x22,0x2f,0x3e,0x20,0x3c,0x2f,0x72,
0x64,0x66,0x3a,0x44,0x65,0x73,0x63,0x72,0x69,0x70,
0x74,0x69,0x6f,0x6e,0x3e,0x20,0x3c,0x2f,0x72,0x64,
0x66,0x3a,0x52,0x44,0x46,0x3e,0x20,0x3c,0x2f,0x78,
0x3a,0x78,0x6d,0x70,0x6d,0x65,0x74,0x61,0x3e,0x20,
0x3c,0x3f,0x78,0x70,0x61,0x63,0x6b,0x65,0x74,0x20,
0x65,0x6e,0x64,0x3d,0x22,0x72,0x22,0x3f,0x3e,0xf6,
0x2d,0xdd,0xd0,0x00,0x00,0x05,0xea,0x49,0x44,0x41,
0x54,0x78,0xda,0xc4,0x57,0x7d,0x4c,0x53,0x57,0x14,
0x3f,0xfd,0xa0,0x08,0x88,0x54,0x04,0xac,0xb0,0x21,
0x02,0x82,0x8a,0x6e,0x2a,0x7e,0x32,0xb7,0x4e,0x1c,
0x59,0xf6,0x91,0x65,0x43,0x8d,0x46,0xe3,0x5c,0xf6,
0x19,0xa3,0x59,0xb2,0xfd,0xb1,0xec,0x0f,0x23,0xb0,
0xb9,0xcc,0x6d,0x99,0x9b,0x4b,0x96,0x65,0x59,0x0c,
0xcb,0x66,0x8c,0x46,0x61,0x8b,0xcc,0x05,0x25,0x6e,
0xe8,0x36,0x17,0x85,0x12,0x84,0x4d,0x21,0x32,0xf9,
0x10,0xd0,0x02,0x16,0x81,0x42,0xa5,0xed,0x6b,0xf7,
0x3b,0x97,0xf7,0xea,0x6b,0x69,0x0b,0xfa,0x8f,0x27,
0xf9,0xf5,0xbe,0xfb,0xee,0xbd,0xe7,0x9c,0x7b,0xce,
0xef,0x9e,0xfb,0xaa,0xf1,0x7a,0xbd,0xf4,0x20,0x45,
0xcf,0x3f,0xc5,0x65,0x96,0xfb,0x5d,0x1f,0x0f,0xd8,
0xee,0x77,0x71,0xf1,0xfa,0xdc,0x31,0x07,0x58,0x8a,
0xd0,0x09,0x27,0x25,0x65,0x16,0x13,0x9a,0x17,0x64,
0x2c,0x02,0x52,0x55,0xc3,0x12,0x50,0x0f,0x9c,0x05,
0x2a,0xa1,0xab,0x6a,0x22,0xe3,0x25,0xf2,0xa6,0x7d,
0x0e,0x84,0x4a,0xc5,0x07,0xe5,0x75,0x6c,0x78,0x0f,
0xf0,0x26,0xa0,0x0b,0xa1,0x8f,0xdf,0xe7,0xca,0x78,
0x17,0xca,0x3b,0x78,0x4f,0x7b,0x0a,0x97,0x7e,0x3f,
0xa9,0x14,0x78,0x29,0xb8,0xf1,0x92,0x72,0xcb,0x0e,
0x34,0x9f,0x02,0x53,0x35,0x1a,0x0d,0x65,0x24,0xc5,
0xd2,0x5c,0x53,0x1c,0xa5,0x25,0xc6,0x52,0xd2,0xb4,
0x28,0xdf,0xbc,0x51,0x97,0x44,0xdd,0xb7,0x47,0xe8,
0x5a,0xcf,0x20,0x35,0x5e,0xb7,0xd1,0xc0,0x88,0x93,
0xa3,0x53,0x8a,0xf5,0xbb,0xd0,0x6e,0x2e,0x2a,0xcc,
0x6d,0x09,0xd4,0xad,0xd8,0x1c,0x73,0xc0,0x3b,0x06,
0x95,0x61,0xde,0xd1,0x97,0x00,0x2b,0xa0,0xec,0x59,
0x46,0x5a,0x97,0x93,0x42,0x89,0xd3,0xa6,0xdc,0x55,
0xa0,0x9a,0x6f,0xd0,0xeb,0x28,0x2d,0x21,0x56,0x60,
0xed,0xfc,0x64,0xaa,0x6f,0xbf,0x45,0xbf,0x5f,0xe9,
0xa6,0x21,0x87,0x8b,0x23,0x72,0x01,0x1c,0x2b,0x84,
0x13,0x67,0xfd,0x1c,0x90,0xd7,0x6b,0x95,0xf0,0x07,
0xa0,0x14,0xd8,0xa5,0xd5,0x20,0xe1,0x4b,0x67,0xd3,
0xa6,0x55,0xe9,0x94,0x10,0x1b,0x19,0x6c,0xde,0x38,
0xb0,0x2c,0x9e,0x3d,0x83,0x76,0x3e,0x95,0x23,0x22,
0x86,0x77,0xf1,0xc0,0xe9,0xe2,0xb2,0xda,0xfc,0x60,
0x73,0x7d,0x0e,0x78,0xd0,0x7a,0xc4,0x89,0xa8,0x7d,
0x1f,0xfd,0x6d,0x3a,0x58,0xdf,0xfe,0x44,0x36,0x3d,
0x02,0x65,0xca,0xd8,0xbd,0x40,0xaf,0xd7,0xd2,0xe6,
0xbc,0x4c,0xe1,0x0c,0xf4,0x19,0x80,0x23,0xd0,0x9d,
0xa9,0x8c,0x2b,0x0e,0x88,0x14,0x78,0x64,0x8f,0x4a,
0xca,0x6a,0x17,0xa3,0xbb,0x97,0xdf,0x15,0xae,0x98,
0x43,0xc9,0xd3,0xa3,0x43,0x92,0xb3,0x6f,0xe8,0x0e,
0x75,0xd9,0x86,0x45,0x28,0x53,0xe2,0x63,0xfc,0xd2,
0xa3,0x96,0xe7,0x96,0xa4,0x52,0xff,0xf0,0x28,0xf3,
0x23,0x11,0xdd,0x1f,0xa0,0x2f,0x4f,0xb1,0x39,0x8e,
0x03,0x00,0x13,0x4e,0x97,0x97,0x65,0xa2,0x2c,0x93,
0x91,0x82,0xd9,0xee,0xb8,0x65,0xa7,0xd3,0x0d,0x9d,
0xc2,0x01,0xb5,0x18,0xa3,0x0d,0xb4,0x6e,0x61,0x8a,
0x20,0xa9,0xbf,0x68,0x68,0xe3,0xca,0x0c,0x3a,0x50,
0xd9,0x48,0x0e,0xa7,0x7b,0x75,0xf1,0xf1,0xda,0x0d,
0x45,0xeb,0x97,0x1d,0xf7,0xe3,0x80,0xc7,0xe3,0xa5,
0xa2,0x63,0x35,0xf9,0x68,0x0b,0x22,0x41,0xa8,0xc7,
0xb2,0x66,0x06,0xcd,0x6f,0x5b,0xef,0x10,0x1d,0x39,
0xdf,0x42,0xbd,0x83,0x0e,0xd1,0x4f,0xc2,0xae,0x19,
0xfc,0xcc,0xbb,0x2c,0xbb,0xd8,0x4a,0x2d,0x37,0x07,
0xc6,0xad,0x33,0x20,0x1d,0xe6,0x79,0xb3,0x84,0x1d,
0x60,0xaf,0x48,0xb9,0xc7,0x2f,0x02,0xa2,0xc3,0xe7,
0x9c,0x56,0x64,0x24,0x52,0x64,0x84,0xce,0x17,0x22,
0x75,0x9d,0xf8,0xb5,0xbe,0x83,0x5c,0x92,0x47,0xf4,
0x5f,0x5c,0x96,0x46,0xf3,0x92,0x8d,0xe2,0xf9,0x2a,
0x8c,0x96,0xd7,0xb4,0x41,0xa9,0x87,0xaa,0xc1,0xfe,
0x39,0x20,0x5f,0xa0,0x2c,0x99,0x93,0x40,0xe7,0x9a,
0x6e,0xe0,0x64,0x38,0xb3,0xb1,0x59,0xb3,0x5c,0xb4,
0xe4,0x08,0x78,0x3d,0x3a,0xa0,0x00,0xa0,0xec,0x14,
0xa3,0x38,0xa1,0x81,0xe8,0xb3,0x8f,0x22,0xec,0x0e,
0x92,0x60,0x24,0x1d,0x06,0xb2,0x93,0xef,0xce,0xcb,
0x44,0xd8,0x8d,0xd1,0x11,0x62,0xec,0x06,0xea,0x81,
0xe0,0x54,0xc0,0x7a,0x2d,0x48,0xbd,0xe0,0x21,0x23,
0xdb,0x62,0xbc,0xc4,0xad,0x3a,0x05,0x8b,0x80,0xf8,
0xd8,0x29,0x11,0x20,0x53,0x94,0x50,0x10,0x88,0x98,
0x48,0xbd,0x00,0x13,0x63,0xcd,0x3c,0x93,0xdf,0x98,
0xfd,0x8e,0x8b,0x06,0x47,0x9c,0x24,0x21,0x3a,0x11,
0x30,0xe4,0x95,0x49,0x16,0x88,0xd4,0x84,0xa9,0x4a,
0x1a,0x56,0xf8,0xa5,0x00,0x83,0x99,0xdc,0x26,0xc5,
0x85,0x66,0x3d,0xe7,0x71,0x47,0x41,0x0e,0xb9,0x61,
0x84,0x53,0xa4,0xcc,0x03,0xb1,0x04,0x2f,0xd8,0x09,
0x91,0xc2,0xcc,0xa4,0x90,0x3a,0x94,0xcd,0x71,0xd0,
0xfc,0x4a,0x31,0xbc,0xe1,0x23,0x42,0x53,0x11,0x01,
0x4f,0x98,0xdb,0x99,0xcb,0x71,0x84,0x5e,0xe7,0x9b,
0xc3,0x46,0x7f,0x3c,0xdb,0x2c,0x48,0xc9,0xb2,0x30,
0x75,0x06,0xad,0xce,0x32,0x85,0xd4,0x11,0x1d,0x19,
0xa1,0x90,0x2f,0xd1,0xdf,0x01,0xd9,0x63,0x75,0x85,
0x9a,0x48,0x78,0xcd,0xa1,0x73,0x4d,0xc8,0xf9,0xb0,
0xe8,0x2f,0xcf,0x98,0x49,0xcf,0xe7,0xa6,0x85,0x5d,
0xcf,0x95,0x55,0x45,0x6e,0x2e,0xf7,0x92,0x12,0x81,
0x01,0x6e,0x47,0x46,0x5d,0xe3,0xd8,0x1f,0x4a,0x1a,
0xda,0xfb,0xa8,0xa3,0x6f,0x48,0x3c,0xaf,0x9c,0x6b,
0xa2,0x67,0x51,0xb2,0x27,0x5a,0x3b,0xcc,0xfa,0xc7,
0x22,0x30,0x20,0x5f,0xe1,0x4a,0x04,0x3c,0xad,0xdc,
0xf6,0x0c,0x8e,0x84,0x4d,0x81,0x5a,0xba,0x51,0x05,
0x99,0x0f,0x5a,0xa4,0x25,0x7f,0xd1,0xc3,0x93,0x5a,
0x67,0xbd,0xed,0x20,0x99,0xfd,0xad,0x81,0x1c,0xe0,
0x8f,0x09,0x67,0xf7,0xad,0x61,0x43,0xbf,0xdd,0x41,
0x71,0xd1,0x91,0x13,0x2a,0xe3,0xd2,0x2b,0x49,0x12,
0x65,0xa6,0x4c,0x27,0x83,0x4e,0x23,0x6a,0xc0,0x44,
0xd2,0xde,0x3b,0xa8,0x44,0xc0,0x12,0xe8,0x00,0xb3,
0xe8,0x6f,0x0f,0x79,0xcd,0x57,0x3a,0xfb,0x69,0xf9,
0xdc,0x99,0x13,0x2a,0x5b,0x38,0x3b,0x81,0xb2,0xd8,
0x38,0x48,0x29,0x4d,0x32,0x6d,0x97,0xda,0xfa,0x14,
0x07,0xaa,0xfc,0x1c,0x90,0xc6,0xc2,0xc2,0x5f,0x2f,
0xe6,0xbf,0x9a,0xba,0x69,0x49,0x7a,0xa2,0x08,0x6d,
0x38,0x69,0xee,0xb2,0x51,0x45,0xcd,0x7f,0xe2,0xe8,
0x6e,0x35,0xcf,0x27,0x9d,0x56,0x1b,0x76,0xfe,0x35,
0xeb,0x00,0x2e,0xaf,0x21,0x25,0xff,0x27,0x7c,0xc4,
0x14,0x0e,0x20,0x97,0xc0,0x61,0xa0,0xc5,0x0a,0x56,
0x5f,0xb8,0x7a,0x53,0xec,0x2a,0x1c,0xaa,0xea,0xdb,
0xa8,0xa5,0xbb,0x9f,0xce,0x5f,0xe9,0xa2,0x66,0xb4,
0xe1,0xe6,0xba,0x90,0x9e,0x93,0x96,0x56,0xc5,0xce,
0xbe,0xa2,0x4d,0xab,0x1c,0x92,0xa4,0xaa,0x84,0x5c,
0xdf,0x8b,0x37,0xaf,0x76,0xa2,0xdd,0xcd,0xcf,0x95,
0x75,0x6d,0x64,0xc3,0x6d,0x27,0x57,0xad,0xa0,0xc8,
0x49,0x4d,0xc0,0xa7,0x98,0x9b,0xa2,0x0c,0x7a,0x32,
0x19,0x63,0xc2,0xce,0xad,0x6e,0xec,0xa4,0xeb,0x38,
0x31,0xd0,0xdd,0x05,0x1c,0xe0,0x77,0x2e,0xb5,0x03,
0x6e,0xb7,0x24,0x8e,0xd0,0x87,0x5b,0xf2,0x8e,0xe2,
0xf9,0xa7,0xa1,0x91,0x51,0x2a,0x3d,0xf3,0x8f,0x28,
0x34,0xc1,0x4a,0x2a,0x23,0x6f,0x7e,0x0a,0x7d,0xfe,
0x5a,0x3e,0xed,0x7b,0xc5,0x4c,0x31,0xa2,0x80,0x05,
0x9f,0x57,0x77,0xcd,0x4a,0xa7,0xea,0x5a,0xd9,0x86,
0x04,0xbc,0x0e,0x1b,0x0e,0x7e,0xcf,0x36,0x55,0x11,
0x90,0x7c,0x0b,0xf0,0xbc,0x15,0x68,0xbc,0xde,0x37,
0x48,0x07,0x7e,0xb1,0x10,0xa7,0x24,0x94,0xf2,0x28,
0xdc,0x0d,0xa1,0xc6,0x18,0x67,0x1a,0xda,0xe9,0x50,
0xf5,0x65,0x1a,0x75,0xbb,0x59,0xef,0x7b,0x1f,0x6d,
0x5b,0x53,0xa9,0xb2,0xe3,0x1f,0x01,0xe5,0xa3,0xe4,
0x93,0xed,0x66,0x07,0xfa,0xcf,0x00,0x16,0xab,0xcd,
0x4e,0xfb,0x7f,0xae,0xa1,0x3f,0xfe,0xed,0x14,0x67,
0x5e,0xf5,0xe1,0x12,0x16,0x56,0xdc,0x88,0x07,0x4f,
0x37,0x50,0xc5,0x85,0x16,0xa1,0x1b,0x28,0x82,0xde,
0xfd,0xea,0x39,0x4a,0x04,0xc4,0x29,0x60,0xe5,0xea,
0x2a,0xf6,0xd9,0xab,0x6b,0xbb,0xde,0xf9,0xee,0xcc,
0xe3,0xfc,0x09,0x85,0xb1,0x0d,0xc7,0xfe,0x6c,0xa2,
0xdf,0x2e,0xb5,0xd3,0xca,0xec,0x64,0x7a,0x34,0x3d,
0x49,0x30,0x3f,0x50,0xf8,0x2a,0x6e,0xee,0xb4,0xd1,
0xa5,0xd6,0x1e,0xaa,0x05,0x89,0x65,0x7d,0x7c,0xbc,
0x5f,0xfe,0xe2,0x8d,0x75,0xc7,0x03,0xab,0xa4,0x5b,
0xe6,0x80,0x5e,0x59,0x1c,0x28,0x58,0xc4,0x8b,0x37,
0xbe,0xfd,0x6d,0x15,0xff,0x13,0xfa,0xb8,0x67,0x60,
0x78,0x41,0xc5,0xc5,0xab,0xc4,0x98,0x11,0x1b,0x45,
0xd1,0xc8,0x3b,0xb7,0x76,0x87,0x13,0x21,0x96,0xe8,
0x66,0xbf,0x9d,0x5c,0x6e,0x3f,0x3d,0x07,0xf9,0x7f,
0xcd,0x57,0x6f,0x15,0x74,0x04,0x3b,0x96,0x8a,0xcd,
0x90,0x0e,0x28,0x02,0x05,0x27,0x76,0x7e,0x73,0xea,
0xa4,0xfc,0x97,0x6c,0x0b,0x50,0x00,0x67,0xe2,0xf8,
0x34,0xb7,0x59,0x6f,0x8f,0x2b,0x0f,0xc0,0x51,0xc6,
0xd7,0x3b,0x9e,0xbe,0x1c,0xae,0x2e,0x28,0x36,0x35,
0x7c,0x7b,0x3d,0xb9,0xfb,0xf0,0xbd,0xfe,0xaf,0xcc,
0x55,0xb5,0xbd,0xfc,0xad,0xaa,0x6a,0x27,0x2d,0xd5,
0x7b,0xb7,0x8c,0x39,0xf0,0x20,0xe5,0x7f,0x01,0x06,
0x00,0x0a,0x2a,0x9d,0x13,0x76,0x62,0x16,0xe8,0x00,
0x00,0x00,0x00,0x49,0x45,0x4e,0x44,0xae,0x42,0x60,
0x82,

  };

  static FXIcon *VFLL6;
  static unsigned char YQRZ_32 []= {
0x89,0x50,0x4e,0x47,0x0d,0x0a,0x1a,0x0a,0x00,0x00,
0x00,0x0d,0x49,0x48,0x44,0x52,0x00,0x00,0x00,0x20,
0x00,0x00,0x00,0x20,0x08,0x06,0x00,0x00,0x00,0x73,
0x7a,0x7a,0xf4,0x00,0x00,0x00,0x19,0x74,0x45,0x58,
0x74,0x53,0x6f,0x66,0x74,0x77,0x61,0x72,0x65,0x00,
0x41,0x64,0x6f,0x62,0x65,0x20,0x49,0x6d,0x61,0x67,
0x65,0x52,0x65,0x61,0x64,0x79,0x71,0xc9,0x65,0x3c,
0x00,0x00,0x03,0xad,0x69,0x54,0x58,0x74,0x58,0x4d,
0x4c,0x3a,0x63,0x6f,0x6d,0x2e,0x61,0x64,0x6f,0x62,
0x65,0x2e,0x78,0x6d,0x70,0x00,0x00,0x00,0x00,0x00,
0x3c,0x3f,0x78,0x70,0x61,0x63,0x6b,0x65,0x74,0x20,
0x62,0x65,0x67,0x69,0x6e,0x3d,0x22,0xef,0xbb,0xbf,
0x22,0x20,0x69,0x64,0x3d,0x22,0x57,0x35,0x4d,0x30,
0x4d,0x70,0x43,0x65,0x68,0x69,0x48,0x7a,0x72,0x65,
0x53,0x7a,0x4e,0x54,0x63,0x7a,0x6b,0x63,0x39,0x64,
0x22,0x3f,0x3e,0x20,0x3c,0x78,0x3a,0x78,0x6d,0x70,
0x6d,0x65,0x74,0x61,0x20,0x78,0x6d,0x6c,0x6e,0x73,
0x3a,0x78,0x3d,0x22,0x61,0x64,0x6f,0x62,0x65,0x3a,
0x6e,0x73,0x3a,0x6d,0x65,0x74,0x61,0x2f,0x22,0x20,
0x78,0x3a,0x78,0x6d,0x70,0x74,0x6b,0x3d,0x22,0x41,
0x64,0x6f,0x62,0x65,0x20,0x58,0x4d,0x50,0x20,0x43,
0x6f,0x72,0x65,0x20,0x35,0x2e,0x30,0x2d,0x63,0x30,
0x36,0x30,0x20,0x36,0x31,0x2e,0x31,0x33,0x34,0x37,
0x37,0x37,0x2c,0x20,0x32,0x30,0x31,0x30,0x2f,0x30,
0x32,0x2f,0x31,0x32,0x2d,0x31,0x37,0x3a,0x33,0x32,
0x3a,0x30,0x30,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x22,0x3e,0x20,0x3c,0x72,0x64,0x66,0x3a,0x52,
0x44,0x46,0x20,0x78,0x6d,0x6c,0x6e,0x73,0x3a,0x72,
0x64,0x66,0x3d,0x22,0x68,0x74,0x74,0x70,0x3a,0x2f,
0x2f,0x77,0x77,0x77,0x2e,0x77,0x33,0x2e,0x6f,0x72,
0x67,0x2f,0x31,0x39,0x39,0x39,0x2f,0x30,0x32,0x2f,
0x32,0x32,0x2d,0x72,0x64,0x66,0x2d,0x73,0x79,0x6e,
0x74,0x61,0x78,0x2d,0x6e,0x73,0x23,0x22,0x3e,0x20,
0x3c,0x72,0x64,0x66,0x3a,0x44,0x65,0x73,0x63,0x72,
0x69,0x70,0x74,0x69,0x6f,0x6e,0x20,0x72,0x64,0x66,
0x3a,0x61,0x62,0x6f,0x75,0x74,0x3d,0x22,0x22,0x20,
0x78,0x6d,0x6c,0x6e,0x73,0x3a,0x78,0x6d,0x70,0x52,
0x69,0x67,0x68,0x74,0x73,0x3d,0x22,0x68,0x74,0x74,
0x70,0x3a,0x2f,0x2f,0x6e,0x73,0x2e,0x61,0x64,0x6f,
0x62,0x65,0x2e,0x63,0x6f,0x6d,0x2f,0x78,0x61,0x70,
0x2f,0x31,0x2e,0x30,0x2f,0x72,0x69,0x67,0x68,0x74,
0x73,0x2f,0x22,0x20,0x78,0x6d,0x6c,0x6e,0x73,0x3a,
0x78,0x6d,0x70,0x4d,0x4d,0x3d,0x22,0x68,0x74,0x74,
0x70,0x3a,0x2f,0x2f,0x6e,0x73,0x2e,0x61,0x64,0x6f,
0x62,0x65,0x2e,0x63,0x6f,0x6d,0x2f,0x78,0x61,0x70,
0x2f,0x31,0x2e,0x30,0x2f,0x6d,0x6d,0x2f,0x22,0x20,
0x78,0x6d,0x6c,0x6e,0x73,0x3a,0x73,0x74,0x52,0x65,
0x66,0x3d,0x22,0x68,0x74,0x74,0x70,0x3a,0x2f,0x2f,
0x6e,0x73,0x2e,0x61,0x64,0x6f,0x62,0x65,0x2e,0x63,
0x6f,0x6d,0x2f,0x78,0x61,0x70,0x2f,0x31,0x2e,0x30,
0x2f,0x73,0x54,0x79,0x70,0x65,0x2f,0x52,0x65,0x73,
0x6f,0x75,0x72,0x63,0x65,0x52,0x65,0x66,0x23,0x22,
0x20,0x78,0x6d,0x6c,0x6e,0x73,0x3a,0x78,0x6d,0x70,
0x3d,0x22,0x68,0x74,0x74,0x70,0x3a,0x2f,0x2f,0x6e,
0x73,0x2e,0x61,0x64,0x6f,0x62,0x65,0x2e,0x63,0x6f,
0x6d,0x2f,0x78,0x61,0x70,0x2f,0x31,0x2e,0x30,0x2f,
0x22,0x20,0x78,0x6d,0x70,0x52,0x69,0x67,0x68,0x74,
0x73,0x3a,0x4d,0x61,0x72,0x6b,0x65,0x64,0x3d,0x22,
0x46,0x61,0x6c,0x73,0x65,0x22,0x20,0x78,0x6d,0x70,
0x4d,0x4d,0x3a,0x4f,0x72,0x69,0x67,0x69,0x6e,0x61,
0x6c,0x44,0x6f,0x63,0x75,0x6d,0x65,0x6e,0x74,0x49,
0x44,0x3d,0x22,0x75,0x75,0x69,0x64,0x3a,0x41,0x45,
0x33,0x46,0x34,0x36,0x30,0x38,0x38,0x46,0x38,0x39,
0x44,0x45,0x31,0x31,0x41,0x38,0x36,0x30,0x39,0x39,
0x46,0x31,0x32,0x31,0x41,0x32,0x30,0x34,0x32,0x39,
0x22,0x20,0x78,0x6d,0x70,0x4d,0x4d,0x3a,0x44,0x6f,
0x63,0x75,0x6d,0x65,0x6e,0x74,0x49,0x44,0x3d,0x22,
0x78,0x6d,0x70,0x2e,0x64,0x69,0x64,0x3a,0x39,0x33,
0x31,0x31,0x35,0x34,0x33,0x35,0x42,0x35,0x45,0x46,
0x31,0x31,0x45,0x30,0x42,0x41,0x41,0x33,0x42,0x42,
0x45,0x34,0x44,0x37,0x42,0x31,0x33,0x35,0x31,0x44,
0x22,0x20,0x78,0x6d,0x70,0x4d,0x4d,0x3a,0x49,0x6e,
0x73,0x74,0x61,0x6e,0x63,0x65,0x49,0x44,0x3d,0x22,
0x78,0x6d,0x70,0x2e,0x69,0x69,0x64,0x3a,0x39,0x33,
0x31,0x31,0x35,0x34,0x33,0x34,0x42,0x35,0x45,0x46,
0x31,0x31,0x45,0x30,0x42,0x41,0x41,0x33,0x42,0x42,
0x45,0x34,0x44,0x37,0x42,0x31,0x33,0x35,0x31,0x44,
0x22,0x20,0x78,0x6d,0x70,0x3a,0x43,0x72,0x65,0x61,
0x74,0x6f,0x72,0x54,0x6f,0x6f,0x6c,0x3d,0x22,0x41,
0x64,0x6f,0x62,0x65,0x20,0x50,0x68,0x6f,0x74,0x6f,
0x73,0x68,0x6f,0x70,0x20,0x43,0x53,0x35,0x20,0x57,
0x69,0x6e,0x64,0x6f,0x77,0x73,0x22,0x3e,0x20,0x3c,
0x78,0x6d,0x70,0x4d,0x4d,0x3a,0x44,0x65,0x72,0x69,
0x76,0x65,0x64,0x46,0x72,0x6f,0x6d,0x20,0x73,0x74,
0x52,0x65,0x66,0x3a,0x69,0x6e,0x73,0x74,0x61,0x6e,
0x63,0x65,0x49,0x44,0x3d,0x22,0x78,0x6d,0x70,0x2e,
0x69,0x69,0x64,0x3a,0x33,0x32,0x45,0x42,0x39,0x34,
0x34,0x44,0x45,0x46,0x42,0x35,0x45,0x30,0x31,0x31,
0x39,0x41,0x42,0x34,0x41,0x46,0x44,0x39,0x30,0x33,
0x39,0x46,0x33,0x35,0x45,0x42,0x22,0x20,0x73,0x74,
0x52,0x65,0x66,0x3a,0x64,0x6f,0x63,0x75,0x6d,0x65,
0x6e,0x74,0x49,0x44,0x3d,0x22,0x75,0x75,0x69,0x64,
0x3a,0x41,0x45,0x33,0x46,0x34,0x36,0x30,0x38,0x38,
0x46,0x38,0x39,0x44,0x45,0x31,0x31,0x41,0x38,0x36,
0x30,0x39,0x39,0x46,0x31,0x32,0x31,0x41,0x32,0x30,
0x34,0x32,0x39,0x22,0x2f,0x3e,0x20,0x3c,0x2f,0x72,
0x64,0x66,0x3a,0x44,0x65,0x73,0x63,0x72,0x69,0x70,
0x74,0x69,0x6f,0x6e,0x3e,0x20,0x3c,0x2f,0x72,0x64,
0x66,0x3a,0x52,0x44,0x46,0x3e,0x20,0x3c,0x2f,0x78,
0x3a,0x78,0x6d,0x70,0x6d,0x65,0x74,0x61,0x3e,0x20,
0x3c,0x3f,0x78,0x70,0x61,0x63,0x6b,0x65,0x74,0x20,
0x65,0x6e,0x64,0x3d,0x22,0x72,0x22,0x3f,0x3e,0xda,
0x14,0xa0,0x85,0x00,0x00,0x03,0x0b,0x49,0x44,0x41,
0x54,0x78,0xda,0xc4,0x97,0x5f,0x84,0xd4,0x51,0x14,
0xc7,0x67,0x9b,0x0c,0xc3,0x30,0x2c,0x43,0xc4,0xb2,
0x0f,0xd9,0xc4,0x52,0x86,0x88,0x4d,0x89,0xc9,0xb0,
0x6b,0xd9,0x6d,0x88,0xa5,0x87,0x5e,0x8a,0xb4,0x76,
0x5f,0x62,0xd8,0x64,0x5f,0xd2,0x3c,0xad,0x62,0x2d,
0x2d,0xa9,0x97,0x96,0x28,0xb1,0x1a,0xa5,0x79,0xd9,
0x34,0x4f,0x69,0x49,0x91,0x56,0x43,0x44,0x2c,0xb1,
0x44,0x5a,0x7e,0xbb,0xbf,0xdf,0xaf,0xef,0xc9,0x19,
0x7b,0x7e,0xa7,0x73,0x7f,0xf3,0xc7,0xae,0x3d,0xeb,
0xb3,0x33,0xf7,0xfc,0xce,0xbd,0xf7,0xfc,0xee,0x3d,
0xe7,0xdc,0x3b,0x3d,0x61,0x18,0x26,0xf6,0x53,0x0e,
0xd2,0xbf,0xd9,0x67,0xab,0x9d,0xf6,0xcb,0x80,0x01,
0xd1,0xde,0x00,0xdf,0x3a,0x19,0x60,0xf6,0x42,0x7e,
0xc7,0x81,0x2e,0x64,0x06,0x94,0x45,0xfb,0x3b,0x38,
0x02,0xbc,0x4e,0x07,0x3a,0xd0,0xc5,0xe4,0x87,0xc0,
0x75,0xa5,0xeb,0x03,0x97,0xba,0x79,0x93,0x7f,0x0e,
0x84,0xf6,0x5f,0x3f,0x18,0x05,0x29,0xa5,0x2f,0x83,
0x8c,0x61,0x3f,0x03,0x92,0x4a,0x47,0xed,0x9c,0x35,
0x7a,0x24,0x06,0x8c,0x38,0xec,0x07,0x6f,0xc1,0x61,
0xf0,0x0b,0x2c,0x83,0x37,0xe0,0x22,0x28,0x38,0x5e,
0x86,0xfa,0xbc,0x04,0xf3,0xc0,0x07,0x25,0x30,0x0a,
0x36,0xc1,0x59,0xd0,0x70,0x06,0xa1,0xca,0x04,0x39,
0x39,0x49,0x96,0x97,0xb7,0x9d,0x25,0x2e,0x38,0x1c,
0x5c,0x71,0x39,0x61,0x39,0x50,0x16,0x93,0xef,0x96,
0xd0,0x78,0x2f,0xc0,0x51,0xd3,0x81,0x20,0xea,0x00,
0x2d,0xf7,0x95,0x3d,0x48,0xf9,0xd5,0x98,0x2d,0x88,
0xe8,0x6a,0x9c,0xd7,0xbd,0x8e,0x81,0x68,0x4f,0xab,
0x6c,0x27,0x97,0x7e,0x18,0xa4,0x63,0x1c,0x58,0x72,
0x3a,0x10,0x04,0x11,0x0f,0x28,0x97,0xdf,0x81,0xa2,
0x61,0xff,0x1c,0x5c,0x03,0xeb,0x4a,0xbf,0xc8,0xe9,
0xb9,0x00,0xc6,0x8c,0x7e,0x1b,0xca,0x61,0x95,0x86,
0x58,0x02,0x41,0x06,0x0c,0x29,0x1d,0xb1,0x08,0xc6,
0xc1,0xba,0xf1,0x2c,0xc1,0xfa,0x71,0xb6,0xd3,0xcf,
0x92,0x20,0x2d,0x75,0x11,0x07,0x82,0x30,0x68,0x32,
0x00,0x1e,0x82,0x8c,0xd0,0x11,0x1f,0xc0,0xa4,0xd2,
0x11,0x83,0x86,0x6e,0x92,0xed,0xa5,0x2e,0x0b,0x6e,
0x48,0x5d,0xd4,0x81,0x20,0x1c,0x04,0x8f,0xc1,0x17,
0x50,0xa2,0x2d,0x51,0xdc,0x06,0x9e,0x68,0x4f,0x81,
0x3f,0xe0,0x23,0x7f,0x4e,0x89,0x67,0x1e,0xdb,0xeb,
0x31,0x66,0xc0,0x7b,0x50,0x90,0x5b,0xce,0x2b,0x10,
0x96,0xc1,0x04,0x65,0x83,0x81,0x07,0xaa,0xa2,0x5d,
0x02,0x77,0x41,0x9a,0xdb,0x69,0x6e,0x97,0x84,0x0d,
0xd9,0x6f,0x1a,0x63,0xe5,0xc1,0x6b,0xb2,0x8f,0x0b,
0x42,0x2d,0x0d,0x8e,0xfc,0xa6,0xb8,0x52,0x94,0xf4,
0x4f,0x45,0xa6,0xd0,0xe9,0x78,0xcc,0x61,0x9b,0x8b,
0xab,0x03,0x5a,0x52,0xc6,0x51,0xec,0x3a,0xa2,0xe3,
0xfa,0x25,0xe2,0xd2,0xf0,0x67,0x8c,0x4d,0x1f,0xe7,
0xf7,0xa6,0x28,0x54,0xa7,0x0c,0xbb,0x65,0xf1,0x3d,
0xcd,0xfd,0xda,0x3b,0x0d,0x11,0x95,0xd3,0x60,0xc4,
0x88,0x5e,0x22,0x05,0x8a,0xa2,0x3d,0x07,0x6a,0xca,
0xa6,0xc6,0xfa,0x66,0xbb,0xc8,0xfd,0x12,0x06,0x4b,
0xa0,0x62,0xc5,0x40,0x95,0xb9,0xa3,0x2e,0x1b,0x24,
0xb7,0xf8,0x0d,0x7d,0x2e,0x54,0xe7,0xc1,0x10,0xef,
0xf1,0x67,0x50,0x17,0xb6,0x49,0xb6,0xd7,0x42,0xf1,
0x71,0x13,0xac,0xb5,0xaa,0x84,0x15,0xbe,0x70,0xc8,
0x3d,0x3d,0xce,0x55,0xee,0xaa,0xd0,0xd5,0xd5,0xc4,
0x4d,0x59,0x60,0x7b,0x29,0xbf,0xc1,0x65,0xfe,0xfc,
0x3f,0x06,0x7c,0x51,0x18,0x78,0xaf,0x3d,0x47,0x94,
0xe7,0x1c,0xa5,0x38,0xd1,0xa2,0x14,0xd7,0xad,0xc9,
0x77,0x1c,0xf0,0x03,0x7d,0xa6,0xbb,0x0e,0xa2,0x31,
0x7e,0xfe,0xca,0x38,0x8c,0x8a,0x31,0x19,0x72,0x92,
0xb3,0xc2,0x33,0x1d,0xd8,0x8a,0x3a,0x30,0xd1,0xc6,
0x8d,0xb8,0xc4,0xb4,0x2b,0xbd,0xec,0x64,0xd5,0x74,
0x60,0x7b,0xdb,0x97,0xba,0xfc,0x1e,0xfd,0x04,0x18,
0x76,0x3a,0xb0,0xe5,0x47,0x1c,0x18,0xe1,0x2b,0xd4,
0x6e,0xde,0x8a,0x7e,0x80,0x39,0x67,0x0c,0xa8,0x15,
0x68,0xf0,0xfd,0x6d,0x85,0x0b,0x4a,0x95,0x53,0x28,
0xc9,0xd9,0x71,0x2e,0x66,0x22,0x8a,0x8b,0x27,0xe0,
0x0c,0x5f,0x48,0xb3,0x3c,0x79,0xfc,0xa5,0x74,0x3b,
0x1a,0x03,0x4d,0x27,0x4e,0xf0,0x45,0xc2,0x57,0xd5,
0xee,0x2b,0x5f,0x5c,0xb5,0xdc,0x03,0xd3,0xfc,0xfd,
0x01,0x07,0x1d,0x2d,0xfb,0x27,0xd7,0xe4,0x3b,0x59,
0x10,0x04,0xd6,0x33,0xab,0x3c,0xfb,0x5c,0x27,0xee,
0x1b,0x79,0x5e,0x51,0x3a,0x8f,0x6f,0x50,0xad,0xcf,
0x02,0x87,0x03,0x2e,0x79,0xc4,0x95,0x4e,0xc6,0xc8,
0xbc,0xa3,0x36,0xb4,0x7d,0x18,0x75,0xd2,0x87,0xde,
0xec,0xb4,0xaa,0x15,0x6b,0xdd,0x46,0x67,0xcf,0x7e,
0xff,0x3c,0xff,0x2b,0xc0,0x00,0x24,0x67,0xba,0xfb,
0x88,0xf3,0xd4,0xdd,0x00,0x00,0x00,0x00,0x49,0x45,
0x4e,0x44,0xae,0x42,0x60,0x82,

  };

  static FXIcon *VFLL7;
  static unsigned char YQRZ_45 []= {
0x89,0x50,0x4e,0x47,0x0d,0x0a,0x1a,0x0a,0x00,0x00,
0x00,0x0d,0x49,0x48,0x44,0x52,0x00,0x00,0x00,0x20,
0x00,0x00,0x00,0x20,0x08,0x06,0x00,0x00,0x00,0x73,
0x7a,0x7a,0xf4,0x00,0x00,0x00,0x19,0x74,0x45,0x58,
0x74,0x53,0x6f,0x66,0x74,0x77,0x61,0x72,0x65,0x00,
0x41,0x64,0x6f,0x62,0x65,0x20,0x49,0x6d,0x61,0x67,
0x65,0x52,0x65,0x61,0x64,0x79,0x71,0xc9,0x65,0x3c,
0x00,0x00,0x03,0xad,0x69,0x54,0x58,0x74,0x58,0x4d,
0x4c,0x3a,0x63,0x6f,0x6d,0x2e,0x61,0x64,0x6f,0x62,
0x65,0x2e,0x78,0x6d,0x70,0x00,0x00,0x00,0x00,0x00,
0x3c,0x3f,0x78,0x70,0x61,0x63,0x6b,0x65,0x74,0x20,
0x62,0x65,0x67,0x69,0x6e,0x3d,0x22,0xef,0xbb,0xbf,
0x22,0x20,0x69,0x64,0x3d,0x22,0x57,0x35,0x4d,0x30,
0x4d,0x70,0x43,0x65,0x68,0x69,0x48,0x7a,0x72,0x65,
0x53,0x7a,0x4e,0x54,0x63,0x7a,0x6b,0x63,0x39,0x64,
0x22,0x3f,0x3e,0x20,0x3c,0x78,0x3a,0x78,0x6d,0x70,
0x6d,0x65,0x74,0x61,0x20,0x78,0x6d,0x6c,0x6e,0x73,
0x3a,0x78,0x3d,0x22,0x61,0x64,0x6f,0x62,0x65,0x3a,
0x6e,0x73,0x3a,0x6d,0x65,0x74,0x61,0x2f,0x22,0x20,
0x78,0x3a,0x78,0x6d,0x70,0x74,0x6b,0x3d,0x22,0x41,
0x64,0x6f,0x62,0x65,0x20,0x58,0x4d,0x50,0x20,0x43,
0x6f,0x72,0x65,0x20,0x35,0x2e,0x30,0x2d,0x63,0x30,
0x36,0x30,0x20,0x36,0x31,0x2e,0x31,0x33,0x34,0x37,
0x37,0x37,0x2c,0x20,0x32,0x30,0x31,0x30,0x2f,0x30,
0x32,0x2f,0x31,0x32,0x2d,0x31,0x37,0x3a,0x33,0x32,
0x3a,0x30,0x30,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x22,0x3e,0x20,0x3c,0x72,0x64,0x66,0x3a,0x52,
0x44,0x46,0x20,0x78,0x6d,0x6c,0x6e,0x73,0x3a,0x72,
0x64,0x66,0x3d,0x22,0x68,0x74,0x74,0x70,0x3a,0x2f,
0x2f,0x77,0x77,0x77,0x2e,0x77,0x33,0x2e,0x6f,0x72,
0x67,0x2f,0x31,0x39,0x39,0x39,0x2f,0x30,0x32,0x2f,
0x32,0x32,0x2d,0x72,0x64,0x66,0x2d,0x73,0x79,0x6e,
0x74,0x61,0x78,0x2d,0x6e,0x73,0x23,0x22,0x3e,0x20,
0x3c,0x72,0x64,0x66,0x3a,0x44,0x65,0x73,0x63,0x72,
0x69,0x70,0x74,0x69,0x6f,0x6e,0x20,0x72,0x64,0x66,
0x3a,0x61,0x62,0x6f,0x75,0x74,0x3d,0x22,0x22,0x20,
0x78,0x6d,0x6c,0x6e,0x73,0x3a,0x78,0x6d,0x70,0x52,
0x69,0x67,0x68,0x74,0x73,0x3d,0x22,0x68,0x74,0x74,
0x70,0x3a,0x2f,0x2f,0x6e,0x73,0x2e,0x61,0x64,0x6f,
0x62,0x65,0x2e,0x63,0x6f,0x6d,0x2f,0x78,0x61,0x70,
0x2f,0x31,0x2e,0x30,0x2f,0x72,0x69,0x67,0x68,0x74,
0x73,0x2f,0x22,0x20,0x78,0x6d,0x6c,0x6e,0x73,0x3a,
0x78,0x6d,0x70,0x4d,0x4d,0x3d,0x22,0x68,0x74,0x74,
0x70,0x3a,0x2f,0x2f,0x6e,0x73,0x2e,0x61,0x64,0x6f,
0x62,0x65,0x2e,0x63,0x6f,0x6d,0x2f,0x78,0x61,0x70,
0x2f,0x31,0x2e,0x30,0x2f,0x6d,0x6d,0x2f,0x22,0x20,
0x78,0x6d,0x6c,0x6e,0x73,0x3a,0x73,0x74,0x52,0x65,
0x66,0x3d,0x22,0x68,0x74,0x74,0x70,0x3a,0x2f,0x2f,
0x6e,0x73,0x2e,0x61,0x64,0x6f,0x62,0x65,0x2e,0x63,
0x6f,0x6d,0x2f,0x78,0x61,0x70,0x2f,0x31,0x2e,0x30,
0x2f,0x73,0x54,0x79,0x70,0x65,0x2f,0x52,0x65,0x73,
0x6f,0x75,0x72,0x63,0x65,0x52,0x65,0x66,0x23,0x22,
0x20,0x78,0x6d,0x6c,0x6e,0x73,0x3a,0x78,0x6d,0x70,
0x3d,0x22,0x68,0x74,0x74,0x70,0x3a,0x2f,0x2f,0x6e,
0x73,0x2e,0x61,0x64,0x6f,0x62,0x65,0x2e,0x63,0x6f,
0x6d,0x2f,0x78,0x61,0x70,0x2f,0x31,0x2e,0x30,0x2f,
0x22,0x20,0x78,0x6d,0x70,0x52,0x69,0x67,0x68,0x74,
0x73,0x3a,0x4d,0x61,0x72,0x6b,0x65,0x64,0x3d,0x22,
0x46,0x61,0x6c,0x73,0x65,0x22,0x20,0x78,0x6d,0x70,
0x4d,0x4d,0x3a,0x4f,0x72,0x69,0x67,0x69,0x6e,0x61,
0x6c,0x44,0x6f,0x63,0x75,0x6d,0x65,0x6e,0x74,0x49,
0x44,0x3d,0x22,0x75,0x75,0x69,0x64,0x3a,0x41,0x45,
0x33,0x46,0x34,0x36,0x30,0x38,0x38,0x46,0x38,0x39,
0x44,0x45,0x31,0x31,0x41,0x38,0x36,0x30,0x39,0x39,
0x46,0x31,0x32,0x31,0x41,0x32,0x30,0x34,0x32,0x39,
0x22,0x20,0x78,0x6d,0x70,0x4d,0x4d,0x3a,0x44,0x6f,
0x63,0x75,0x6d,0x65,0x6e,0x74,0x49,0x44,0x3d,0x22,
0x78,0x6d,0x70,0x2e,0x64,0x69,0x64,0x3a,0x38,0x43,
0x46,0x41,0x31,0x32,0x34,0x43,0x42,0x35,0x45,0x46,
0x31,0x31,0x45,0x30,0x42,0x41,0x41,0x33,0x42,0x42,
0x45,0x34,0x44,0x37,0x42,0x31,0x33,0x35,0x31,0x44,
0x22,0x20,0x78,0x6d,0x70,0x4d,0x4d,0x3a,0x49,0x6e,
0x73,0x74,0x61,0x6e,0x63,0x65,0x49,0x44,0x3d,0x22,
0x78,0x6d,0x70,0x2e,0x69,0x69,0x64,0x3a,0x38,0x43,
0x46,0x41,0x31,0x32,0x34,0x42,0x42,0x35,0x45,0x46,
0x31,0x31,0x45,0x30,0x42,0x41,0x41,0x33,0x42,0x42,
0x45,0x34,0x44,0x37,0x42,0x31,0x33,0x35,0x31,0x44,
0x22,0x20,0x78,0x6d,0x70,0x3a,0x43,0x72,0x65,0x61,
0x74,0x6f,0x72,0x54,0x6f,0x6f,0x6c,0x3d,0x22,0x41,
0x64,0x6f,0x62,0x65,0x20,0x50,0x68,0x6f,0x74,0x6f,
0x73,0x68,0x6f,0x70,0x20,0x43,0x53,0x35,0x20,0x57,
0x69,0x6e,0x64,0x6f,0x77,0x73,0x22,0x3e,0x20,0x3c,
0x78,0x6d,0x70,0x4d,0x4d,0x3a,0x44,0x65,0x72,0x69,
0x76,0x65,0x64,0x46,0x72,0x6f,0x6d,0x20,0x73,0x74,
0x52,0x65,0x66,0x3a,0x69,0x6e,0x73,0x74,0x61,0x6e,
0x63,0x65,0x49,0x44,0x3d,0x22,0x78,0x6d,0x70,0x2e,
0x69,0x69,0x64,0x3a,0x33,0x32,0x45,0x42,0x39,0x34,
0x34,0x44,0x45,0x46,0x42,0x35,0x45,0x30,0x31,0x31,
0x39,0x41,0x42,0x34,0x41,0x46,0x44,0x39,0x30,0x33,
0x39,0x46,0x33,0x35,0x45,0x42,0x22,0x20,0x73,0x74,
0x52,0x65,0x66,0x3a,0x64,0x6f,0x63,0x75,0x6d,0x65,
0x6e,0x74,0x49,0x44,0x3d,0x22,0x75,0x75,0x69,0x64,
0x3a,0x41,0x45,0x33,0x46,0x34,0x36,0x30,0x38,0x38,
0x46,0x38,0x39,0x44,0x45,0x31,0x31,0x41,0x38,0x36,
0x30,0x39,0x39,0x46,0x31,0x32,0x31,0x41,0x32,0x30,
0x34,0x32,0x39,0x22,0x2f,0x3e,0x20,0x3c,0x2f,0x72,
0x64,0x66,0x3a,0x44,0x65,0x73,0x63,0x72,0x69,0x70,
0x74,0x69,0x6f,0x6e,0x3e,0x20,0x3c,0x2f,0x72,0x64,
0x66,0x3a,0x52,0x44,0x46,0x3e,0x20,0x3c,0x2f,0x78,
0x3a,0x78,0x6d,0x70,0x6d,0x65,0x74,0x61,0x3e,0x20,
0x3c,0x3f,0x78,0x70,0x61,0x63,0x6b,0x65,0x74,0x20,
0x65,0x6e,0x64,0x3d,0x22,0x72,0x22,0x3f,0x3e,0x1a,
0x6a,0xcd,0xe1,0x00,0x00,0x02,0xeb,0x49,0x44,0x41,
0x54,0x78,0xda,0xc4,0x57,0x41,0x68,0x13,0x41,0x14,
0xfd,0x93,0xa4,0x8d,0xb5,0x9a,0x9a,0xa4,0x11,0x31,
0x07,0x41,0x3c,0xb4,0xd1,0x48,0x60,0xd5,0x52,0x3d,
0xe8,0x45,0x2f,0x22,0x48,0x2b,0x3d,0x15,0x11,0x2c,
0x54,0x90,0x22,0x82,0x20,0x1e,0x84,0x9e,0x3c,0x56,
0x88,0xb4,0x14,0x11,0x11,0xf1,0xe0,0xa5,0x50,0xf1,
0xea,0x45,0x10,0x69,0xc1,0x1e,0xb4,0x57,0x11,0x3d,
0x08,0x05,0x0d,0x81,0x1e,0x4c,0xb2,0xbb,0x33,0xdf,
0x99,0x49,0x67,0xbb,0xd9,0xdd,0x49,0xb2,0x69,0x4a,
0x07,0x1e,0x33,0x99,0x99,0xff,0xfe,0x9b,0xff,0x77,
0xff,0x64,0x09,0x22,0xc2,0x5e,0xb6,0x98,0x1a,0xcc,
0x2e,0xad,0x75,0x62,0xaf,0xd4,0x93,0xb0,0x86,0xb3,
0xe3,0x86,0xec,0x23,0x3b,0x10,0x8f,0xd1,0x08,0x01,
0x01,0x97,0x90,0xce,0x23,0x10,0x32,0x15,0xd2,0x79,
0x64,0xeb,0xdc,0xc8,0x7b,0xca,0x24,0x01,0xe9,0x58,
0x00,0x6b,0x5f,0x80,0xdc,0x88,0x14,0xe1,0x6a,0xef,
0x1b,0x39,0xf1,0xae,0x32,0xa9,0x42,0x10,0x5a,0x84,
0x23,0x80,0x32,0x16,0x26,0xe7,0x30,0xbe,0xff,0x35,
0x18,0xa7,0xb2,0xf5,0x1f,0xeb,0xaf,0x60,0xe9,0xdf,
0x4d,0x67,0xcf,0x93,0xd1,0x55,0x9f,0x21,0x39,0x76,
0x2f,0xf6,0xe8,0xed,0x0a,0xf5,0xce,0x3b,0xcf,0x00,
0xa5,0xac,0x15,0x50,0x8d,0xc7,0xfb,0x5e,0x82,0x71,
0xf2,0x30,0x0f,0x9b,0x25,0x21,0xc6,0x57,0xd8,0xe2,
0x9a,0x5a,0x7f,0xf8,0xe9,0x2c,0x80,0x5d,0x6e,0x04,
0x40,0xdc,0xcd,0xe7,0x17,0xc0,0xb0,0x19,0x50,0x8d,
0x27,0xfa,0x5f,0x70,0x87,0x19,0x7e,0x4e,0xab,0x01,
0x97,0xce,0x65,0x8d,0xb1,0x7d,0xcf,0x4b,0x6a,0xdf,
0x83,0xcf,0x97,0x01,0xac,0xd2,0x36,0x00,0xa2,0x6e,
0xce,0x30,0x11,0xd8,0x3e,0x79,0xef,0xc2,0x6f,0x23,
0x97,0xe6,0xa7,0x36,0x03,0x31,0x92,0xcf,0xa4,0xc7,
0x7a,0x17,0x36,0xd4,0xfe,0xfb,0x2b,0x13,0x00,0xe6,
0xdf,0x3a,0x3c,0x3e,0x7c,0x02,0x6c,0x46,0x83,0x80,
0x6a,0x7c,0x23,0x5e,0xfc,0x3e,0x5a,0x18,0xcc,0x02,
0x72,0x67,0x4d,0x70,0xbe,0x30,0x78,0x64,0x32,0x31,
0xef,0xd8,0xcd,0x7c,0x99,0x06,0xbb,0xba,0xe1,0xf3,
0xe1,0x17,0xc0,0x55,0x79,0x80,0x6a,0x3c,0xd9,0xff,
0xac,0x74,0xa1,0x90,0x39,0xa1,0x72,0xde,0x0a,0xc6,
0x70,0x8a,0xdc,0x4a,0xcc,0x3b,0x5c,0x33,0x5f,0x1f,
0xfb,0x7c,0xb4,0x12,0xe0,0x38,0xbf,0x3d,0x50,0xac,
0x8d,0xe4,0x93,0xe9,0x56,0x27,0xf7,0xc2,0xc8,0x0d,
0xc0,0xd4,0xa1,0xa2,0xc3,0x39,0xb5,0xf8,0x61,0xb3,
0xa9,0x00,0x4a,0xa9,0x02,0xaa,0xf1,0x9d,0xd4,0x53,
0x38,0x93,0x4b,0xc4,0x75,0x39,0x6f,0x05,0x63,0x38,
0x21,0x39,0x82,0xb8,0x7d,0x02,0x6a,0x96,0x2d,0x80,
0x5b,0x3d,0xdc,0x4d,0xcf,0x81,0x31,0x94,0x80,0x76,
0xc3,0xae,0x01,0x0a,0x0e,0xc1,0xa5,0x78,0x95,0x0f,
0x5f,0x21,0xaa,0xd4,0x2c,0x77,0x29,0x24,0xc6,0xd0,
0x01,0x94,0x27,0xd9,0x59,0x93,0x55,0x91,0x73,0x41,
0xe5,0x97,0x45,0x5c,0x85,0xcc,0xa9,0x98,0x8e,0x80,
0x9a,0x69,0x37,0x18,0x21,0xab,0x75,0xf5,0xda,0xdd,
0xe2,0x27,0xde,0x8b,0x2b,0xe6,0x4e,0x81,0xbb,0x8e,
0x33,0xda,0x65,0x01,0x56,0xc3,0x01,0x1d,0x11,0x44,
0x77,0x0b,0x5a,0xab,0x17,0xbb,0xfa,0x4f,0xa5,0x67,
0xe4,0x23,0x69,0x7a,0x19,0x79,0x1b,0xa3,0xd5,0xc0,
0xf9,0x4a,0x8d,0x41,0x79,0xd3,0xe6,0xaf,0x52,0xa3,
0xbe,0x58,0x94,0x40,0x32,0x11,0x83,0xbe,0x78,0xa4,
0xb3,0xeb,0xb8,0x5d,0x01,0xa5,0xb2,0x05,0xc7,0xaf,
0x7f,0x0b,0x5c,0xfb,0xb1,0x7c,0x1a,0x8e,0x66,0x7a,
0x76,0x57,0x80,0x69,0x59,0xfc,0xcf,0x0b,0xd5,0xac,
0x55,0xb9,0x1d,0xed,0x96,0x80,0xe0,0x87,0x10,0xa9,
0x0d,0x8c,0xd9,0x9a,0x35,0x93,0xdb,0xb1,0xdd,0x15,
0xc0,0xf8,0x45,0x82,0x68,0x6b,0xd6,0xcc,0xd0,0x11,
0xd0,0x3e,0x31,0x92,0x2c,0x10,0x3c,0x05,0x3c,0x02,
0x6e,0xfc,0x59,0x9f,0x93,0xbd,0x58,0xd3,0xd9,0x75,
0x10,0x01,0x53,0x33,0x8f,0xbe,0x14,0x30,0x56,0x4f,
0x0b,0xa3,0x5c,0x00,0x25,0xdd,0x49,0x01,0x8a,0x5a,
0x1e,0x38,0x8f,0xbe,0xf4,0xa4,0x86,0xa6,0xe5,0x1c,
0xca,0xe8,0xec,0xb2,0x00,0x51,0xc3,0xb4,0x21,0x45,
0x91,0x12,0xe8,0x92,0x00,0x0d,0x93,0x28,0x9c,0x22,
0xd4,0x3a,0x9b,0xb0,0x5f,0x7a,0xb1,0xb0,0x25,0x35,
0x1a,0x01,0x6d,0x1d,0x88,0x76,0xf0,0x9d,0xd5,0x4c,
0xc0,0x7b,0x8e,0x6b,0xde,0xc9,0xe4,0x41,0x80,0x9f,
0xcb,0x79,0xf0,0xbe,0xee,0xc2,0xb9,0x58,0x6b,0xc2,
0x15,0x7c,0x5f,0xef,0xf5,0xd7,0x71,0x04,0xf6,0xb8,
0xfd,0x17,0x60,0x00,0x39,0x6b,0x16,0x6d,0x8c,0xf5,
0x59,0x20,0x00,0x00,0x00,0x00,0x49,0x45,0x4e,0x44,
0xae,0x42,0x60,0x82,

  };

  static FXIcon *VFLL8;
  static unsigned char YQRZ_51 []= {
0x89,0x50,0x4e,0x47,0x0d,0x0a,0x1a,0x0a,0x00,0x00,
0x00,0x0d,0x49,0x48,0x44,0x52,0x00,0x00,0x00,0x20,
0x00,0x00,0x00,0x20,0x08,0x06,0x00,0x00,0x00,0x73,
0x7a,0x7a,0xf4,0x00,0x00,0x00,0x19,0x74,0x45,0x58,
0x74,0x53,0x6f,0x66,0x74,0x77,0x61,0x72,0x65,0x00,
0x41,0x64,0x6f,0x62,0x65,0x20,0x49,0x6d,0x61,0x67,
0x65,0x52,0x65,0x61,0x64,0x79,0x71,0xc9,0x65,0x3c,
0x00,0x00,0x03,0xad,0x69,0x54,0x58,0x74,0x58,0x4d,
0x4c,0x3a,0x63,0x6f,0x6d,0x2e,0x61,0x64,0x6f,0x62,
0x65,0x2e,0x78,0x6d,0x70,0x00,0x00,0x00,0x00,0x00,
0x3c,0x3f,0x78,0x70,0x61,0x63,0x6b,0x65,0x74,0x20,
0x62,0x65,0x67,0x69,0x6e,0x3d,0x22,0xef,0xbb,0xbf,
0x22,0x20,0x69,0x64,0x3d,0x22,0x57,0x35,0x4d,0x30,
0x4d,0x70,0x43,0x65,0x68,0x69,0x48,0x7a,0x72,0x65,
0x53,0x7a,0x4e,0x54,0x63,0x7a,0x6b,0x63,0x39,0x64,
0x22,0x3f,0x3e,0x20,0x3c,0x78,0x3a,0x78,0x6d,0x70,
0x6d,0x65,0x74,0x61,0x20,0x78,0x6d,0x6c,0x6e,0x73,
0x3a,0x78,0x3d,0x22,0x61,0x64,0x6f,0x62,0x65,0x3a,
0x6e,0x73,0x3a,0x6d,0x65,0x74,0x61,0x2f,0x22,0x20,
0x78,0x3a,0x78,0x6d,0x70,0x74,0x6b,0x3d,0x22,0x41,
0x64,0x6f,0x62,0x65,0x20,0x58,0x4d,0x50,0x20,0x43,
0x6f,0x72,0x65,0x20,0x35,0x2e,0x30,0x2d,0x63,0x30,
0x36,0x30,0x20,0x36,0x31,0x2e,0x31,0x33,0x34,0x37,
0x37,0x37,0x2c,0x20,0x32,0x30,0x31,0x30,0x2f,0x30,
0x32,0x2f,0x31,0x32,0x2d,0x31,0x37,0x3a,0x33,0x32,
0x3a,0x30,0x30,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x22,0x3e,0x20,0x3c,0x72,0x64,0x66,0x3a,0x52,
0x44,0x46,0x20,0x78,0x6d,0x6c,0x6e,0x73,0x3a,0x72,
0x64,0x66,0x3d,0x22,0x68,0x74,0x74,0x70,0x3a,0x2f,
0x2f,0x77,0x77,0x77,0x2e,0x77,0x33,0x2e,0x6f,0x72,
0x67,0x2f,0x31,0x39,0x39,0x39,0x2f,0x30,0x32,0x2f,
0x32,0x32,0x2d,0x72,0x64,0x66,0x2d,0x73,0x79,0x6e,
0x74,0x61,0x78,0x2d,0x6e,0x73,0x23,0x22,0x3e,0x20,
0x3c,0x72,0x64,0x66,0x3a,0x44,0x65,0x73,0x63,0x72,
0x69,0x70,0x74,0x69,0x6f,0x6e,0x20,0x72,0x64,0x66,
0x3a,0x61,0x62,0x6f,0x75,0x74,0x3d,0x22,0x22,0x20,
0x78,0x6d,0x6c,0x6e,0x73,0x3a,0x78,0x6d,0x70,0x52,
0x69,0x67,0x68,0x74,0x73,0x3d,0x22,0x68,0x74,0x74,
0x70,0x3a,0x2f,0x2f,0x6e,0x73,0x2e,0x61,0x64,0x6f,
0x62,0x65,0x2e,0x63,0x6f,0x6d,0x2f,0x78,0x61,0x70,
0x2f,0x31,0x2e,0x30,0x2f,0x72,0x69,0x67,0x68,0x74,
0x73,0x2f,0x22,0x20,0x78,0x6d,0x6c,0x6e,0x73,0x3a,
0x78,0x6d,0x70,0x4d,0x4d,0x3d,0x22,0x68,0x74,0x74,
0x70,0x3a,0x2f,0x2f,0x6e,0x73,0x2e,0x61,0x64,0x6f,
0x62,0x65,0x2e,0x63,0x6f,0x6d,0x2f,0x78,0x61,0x70,
0x2f,0x31,0x2e,0x30,0x2f,0x6d,0x6d,0x2f,0x22,0x20,
0x78,0x6d,0x6c,0x6e,0x73,0x3a,0x73,0x74,0x52,0x65,
0x66,0x3d,0x22,0x68,0x74,0x74,0x70,0x3a,0x2f,0x2f,
0x6e,0x73,0x2e,0x61,0x64,0x6f,0x62,0x65,0x2e,0x63,
0x6f,0x6d,0x2f,0x78,0x61,0x70,0x2f,0x31,0x2e,0x30,
0x2f,0x73,0x54,0x79,0x70,0x65,0x2f,0x52,0x65,0x73,
0x6f,0x75,0x72,0x63,0x65,0x52,0x65,0x66,0x23,0x22,
0x20,0x78,0x6d,0x6c,0x6e,0x73,0x3a,0x78,0x6d,0x70,
0x3d,0x22,0x68,0x74,0x74,0x70,0x3a,0x2f,0x2f,0x6e,
0x73,0x2e,0x61,0x64,0x6f,0x62,0x65,0x2e,0x63,0x6f,
0x6d,0x2f,0x78,0x61,0x70,0x2f,0x31,0x2e,0x30,0x2f,
0x22,0x20,0x78,0x6d,0x70,0x52,0x69,0x67,0x68,0x74,
0x73,0x3a,0x4d,0x61,0x72,0x6b,0x65,0x64,0x3d,0x22,
0x46,0x61,0x6c,0x73,0x65,0x22,0x20,0x78,0x6d,0x70,
0x4d,0x4d,0x3a,0x4f,0x72,0x69,0x67,0x69,0x6e,0x61,
0x6c,0x44,0x6f,0x63,0x75,0x6d,0x65,0x6e,0x74,0x49,
0x44,0x3d,0x22,0x75,0x75,0x69,0x64,0x3a,0x41,0x45,
0x33,0x46,0x34,0x36,0x30,0x38,0x38,0x46,0x38,0x39,
0x44,0x45,0x31,0x31,0x41,0x38,0x36,0x30,0x39,0x39,
0x46,0x31,0x32,0x31,0x41,0x32,0x30,0x34,0x32,0x39,
0x22,0x20,0x78,0x6d,0x70,0x4d,0x4d,0x3a,0x44,0x6f,
0x63,0x75,0x6d,0x65,0x6e,0x74,0x49,0x44,0x3d,0x22,
0x78,0x6d,0x70,0x2e,0x64,0x69,0x64,0x3a,0x38,0x46,
0x38,0x30,0x32,0x46,0x38,0x45,0x42,0x35,0x45,0x46,
0x31,0x31,0x45,0x30,0x42,0x41,0x41,0x33,0x42,0x42,
0x45,0x34,0x44,0x37,0x42,0x31,0x33,0x35,0x31,0x44,
0x22,0x20,0x78,0x6d,0x70,0x4d,0x4d,0x3a,0x49,0x6e,
0x73,0x74,0x61,0x6e,0x63,0x65,0x49,0x44,0x3d,0x22,
0x78,0x6d,0x70,0x2e,0x69,0x69,0x64,0x3a,0x38,0x46,
0x32,0x46,0x31,0x46,0x41,0x34,0x42,0x35,0x45,0x46,
0x31,0x31,0x45,0x30,0x42,0x41,0x41,0x33,0x42,0x42,
0x45,0x34,0x44,0x37,0x42,0x31,0x33,0x35,0x31,0x44,
0x22,0x20,0x78,0x6d,0x70,0x3a,0x43,0x72,0x65,0x61,
0x74,0x6f,0x72,0x54,0x6f,0x6f,0x6c,0x3d,0x22,0x41,
0x64,0x6f,0x62,0x65,0x20,0x50,0x68,0x6f,0x74,0x6f,
0x73,0x68,0x6f,0x70,0x20,0x43,0x53,0x35,0x20,0x57,
0x69,0x6e,0x64,0x6f,0x77,0x73,0x22,0x3e,0x20,0x3c,
0x78,0x6d,0x70,0x4d,0x4d,0x3a,0x44,0x65,0x72,0x69,
0x76,0x65,0x64,0x46,0x72,0x6f,0x6d,0x20,0x73,0x74,
0x52,0x65,0x66,0x3a,0x69,0x6e,0x73,0x74,0x61,0x6e,
0x63,0x65,0x49,0x44,0x3d,0x22,0x78,0x6d,0x70,0x2e,
0x69,0x69,0x64,0x3a,0x33,0x32,0x45,0x42,0x39,0x34,
0x34,0x44,0x45,0x46,0x42,0x35,0x45,0x30,0x31,0x31,
0x39,0x41,0x42,0x34,0x41,0x46,0x44,0x39,0x30,0x33,
0x39,0x46,0x33,0x35,0x45,0x42,0x22,0x20,0x73,0x74,
0x52,0x65,0x66,0x3a,0x64,0x6f,0x63,0x75,0x6d,0x65,
0x6e,0x74,0x49,0x44,0x3d,0x22,0x75,0x75,0x69,0x64,
0x3a,0x41,0x45,0x33,0x46,0x34,0x36,0x30,0x38,0x38,
0x46,0x38,0x39,0x44,0x45,0x31,0x31,0x41,0x38,0x36,
0x30,0x39,0x39,0x46,0x31,0x32,0x31,0x41,0x32,0x30,
0x34,0x32,0x39,0x22,0x2f,0x3e,0x20,0x3c,0x2f,0x72,
0x64,0x66,0x3a,0x44,0x65,0x73,0x63,0x72,0x69,0x70,
0x74,0x69,0x6f,0x6e,0x3e,0x20,0x3c,0x2f,0x72,0x64,
0x66,0x3a,0x52,0x44,0x46,0x3e,0x20,0x3c,0x2f,0x78,
0x3a,0x78,0x6d,0x70,0x6d,0x65,0x74,0x61,0x3e,0x20,
0x3c,0x3f,0x78,0x70,0x61,0x63,0x6b,0x65,0x74,0x20,
0x65,0x6e,0x64,0x3d,0x22,0x72,0x22,0x3f,0x3e,0xd1,
0x1f,0xef,0x65,0x00,0x00,0x00,0xe8,0x49,0x44,0x41,
0x54,0x78,0xda,0xec,0x97,0x41,0x12,0x82,0x30,0x0c,
0x45,0x89,0xe3,0x8d,0x3c,0x87,0x6e,0xbc,0x83,0x5e,
0x83,0x6b,0xe8,0x1d,0x5c,0x79,0x0e,0x4f,0x24,0x4d,
0xa4,0x16,0x18,0x40,0x28,0xa5,0x9a,0xfc,0x8d,0x99,
0x29,0x1b,0x60,0xf2,0xf8,0xff,0x97,0x4c,0x49,0x44,
0x0a,0x64,0x6d,0x0a,0x70,0xc1,0x01,0xb6,0xfe,0x52,
0xde,0x1e,0x6b,0xdf,0xf3,0xbe,0xdd,0xeb,0x75,0xc8,
0x6d,0x5c,0x1e,0x77,0xd9,0x0a,0xb4,0xa1,0xd9,0x23,
0x2c,0x10,0x15,0x0b,0x12,0x77,0x82,0xa8,0x65,0x20,
0x01,0x40,0x54,0x43,0xc8,0x82,0x69,0xde,0x65,0xc0,
0x2b,0x30,0xb3,0x24,0x72,0xaf,0xbd,0x9f,0x53,0x97,
0x91,0x02,0x62,0xfd,0xe5,0xa7,0x7a,0x9d,0x63,0x00,
0x62,0xfa,0x23,0x62,0x66,0x48,0xf3,0x29,0x05,0xcc,
0x27,0x53,0xa3,0x80,0x40,0x9a,0xf7,0x01,0x60,0x33,
0xb9,0xb1,0x80,0xb1,0xd3,0xb0,0x56,0x80,0xa0,0x16,
0xb8,0x90,0x01,0x02,0x86,0xb0,0xb3,0x80,0x30,0xdb,
0x70,0x38,0x0c,0x4c,0x21,0x82,0x05,0x9f,0x21,0x34,
0x83,0x78,0x03,0x54,0x6e,0x72,0x17,0x68,0x42,0x5c,
0x07,0x00,0xcf,0xca,0xcd,0x3d,0xb8,0x04,0x41,0x3f,
0x51,0x20,0x02,0xa0,0x6e,0x47,0x0a,0x80,0x2a,0x44,
0xc8,0xc0,0x32,0x80,0x1a,0x44,0x2c,0x84,0x26,0x10,
0x01,0x80,0x57,0xcd,0x02,0xea,0x1d,0x4c,0xbe,0x2e,
0xfa,0x1f,0x4e,0xd1,0x00,0x2f,0x01,0x06,0x00,0xcf,
0x13,0xcb,0x68,0x81,0x12,0x84,0xaa,0x00,0x00,0x00,
0x00,0x49,0x45,0x4e,0x44,0xae,0x42,0x60,0x82,

  };

  static FXIcon *VFLL9;
  static unsigned char YQRZ_57 []= {
0x89,0x50,0x4e,0x47,0x0d,0x0a,0x1a,0x0a,0x00,0x00,
0x00,0x0d,0x49,0x48,0x44,0x52,0x00,0x00,0x00,0x20,
0x00,0x00,0x00,0x20,0x08,0x06,0x00,0x00,0x00,0x73,
0x7a,0x7a,0xf4,0x00,0x00,0x00,0x19,0x74,0x45,0x58,
0x74,0x53,0x6f,0x66,0x74,0x77,0x61,0x72,0x65,0x00,
0x41,0x64,0x6f,0x62,0x65,0x20,0x49,0x6d,0x61,0x67,
0x65,0x52,0x65,0x61,0x64,0x79,0x71,0xc9,0x65,0x3c,
0x00,0x00,0x03,0xad,0x69,0x54,0x58,0x74,0x58,0x4d,
0x4c,0x3a,0x63,0x6f,0x6d,0x2e,0x61,0x64,0x6f,0x62,
0x65,0x2e,0x78,0x6d,0x70,0x00,0x00,0x00,0x00,0x00,
0x3c,0x3f,0x78,0x70,0x61,0x63,0x6b,0x65,0x74,0x20,
0x62,0x65,0x67,0x69,0x6e,0x3d,0x22,0xef,0xbb,0xbf,
0x22,0x20,0x69,0x64,0x3d,0x22,0x57,0x35,0x4d,0x30,
0x4d,0x70,0x43,0x65,0x68,0x69,0x48,0x7a,0x72,0x65,
0x53,0x7a,0x4e,0x54,0x63,0x7a,0x6b,0x63,0x39,0x64,
0x22,0x3f,0x3e,0x20,0x3c,0x78,0x3a,0x78,0x6d,0x70,
0x6d,0x65,0x74,0x61,0x20,0x78,0x6d,0x6c,0x6e,0x73,
0x3a,0x78,0x3d,0x22,0x61,0x64,0x6f,0x62,0x65,0x3a,
0x6e,0x73,0x3a,0x6d,0x65,0x74,0x61,0x2f,0x22,0x20,
0x78,0x3a,0x78,0x6d,0x70,0x74,0x6b,0x3d,0x22,0x41,
0x64,0x6f,0x62,0x65,0x20,0x58,0x4d,0x50,0x20,0x43,
0x6f,0x72,0x65,0x20,0x35,0x2e,0x30,0x2d,0x63,0x30,
0x36,0x30,0x20,0x36,0x31,0x2e,0x31,0x33,0x34,0x37,
0x37,0x37,0x2c,0x20,0x32,0x30,0x31,0x30,0x2f,0x30,
0x32,0x2f,0x31,0x32,0x2d,0x31,0x37,0x3a,0x33,0x32,
0x3a,0x30,0x30,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x22,0x3e,0x20,0x3c,0x72,0x64,0x66,0x3a,0x52,
0x44,0x46,0x20,0x78,0x6d,0x6c,0x6e,0x73,0x3a,0x72,
0x64,0x66,0x3d,0x22,0x68,0x74,0x74,0x70,0x3a,0x2f,
0x2f,0x77,0x77,0x77,0x2e,0x77,0x33,0x2e,0x6f,0x72,
0x67,0x2f,0x31,0x39,0x39,0x39,0x2f,0x30,0x32,0x2f,
0x32,0x32,0x2d,0x72,0x64,0x66,0x2d,0x73,0x79,0x6e,
0x74,0x61,0x78,0x2d,0x6e,0x73,0x23,0x22,0x3e,0x20,
0x3c,0x72,0x64,0x66,0x3a,0x44,0x65,0x73,0x63,0x72,
0x69,0x70,0x74,0x69,0x6f,0x6e,0x20,0x72,0x64,0x66,
0x3a,0x61,0x62,0x6f,0x75,0x74,0x3d,0x22,0x22,0x20,
0x78,0x6d,0x6c,0x6e,0x73,0x3a,0x78,0x6d,0x70,0x52,
0x69,0x67,0x68,0x74,0x73,0x3d,0x22,0x68,0x74,0x74,
0x70,0x3a,0x2f,0x2f,0x6e,0x73,0x2e,0x61,0x64,0x6f,
0x62,0x65,0x2e,0x63,0x6f,0x6d,0x2f,0x78,0x61,0x70,
0x2f,0x31,0x2e,0x30,0x2f,0x72,0x69,0x67,0x68,0x74,
0x73,0x2f,0x22,0x20,0x78,0x6d,0x6c,0x6e,0x73,0x3a,
0x78,0x6d,0x70,0x4d,0x4d,0x3d,0x22,0x68,0x74,0x74,
0x70,0x3a,0x2f,0x2f,0x6e,0x73,0x2e,0x61,0x64,0x6f,
0x62,0x65,0x2e,0x63,0x6f,0x6d,0x2f,0x78,0x61,0x70,
0x2f,0x31,0x2e,0x30,0x2f,0x6d,0x6d,0x2f,0x22,0x20,
0x78,0x6d,0x6c,0x6e,0x73,0x3a,0x73,0x74,0x52,0x65,
0x66,0x3d,0x22,0x68,0x74,0x74,0x70,0x3a,0x2f,0x2f,
0x6e,0x73,0x2e,0x61,0x64,0x6f,0x62,0x65,0x2e,0x63,
0x6f,0x6d,0x2f,0x78,0x61,0x70,0x2f,0x31,0x2e,0x30,
0x2f,0x73,0x54,0x79,0x70,0x65,0x2f,0x52,0x65,0x73,
0x6f,0x75,0x72,0x63,0x65,0x52,0x65,0x66,0x23,0x22,
0x20,0x78,0x6d,0x6c,0x6e,0x73,0x3a,0x78,0x6d,0x70,
0x3d,0x22,0x68,0x74,0x74,0x70,0x3a,0x2f,0x2f,0x6e,
0x73,0x2e,0x61,0x64,0x6f,0x62,0x65,0x2e,0x63,0x6f,
0x6d,0x2f,0x78,0x61,0x70,0x2f,0x31,0x2e,0x30,0x2f,
0x22,0x20,0x78,0x6d,0x70,0x52,0x69,0x67,0x68,0x74,
0x73,0x3a,0x4d,0x61,0x72,0x6b,0x65,0x64,0x3d,0x22,
0x46,0x61,0x6c,0x73,0x65,0x22,0x20,0x78,0x6d,0x70,
0x4d,0x4d,0x3a,0x4f,0x72,0x69,0x67,0x69,0x6e,0x61,
0x6c,0x44,0x6f,0x63,0x75,0x6d,0x65,0x6e,0x74,0x49,
0x44,0x3d,0x22,0x75,0x75,0x69,0x64,0x3a,0x41,0x45,
0x33,0x46,0x34,0x36,0x30,0x38,0x38,0x46,0x38,0x39,
0x44,0x45,0x31,0x31,0x41,0x38,0x36,0x30,0x39,0x39,
0x46,0x31,0x32,0x31,0x41,0x32,0x30,0x34,0x32,0x39,
0x22,0x20,0x78,0x6d,0x70,0x4d,0x4d,0x3a,0x44,0x6f,
0x63,0x75,0x6d,0x65,0x6e,0x74,0x49,0x44,0x3d,0x22,
0x78,0x6d,0x70,0x2e,0x64,0x69,0x64,0x3a,0x39,0x31,
0x39,0x36,0x33,0x45,0x35,0x30,0x42,0x35,0x45,0x46,
0x31,0x31,0x45,0x30,0x42,0x41,0x41,0x33,0x42,0x42,
0x45,0x34,0x44,0x37,0x42,0x31,0x33,0x35,0x31,0x44,
0x22,0x20,0x78,0x6d,0x70,0x4d,0x4d,0x3a,0x49,0x6e,
0x73,0x74,0x61,0x6e,0x63,0x65,0x49,0x44,0x3d,0x22,
0x78,0x6d,0x70,0x2e,0x69,0x69,0x64,0x3a,0x39,0x31,
0x39,0x36,0x33,0x45,0x34,0x46,0x42,0x35,0x45,0x46,
0x31,0x31,0x45,0x30,0x42,0x41,0x41,0x33,0x42,0x42,
0x45,0x34,0x44,0x37,0x42,0x31,0x33,0x35,0x31,0x44,
0x22,0x20,0x78,0x6d,0x70,0x3a,0x43,0x72,0x65,0x61,
0x74,0x6f,0x72,0x54,0x6f,0x6f,0x6c,0x3d,0x22,0x41,
0x64,0x6f,0x62,0x65,0x20,0x50,0x68,0x6f,0x74,0x6f,
0x73,0x68,0x6f,0x70,0x20,0x43,0x53,0x35,0x20,0x57,
0x69,0x6e,0x64,0x6f,0x77,0x73,0x22,0x3e,0x20,0x3c,
0x78,0x6d,0x70,0x4d,0x4d,0x3a,0x44,0x65,0x72,0x69,
0x76,0x65,0x64,0x46,0x72,0x6f,0x6d,0x20,0x73,0x74,
0x52,0x65,0x66,0x3a,0x69,0x6e,0x73,0x74,0x61,0x6e,
0x63,0x65,0x49,0x44,0x3d,0x22,0x78,0x6d,0x70,0x2e,
0x69,0x69,0x64,0x3a,0x33,0x32,0x45,0x42,0x39,0x34,
0x34,0x44,0x45,0x46,0x42,0x35,0x45,0x30,0x31,0x31,
0x39,0x41,0x42,0x34,0x41,0x46,0x44,0x39,0x30,0x33,
0x39,0x46,0x33,0x35,0x45,0x42,0x22,0x20,0x73,0x74,
0x52,0x65,0x66,0x3a,0x64,0x6f,0x63,0x75,0x6d,0x65,
0x6e,0x74,0x49,0x44,0x3d,0x22,0x75,0x75,0x69,0x64,
0x3a,0x41,0x45,0x33,0x46,0x34,0x36,0x30,0x38,0x38,
0x46,0x38,0x39,0x44,0x45,0x31,0x31,0x41,0x38,0x36,
0x30,0x39,0x39,0x46,0x31,0x32,0x31,0x41,0x32,0x30,
0x34,0x32,0x39,0x22,0x2f,0x3e,0x20,0x3c,0x2f,0x72,
0x64,0x66,0x3a,0x44,0x65,0x73,0x63,0x72,0x69,0x70,
0x74,0x69,0x6f,0x6e,0x3e,0x20,0x3c,0x2f,0x72,0x64,
0x66,0x3a,0x52,0x44,0x46,0x3e,0x20,0x3c,0x2f,0x78,
0x3a,0x78,0x6d,0x70,0x6d,0x65,0x74,0x61,0x3e,0x20,
0x3c,0x3f,0x78,0x70,0x61,0x63,0x6b,0x65,0x74,0x20,
0x65,0x6e,0x64,0x3d,0x22,0x72,0x22,0x3f,0x3e,0x0e,
0x57,0x54,0x67,0x00,0x00,0x01,0x19,0x49,0x44,0x41,
0x54,0x78,0xda,0xc4,0x97,0x51,0x0e,0x83,0x20,0x10,
0x44,0xc7,0xc6,0x73,0xf5,0x12,0x3d,0x50,0xf1,0xbb,
0xe9,0xed,0x7a,0x9a,0x46,0x9d,0x82,0xa0,0x85,0x82,
0x81,0xfe,0x0c,0x24,0xba,0x08,0x6b,0x1e,0xee,0x0e,
0xab,0x0e,0x24,0xd1,0xb3,0x5d,0xd0,0xb9,0x8d,0xee,
0xf4,0x7c,0xdd,0x40,0xae,0x20,0x16,0x80,0xcb,0x66,
0x79,0xd8,0xd9,0x7a,0x84,0x6b,0xce,0x77,0x3b,0x66,
0xdc,0xd8,0x3e,0xb7,0xfb,0x6d,0xf7,0xd1,0xce,0x61,
0x99,0xbc,0xef,0x1a,0x7c,0x56,0x3b,0xe7,0xfa,0xf4,
0x63,0xe4,0x76,0x3c,0xae,0xef,0x6f,0x04,0x1a,0xe1,
0xa8,0xc0,0xc3,0x7c,0x1d,0x6e,0x07,0xd3,0x14,0x34,
0xc2,0x51,0x81,0xa3,0x15,0x1e,0xeb,0xce,0x6b,0x40,
0x0c,0xcf,0x16,0xa0,0x86,0xe7,0x0b,0x10,0xc3,0xe3,
0x9d,0x1f,0x45,0x40,0x07,0x2f,0x44,0x40,0x0b,0xcf,
0x22,0xd0,0x0b,0x9e,0x6a,0x40,0x08,0xcf,0x35,0xe0,
0x2b,0x9c,0xf5,0x99,0x79,0x58,0x2e,0x5b,0x1f,0xce,
0x86,0x7e,0x23,0xfc,0xa7,0xe1,0xb0,0xc0,0xe1,0x76,
0xff,0x15,0xa1,0x51,0x3d,0x79,0x38,0x4c,0x26,0x42,
0x21,0xfc,0x7c,0x1b,0xaa,0xe0,0x25,0x0d,0x68,0xe1,
0x44,0xb9,0x12,0xaa,0xe0,0x04,0x4e,0xde,0x86,0x22,
0x78,0x5e,0x88,0xd4,0xf0,0x42,0x0a,0x8c,0x12,0x1e,
0x6f,0xc3,0x31,0xa4,0xc0,0x7d,0x46,0x4d,0x8d,0x45,
0xa6,0x06,0x1f,0x2a,0xf0,0x82,0x06,0x74,0x4f,0xee,
0xe1,0x79,0x0a,0xb4,0xf0,0x93,0x42,0xa4,0x87,0x27,
0x11,0x50,0xc2,0x0b,0xdb,0xb0,0x0f,0x3c,0x4a,0x41,
0x1f,0x78,0x2a,0xc2,0x0e,0xf0,0x68,0x01,0x7d,0xe0,
0xc9,0x02,0x1a,0xe1,0xa6,0x56,0xe1,0xfe,0x81,0xbb,
0x36,0xf4,0xfe,0x3d,0xff,0x08,0x30,0x00,0xa4,0x5d,
0x54,0x49,0x02,0xaa,0x2d,0x83,0x00,0x00,0x00,0x00,
0x49,0x45,0x4e,0x44,0xae,0x42,0x60,0x82,

  };

  static FXIcon *VFLL10;
  static unsigned char YQRZ_63 []= {
0x89,0x50,0x4e,0x47,0x0d,0x0a,0x1a,0x0a,0x00,0x00,
0x00,0x0d,0x49,0x48,0x44,0x52,0x00,0x00,0x00,0x20,
0x00,0x00,0x00,0x20,0x08,0x06,0x00,0x00,0x00,0x73,
0x7a,0x7a,0xf4,0x00,0x00,0x00,0x06,0x62,0x4b,0x47,
0x44,0x00,0xff,0x00,0xff,0x00,0xff,0xa0,0xbd,0xa7,
0x93,0x00,0x00,0x00,0x09,0x70,0x48,0x59,0x73,0x00,
0x00,0x0b,0x13,0x00,0x00,0x0b,0x13,0x01,0x00,0x9a,
0x9c,0x18,0x00,0x00,0x00,0x07,0x74,0x49,0x4d,0x45,
0x07,0xdd,0x08,0x0a,0x15,0x07,0x20,0x53,0x7b,0xd9,
0x60,0x00,0x00,0x02,0x6b,0x49,0x44,0x41,0x54,0x58,
0xc3,0xed,0x57,0x41,0x6b,0x13,0x41,0x18,0x7d,0x23,
0x1e,0x55,0x7a,0xf0,0x20,0x88,0x7a,0xd4,0x9f,0xe0,
0x21,0xd8,0xdd,0x36,0xc9,0x6e,0x09,0x82,0x14,0x7a,
0xf1,0x07,0x24,0x17,0x25,0xb4,0x86,0x52,0xc1,0xe6,
0x60,0x05,0x41,0x9a,0xd0,0xd4,0xf6,0xb0,0xfb,0x03,
0xbc,0x14,0x8a,0x20,0xed,0xc4,0xcd,0x6e,0x02,0xea,
0xc1,0x9f,0xa0,0x47,0x15,0xc1,0x83,0x87,0x22,0xbd,
0xb9,0x99,0xcf,0x43,0x9a,0x25,0xd9,0xcc,0x64,0x67,
0x63,0x24,0x17,0x07,0x1e,0xcc,0x66,0xf7,0x7b,0x79,
0xfb,0xbe,0xef,0x9b,0x99,0x65,0x44,0x84,0x59,0x8e,
0x73,0x98,0xf1,0x98,0xb9,0x80,0xf3,0x43,0x57,0xe5,
0x5b,0xd1,0x74,0x61,0xc9,0x71,0x00,0x14,0x01,0xb8,
0x6d,0x5e,0x2a,0x4d,0x42,0xae,0xe4,0x68,0x7c,0x8a,
0xa6,0x6c,0xb0,0x06,0xe8,0xe1,0xcd,0x5e,0x60,0xc1,
0x19,0x29,0x8c,0xf6,0x51,0x89,0xa5,0xfa,0xf3,0x82,
0xf3,0x1e,0x40,0x46,0xc6,0xc1,0x5e,0x7e,0x96,0xa7,
0x40,0x74,0x09,0x86,0xb5,0x4f,0x22,0x0c,0x11,0x87,
0x61,0xed,0x93,0xe8,0x12,0x74,0x70,0xc6,0x91,0x51,
0x71,0x0c,0x8e,0x21,0x07,0xee,0xbc,0x6e,0xf6,0x2d,
0x53,0x8e,0x80,0x3f,0x18,0xeb,0xc4,0xe2,0xd2,0x5e,
0x52,0x5b,0xb9,0xef,0xee,0xd9,0x25,0xa9,0x80,0xcc,
0xc1,0x1b,0xad,0x9e,0x0c,0x9a,0x65,0xa9,0x88,0x45,
0xbb,0xa1,0x15,0xff,0x61,0xe5,0x2e,0x93,0xa7,0x20,
0x0c,0x5d,0x99,0xfd,0x71,0x98,0xd9,0x1a,0x89,0xdf,
0x84,0x41,0x98,0xd9,0x1a,0xe9,0xc4,0x8a,0x30,0x74,
0x95,0x29,0x38,0xbd,0x7f,0x03,0x39,0xeb,0x85,0xf6,
0xca,0xd4,0x7a,0xbb,0xce,0x00,0xa4,0x8e,0xb9,0xf0,
0xea,0x8b,0x5c,0xc0,0xaf,0x95,0x6b,0x00,0x80,0x9c,
0xfd,0xfc,0x9f,0x2c,0x8f,0xad,0xe6,0x63,0x06,0x00,
0x97,0x0e,0xbe,0xc9,0x05,0x9c,0x2c,0x5f,0x8d,0xe6,
0xd9,0xec,0xd3,0xa9,0x8a,0xf0,0xfd,0x6a,0x94,0xf7,
0xb9,0xc3,0xef,0xf2,0x36,0x24,0x41,0x11,0x5a,0xde,
0x26,0x83,0x10,0x98,0x06,0x5a,0xde,0x26,0x1b,0xe4,
0x56,0x2e,0xc5,0x83,0x0f,0x91,0x20,0x78,0x5e,0x95,
0x11,0x11,0xfe,0x06,0x9e,0x57,0x65,0x71,0x5e,0x65,
0x11,0xfe,0x2c,0x5c,0x91,0xda,0x97,0x5b,0x78,0x32,
0x51,0x3a,0x5a,0xed,0x67,0xd2,0x76,0xbd,0x7c,0xf4,
0x43,0xcf,0x81,0xc8,0x09,0x7f,0x8b,0x81,0x04,0xd2,
0xc0,0xf3,0xb7,0x98,0x8a,0x4f,0xb9,0x19,0xc5,0x6f,
0x0e,0x0d,0x91,0xce,0x04,0xd2,0x7c,0x3e,0xe6,0x80,
0x1c,0x79,0x63,0x83,0xa8,0xdb,0x45,0x1a,0xe4,0x8d,
0x0d,0x52,0xf1,0xa9,0x05,0x48,0x8a,0x28,0x6f,0xae,
0x13,0x91,0xc0,0x24,0xe8,0xc5,0x8e,0x72,0x6a,0x77,
0x41,0x7e,0xbe,0x92,0xfa,0xcd,0x47,0x9c,0x98,0xaf,
0xd0,0xb8,0x1a,0x50,0x0a,0xb0,0x8c,0x0a,0x81,0x08,
0xd3,0x80,0x65,0x0c,0x8b,0x48,0x14,0x60,0x19,0x8f,
0x88,0x84,0xc0,0x34,0xd1,0xe3,0x4c,0x12,0x40,0x80,
0x6d,0xae,0x91,0x6e,0xab,0xf1,0x60,0x9b,0xf1,0x60,
0x5b,0xbb,0x45,0x6d,0x73,0x6d,0xe4,0x10,0x3e,0x24,
0xc0,0x36,0x56,0x1d,0x55,0xef,0xc6,0xc1,0xfd,0x1a,
0x83,0x20,0xe0,0x6c,0xae,0x1b,0x67,0x1b,0xab,0x8e,
0xfa,0x54,0x4c,0x54,0xd4,0xc9,0x29,0x0f,0xea,0x2c,
0xfe,0x33,0x0f,0xea,0x4c,0xb3,0x26,0x8a,0x63,0x6a,
0x40,0xb8,0x49,0xb9,0x3c,0x0e,0xea,0xca,0xfd,0xe1,
0x38,0xa8,0x33,0x8d,0x7a,0x50,0x1f,0x48,0xbe,0xde,
0x9e,0x83,0x6d,0x96,0x95,0x4b,0x58,0xb3,0xd3,0xd0,
0x3a,0x19,0x27,0x71,0x5c,0xff,0x78,0xa2,0x5e,0x88,
0x78,0x7b,0x47,0x56,0x54,0xa7,0xbc,0xbd,0xa3,0xbd,
0x33,0x2a,0x38,0xd0,0xe7,0x48,0xfc,0x32,0xe2,0x9d,
0x5d,0x06,0xa0,0x6f,0x95,0xcb,0x3b,0xbb,0x17,0xd3,
0xee,0x84,0x12,0x0e,0xa9,0x7b,0xec,0xff,0xc7,0xe9,
0xac,0x05,0xfc,0x01,0x6d,0xe9,0x77,0x84,0x3c,0xb4,
0x16,0x35,0x00,0x00,0x00,0x00,0x49,0x45,0x4e,0x44,
0xae,0x42,0x60,0x82,

  };

  static FXIcon *VFLL11;
  static unsigned char YQRZ_64 []= {
0x89,0x50,0x4e,0x47,0x0d,0x0a,0x1a,0x0a,0x00,0x00,
0x00,0x0d,0x49,0x48,0x44,0x52,0x00,0x00,0x00,0x20,
0x00,0x00,0x00,0x20,0x08,0x06,0x00,0x00,0x00,0x73,
0x7a,0x7a,0xf4,0x00,0x00,0x00,0x19,0x74,0x45,0x58,
0x74,0x53,0x6f,0x66,0x74,0x77,0x61,0x72,0x65,0x00,
0x41,0x64,0x6f,0x62,0x65,0x20,0x49,0x6d,0x61,0x67,
0x65,0x52,0x65,0x61,0x64,0x79,0x71,0xc9,0x65,0x3c,
0x00,0x00,0x03,0xad,0x69,0x54,0x58,0x74,0x58,0x4d,
0x4c,0x3a,0x63,0x6f,0x6d,0x2e,0x61,0x64,0x6f,0x62,
0x65,0x2e,0x78,0x6d,0x70,0x00,0x00,0x00,0x00,0x00,
0x3c,0x3f,0x78,0x70,0x61,0x63,0x6b,0x65,0x74,0x20,
0x62,0x65,0x67,0x69,0x6e,0x3d,0x22,0xef,0xbb,0xbf,
0x22,0x20,0x69,0x64,0x3d,0x22,0x57,0x35,0x4d,0x30,
0x4d,0x70,0x43,0x65,0x68,0x69,0x48,0x7a,0x72,0x65,
0x53,0x7a,0x4e,0x54,0x63,0x7a,0x6b,0x63,0x39,0x64,
0x22,0x3f,0x3e,0x20,0x3c,0x78,0x3a,0x78,0x6d,0x70,
0x6d,0x65,0x74,0x61,0x20,0x78,0x6d,0x6c,0x6e,0x73,
0x3a,0x78,0x3d,0x22,0x61,0x64,0x6f,0x62,0x65,0x3a,
0x6e,0x73,0x3a,0x6d,0x65,0x74,0x61,0x2f,0x22,0x20,
0x78,0x3a,0x78,0x6d,0x70,0x74,0x6b,0x3d,0x22,0x41,
0x64,0x6f,0x62,0x65,0x20,0x58,0x4d,0x50,0x20,0x43,
0x6f,0x72,0x65,0x20,0x35,0x2e,0x30,0x2d,0x63,0x30,
0x36,0x30,0x20,0x36,0x31,0x2e,0x31,0x33,0x34,0x37,
0x37,0x37,0x2c,0x20,0x32,0x30,0x31,0x30,0x2f,0x30,
0x32,0x2f,0x31,0x32,0x2d,0x31,0x37,0x3a,0x33,0x32,
0x3a,0x30,0x30,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x22,0x3e,0x20,0x3c,0x72,0x64,0x66,0x3a,0x52,
0x44,0x46,0x20,0x78,0x6d,0x6c,0x6e,0x73,0x3a,0x72,
0x64,0x66,0x3d,0x22,0x68,0x74,0x74,0x70,0x3a,0x2f,
0x2f,0x77,0x77,0x77,0x2e,0x77,0x33,0x2e,0x6f,0x72,
0x67,0x2f,0x31,0x39,0x39,0x39,0x2f,0x30,0x32,0x2f,
0x32,0x32,0x2d,0x72,0x64,0x66,0x2d,0x73,0x79,0x6e,
0x74,0x61,0x78,0x2d,0x6e,0x73,0x23,0x22,0x3e,0x20,
0x3c,0x72,0x64,0x66,0x3a,0x44,0x65,0x73,0x63,0x72,
0x69,0x70,0x74,0x69,0x6f,0x6e,0x20,0x72,0x64,0x66,
0x3a,0x61,0x62,0x6f,0x75,0x74,0x3d,0x22,0x22,0x20,
0x78,0x6d,0x6c,0x6e,0x73,0x3a,0x78,0x6d,0x70,0x52,
0x69,0x67,0x68,0x74,0x73,0x3d,0x22,0x68,0x74,0x74,
0x70,0x3a,0x2f,0x2f,0x6e,0x73,0x2e,0x61,0x64,0x6f,
0x62,0x65,0x2e,0x63,0x6f,0x6d,0x2f,0x78,0x61,0x70,
0x2f,0x31,0x2e,0x30,0x2f,0x72,0x69,0x67,0x68,0x74,
0x73,0x2f,0x22,0x20,0x78,0x6d,0x6c,0x6e,0x73,0x3a,
0x78,0x6d,0x70,0x4d,0x4d,0x3d,0x22,0x68,0x74,0x74,
0x70,0x3a,0x2f,0x2f,0x6e,0x73,0x2e,0x61,0x64,0x6f,
0x62,0x65,0x2e,0x63,0x6f,0x6d,0x2f,0x78,0x61,0x70,
0x2f,0x31,0x2e,0x30,0x2f,0x6d,0x6d,0x2f,0x22,0x20,
0x78,0x6d,0x6c,0x6e,0x73,0x3a,0x73,0x74,0x52,0x65,
0x66,0x3d,0x22,0x68,0x74,0x74,0x70,0x3a,0x2f,0x2f,
0x6e,0x73,0x2e,0x61,0x64,0x6f,0x62,0x65,0x2e,0x63,
0x6f,0x6d,0x2f,0x78,0x61,0x70,0x2f,0x31,0x2e,0x30,
0x2f,0x73,0x54,0x79,0x70,0x65,0x2f,0x52,0x65,0x73,
0x6f,0x75,0x72,0x63,0x65,0x52,0x65,0x66,0x23,0x22,
0x20,0x78,0x6d,0x6c,0x6e,0x73,0x3a,0x78,0x6d,0x70,
0x3d,0x22,0x68,0x74,0x74,0x70,0x3a,0x2f,0x2f,0x6e,
0x73,0x2e,0x61,0x64,0x6f,0x62,0x65,0x2e,0x63,0x6f,
0x6d,0x2f,0x78,0x61,0x70,0x2f,0x31,0x2e,0x30,0x2f,
0x22,0x20,0x78,0x6d,0x70,0x52,0x69,0x67,0x68,0x74,
0x73,0x3a,0x4d,0x61,0x72,0x6b,0x65,0x64,0x3d,0x22,
0x46,0x61,0x6c,0x73,0x65,0x22,0x20,0x78,0x6d,0x70,
0x4d,0x4d,0x3a,0x4f,0x72,0x69,0x67,0x69,0x6e,0x61,
0x6c,0x44,0x6f,0x63,0x75,0x6d,0x65,0x6e,0x74,0x49,
0x44,0x3d,0x22,0x75,0x75,0x69,0x64,0x3a,0x41,0x45,
0x33,0x46,0x34,0x36,0x30,0x38,0x38,0x46,0x38,0x39,
0x44,0x45,0x31,0x31,0x41,0x38,0x36,0x30,0x39,0x39,
0x46,0x31,0x32,0x31,0x41,0x32,0x30,0x34,0x32,0x39,
0x22,0x20,0x78,0x6d,0x70,0x4d,0x4d,0x3a,0x44,0x6f,
0x63,0x75,0x6d,0x65,0x6e,0x74,0x49,0x44,0x3d,0x22,
0x78,0x6d,0x70,0x2e,0x64,0x69,0x64,0x3a,0x39,0x31,
0x39,0x36,0x33,0x45,0x35,0x38,0x42,0x35,0x45,0x46,
0x31,0x31,0x45,0x30,0x42,0x41,0x41,0x33,0x42,0x42,
0x45,0x34,0x44,0x37,0x42,0x31,0x33,0x35,0x31,0x44,
0x22,0x20,0x78,0x6d,0x70,0x4d,0x4d,0x3a,0x49,0x6e,
0x73,0x74,0x61,0x6e,0x63,0x65,0x49,0x44,0x3d,0x22,
0x78,0x6d,0x70,0x2e,0x69,0x69,0x64,0x3a,0x39,0x31,
0x39,0x36,0x33,0x45,0x35,0x37,0x42,0x35,0x45,0x46,
0x31,0x31,0x45,0x30,0x42,0x41,0x41,0x33,0x42,0x42,
0x45,0x34,0x44,0x37,0x42,0x31,0x33,0x35,0x31,0x44,
0x22,0x20,0x78,0x6d,0x70,0x3a,0x43,0x72,0x65,0x61,
0x74,0x6f,0x72,0x54,0x6f,0x6f,0x6c,0x3d,0x22,0x41,
0x64,0x6f,0x62,0x65,0x20,0x50,0x68,0x6f,0x74,0x6f,
0x73,0x68,0x6f,0x70,0x20,0x43,0x53,0x35,0x20,0x57,
0x69,0x6e,0x64,0x6f,0x77,0x73,0x22,0x3e,0x20,0x3c,
0x78,0x6d,0x70,0x4d,0x4d,0x3a,0x44,0x65,0x72,0x69,
0x76,0x65,0x64,0x46,0x72,0x6f,0x6d,0x20,0x73,0x74,
0x52,0x65,0x66,0x3a,0x69,0x6e,0x73,0x74,0x61,0x6e,
0x63,0x65,0x49,0x44,0x3d,0x22,0x78,0x6d,0x70,0x2e,
0x69,0x69,0x64,0x3a,0x33,0x32,0x45,0x42,0x39,0x34,
0x34,0x44,0x45,0x46,0x42,0x35,0x45,0x30,0x31,0x31,
0x39,0x41,0x42,0x34,0x41,0x46,0x44,0x39,0x30,0x33,
0x39,0x46,0x33,0x35,0x45,0x42,0x22,0x20,0x73,0x74,
0x52,0x65,0x66,0x3a,0x64,0x6f,0x63,0x75,0x6d,0x65,
0x6e,0x74,0x49,0x44,0x3d,0x22,0x75,0x75,0x69,0x64,
0x3a,0x41,0x45,0x33,0x46,0x34,0x36,0x30,0x38,0x38,
0x46,0x38,0x39,0x44,0x45,0x31,0x31,0x41,0x38,0x36,
0x30,0x39,0x39,0x46,0x31,0x32,0x31,0x41,0x32,0x30,
0x34,0x32,0x39,0x22,0x2f,0x3e,0x20,0x3c,0x2f,0x72,
0x64,0x66,0x3a,0x44,0x65,0x73,0x63,0x72,0x69,0x70,
0x74,0x69,0x6f,0x6e,0x3e,0x20,0x3c,0x2f,0x72,0x64,
0x66,0x3a,0x52,0x44,0x46,0x3e,0x20,0x3c,0x2f,0x78,
0x3a,0x78,0x6d,0x70,0x6d,0x65,0x74,0x61,0x3e,0x20,
0x3c,0x3f,0x78,0x70,0x61,0x63,0x6b,0x65,0x74,0x20,
0x65,0x6e,0x64,0x3d,0x22,0x72,0x22,0x3f,0x3e,0xaf,
0x5b,0x6a,0xc5,0x00,0x00,0x01,0x7b,0x49,0x44,0x41,
0x54,0x78,0xda,0xec,0x57,0x41,0x6a,0xc3,0x30,0x10,
0xf4,0x26,0x81,0x5e,0x1a,0x68,0x1f,0xd0,0x06,0x7a,
0x69,0x9e,0x10,0x4a,0x5f,0x92,0x6b,0x21,0x3d,0x95,
0xbe,0xa5,0xe4,0x64,0x41,0xaf,0xfd,0x48,0x20,0xf4,
0x09,0xcd,0xa5,0x90,0xf4,0x01,0xcd,0x21,0x97,0x10,
0xa2,0x8d,0x12,0x6c,0x23,0xb9,0x2b,0x6b,0x25,0xcb,
0xf1,0xa5,0x82,0x05,0x39,0xb6,0x47,0xe3,0xdd,0xd1,
0x68,0x03,0x88,0x98,0xb4,0x39,0x3a,0x49,0xcb,0xa3,
0x75,0x02,0x3d,0xe3,0xea,0x75,0xa8,0x5f,0xa5,0x2a,
0x26,0x2a,0x84,0x8a,0xe7,0x40,0x7c,0x1a,0xe3,0xed,
0xab,0x98,0x82,0xae,0x01,0x7c,0xb9,0x2f,0xa6,0x04,
0x18,0x78,0x2e,0x3e,0x53,0xf1,0x48,0x61,0xc0,0x74,
0x41,0x67,0x40,0xee,0xd1,0xb6,0x78,0xfe,0x3b,0x97,
0x44,0x25,0x46,0xd7,0x56,0x02,0xb9,0x97,0x29,0x03,
0x18,0x02,0x17,0x2f,0xca,0xd2,0xd5,0xca,0x61,0x94,
0x60,0xfb,0x74,0xc7,0xdd,0x93,0x10,0xb8,0xf8,0x69,
0x5c,0xbc,0x7f,0x03,0x9d,0x81,0x1d,0x8a,0x4c,0x34,
0x49,0x40,0x26,0xb8,0xe4,0x85,0xf1,0x25,0x7a,0x06,
0x36,0xe3,0x81,0x0f,0x90,0x9e,0x09,0xaf,0x77,0x2e,
0x3f,0x96,0x16,0x0d,0xec,0x64,0x0e,0x8a,0x35,0xc5,
0xc6,0x2e,0x5d,0x59,0x84,0xfa,0x83,0xb1,0x3d,0x1a,
0x9c,0x46,0x84,0x12,0xcb,0x2f,0x60,0x93,0x8b,0xbb,
0x08,0xc4,0x22,0x01,0x6c,0x2b,0xce,0x8c,0x28,0x26,
0x09,0xa7,0x71,0xb9,0x32,0x50,0x87,0x04,0xcb,0x35,
0xb9,0x04,0xce,0x73,0x1a,0xa2,0x8c,0xb6,0xdd,0xd8,
0x67,0x87,0x49,0x80,0xee,0x8e,0xea,0xa4,0x05,0xbd,
0x44,0x48,0x94,0x20,0x46,0x4d,0x30,0x74,0x1b,0xc6,
0x14,0x04,0xfa,0x1a,0x51,0x13,0x6a,0x44,0xa7,0x15,
0x67,0x12,0x68,0xf2,0x30,0xfa,0x43,0xc2,0x6c,0x4a,
0x25,0xa6,0xc9,0x31,0x0b,0xbc,0x00,0xcb,0xdc,0x15,
0xa9,0xf5,0x38,0xfe,0x79,0xb8,0x3e,0x4b,0x43,0x72,
0x33,0xff,0x05,0x32,0x03,0x8a,0x8c,0x38,0x12,0x72,
0x04,0x04,0xde,0xcb,0xc3,0xde,0x90,0xac,0x46,0x57,
0xae,0xaf,0xa8,0xdb,0x94,0x9e,0x30,0x6e,0x3f,0xd7,
0x95,0x46,0x44,0xf9,0xfe,0x46,0x45,0xdf,0x53,0x9c,
0xac,0xd6,0xbe,0x53,0x01,0x20,0xb4,0x1e,0xae,0x1f,
0xb0,0xed,0xca,0x18,0x64,0xf6,0xe0,0xff,0xcf,0x69,
0xdb,0x04,0x0e,0x02,0x0c,0x00,0x57,0xe6,0x0d,0x82,
0xac,0x8a,0xfc,0x06,0x00,0x00,0x00,0x00,0x49,0x45,
0x4e,0x44,0xae,0x42,0x60,0x82,

  };

  static FXIcon *VFLL12;
  static unsigned char YQRZ_70 []= {
0x89,0x50,0x4e,0x47,0x0d,0x0a,0x1a,0x0a,0x00,0x00,
0x00,0x0d,0x49,0x48,0x44,0x52,0x00,0x00,0x00,0x20,
0x00,0x00,0x00,0x20,0x08,0x06,0x00,0x00,0x00,0x73,
0x7a,0x7a,0xf4,0x00,0x00,0x00,0x19,0x74,0x45,0x58,
0x74,0x53,0x6f,0x66,0x74,0x77,0x61,0x72,0x65,0x00,
0x41,0x64,0x6f,0x62,0x65,0x20,0x49,0x6d,0x61,0x67,
0x65,0x52,0x65,0x61,0x64,0x79,0x71,0xc9,0x65,0x3c,
0x00,0x00,0x03,0xad,0x69,0x54,0x58,0x74,0x58,0x4d,
0x4c,0x3a,0x63,0x6f,0x6d,0x2e,0x61,0x64,0x6f,0x62,
0x65,0x2e,0x78,0x6d,0x70,0x00,0x00,0x00,0x00,0x00,
0x3c,0x3f,0x78,0x70,0x61,0x63,0x6b,0x65,0x74,0x20,
0x62,0x65,0x67,0x69,0x6e,0x3d,0x22,0xef,0xbb,0xbf,
0x22,0x20,0x69,0x64,0x3d,0x22,0x57,0x35,0x4d,0x30,
0x4d,0x70,0x43,0x65,0x68,0x69,0x48,0x7a,0x72,0x65,
0x53,0x7a,0x4e,0x54,0x63,0x7a,0x6b,0x63,0x39,0x64,
0x22,0x3f,0x3e,0x20,0x3c,0x78,0x3a,0x78,0x6d,0x70,
0x6d,0x65,0x74,0x61,0x20,0x78,0x6d,0x6c,0x6e,0x73,
0x3a,0x78,0x3d,0x22,0x61,0x64,0x6f,0x62,0x65,0x3a,
0x6e,0x73,0x3a,0x6d,0x65,0x74,0x61,0x2f,0x22,0x20,
0x78,0x3a,0x78,0x6d,0x70,0x74,0x6b,0x3d,0x22,0x41,
0x64,0x6f,0x62,0x65,0x20,0x58,0x4d,0x50,0x20,0x43,
0x6f,0x72,0x65,0x20,0x35,0x2e,0x30,0x2d,0x63,0x30,
0x36,0x30,0x20,0x36,0x31,0x2e,0x31,0x33,0x34,0x37,
0x37,0x37,0x2c,0x20,0x32,0x30,0x31,0x30,0x2f,0x30,
0x32,0x2f,0x31,0x32,0x2d,0x31,0x37,0x3a,0x33,0x32,
0x3a,0x30,0x30,0x20,0x20,0x20,0x20,0x20,0x20,0x20,
0x20,0x22,0x3e,0x20,0x3c,0x72,0x64,0x66,0x3a,0x52,
0x44,0x46,0x20,0x78,0x6d,0x6c,0x6e,0x73,0x3a,0x72,
0x64,0x66,0x3d,0x22,0x68,0x74,0x74,0x70,0x3a,0x2f,
0x2f,0x77,0x77,0x77,0x2e,0x77,0x33,0x2e,0x6f,0x72,
0x67,0x2f,0x31,0x39,0x39,0x39,0x2f,0x30,0x32,0x2f,
0x32,0x32,0x2d,0x72,0x64,0x66,0x2d,0x73,0x79,0x6e,
0x74,0x61,0x78,0x2d,0x6e,0x73,0x23,0x22,0x3e,0x20,
0x3c,0x72,0x64,0x66,0x3a,0x44,0x65,0x73,0x63,0x72,
0x69,0x70,0x74,0x69,0x6f,0x6e,0x20,0x72,0x64,0x66,
0x3a,0x61,0x62,0x6f,0x75,0x74,0x3d,0x22,0x22,0x20,
0x78,0x6d,0x6c,0x6e,0x73,0x3a,0x78,0x6d,0x70,0x52,
0x69,0x67,0x68,0x74,0x73,0x3d,0x22,0x68,0x74,0x74,
0x70,0x3a,0x2f,0x2f,0x6e,0x73,0x2e,0x61,0x64,0x6f,
0x62,0x65,0x2e,0x63,0x6f,0x6d,0x2f,0x78,0x61,0x70,
0x2f,0x31,0x2e,0x30,0x2f,0x72,0x69,0x67,0x68,0x74,
0x73,0x2f,0x22,0x20,0x78,0x6d,0x6c,0x6e,0x73,0x3a,
0x78,0x6d,0x70,0x4d,0x4d,0x3d,0x22,0x68,0x74,0x74,
0x70,0x3a,0x2f,0x2f,0x6e,0x73,0x2e,0x61,0x64,0x6f,
0x62,0x65,0x2e,0x63,0x6f,0x6d,0x2f,0x78,0x61,0x70,
0x2f,0x31,0x2e,0x30,0x2f,0x6d,0x6d,0x2f,0x22,0x20,
0x78,0x6d,0x6c,0x6e,0x73,0x3a,0x73,0x74,0x52,0x65,
0x66,0x3d,0x22,0x68,0x74,0x74,0x70,0x3a,0x2f,0x2f,
0x6e,0x73,0x2e,0x61,0x64,0x6f,0x62,0x65,0x2e,0x63,
0x6f,0x6d,0x2f,0x78,0x61,0x70,0x2f,0x31,0x2e,0x30,
0x2f,0x73,0x54,0x79,0x70,0x65,0x2f,0x52,0x65,0x73,
0x6f,0x75,0x72,0x63,0x65,0x52,0x65,0x66,0x23,0x22,
0x20,0x78,0x6d,0x6c,0x6e,0x73,0x3a,0x78,0x6d,0x70,
0x3d,0x22,0x68,0x74,0x74,0x70,0x3a,0x2f,0x2f,0x6e,
0x73,0x2e,0x61,0x64,0x6f,0x62,0x65,0x2e,0x63,0x6f,
0x6d,0x2f,0x78,0x61,0x70,0x2f,0x31,0x2e,0x30,0x2f,
0x22,0x20,0x78,0x6d,0x70,0x52,0x69,0x67,0x68,0x74,
0x73,0x3a,0x4d,0x61,0x72,0x6b,0x65,0x64,0x3d,0x22,
0x46,0x61,0x6c,0x73,0x65,0x22,0x20,0x78,0x6d,0x70,
0x4d,0x4d,0x3a,0x4f,0x72,0x69,0x67,0x69,0x6e,0x61,
0x6c,0x44,0x6f,0x63,0x75,0x6d,0x65,0x6e,0x74,0x49,
0x44,0x3d,0x22,0x75,0x75,0x69,0x64,0x3a,0x41,0x45,
0x33,0x46,0x34,0x36,0x30,0x38,0x38,0x46,0x38,0x39,
0x44,0x45,0x31,0x31,0x41,0x38,0x36,0x30,0x39,0x39,
0x46,0x31,0x32,0x31,0x41,0x32,0x30,0x34,0x32,0x39,
0x22,0x20,0x78,0x6d,0x70,0x4d,0x4d,0x3a,0x44,0x6f,
0x63,0x75,0x6d,0x65,0x6e,0x74,0x49,0x44,0x3d,0x22,
0x78,0x6d,0x70,0x2e,0x64,0x69,0x64,0x3a,0x39,0x34,
0x37,0x32,0x33,0x30,0x33,0x46,0x42,0x35,0x45,0x46,
0x31,0x31,0x45,0x30,0x42,0x41,0x41,0x33,0x42,0x42,
0x45,0x34,0x44,0x37,0x42,0x31,0x33,0x35,0x31,0x44,
0x22,0x20,0x78,0x6d,0x70,0x4d,0x4d,0x3a,0x49,0x6e,
0x73,0x74,0x61,0x6e,0x63,0x65,0x49,0x44,0x3d,0x22,
0x78,0x6d,0x70,0x2e,0x69,0x69,0x64,0x3a,0x39,0x34,
0x37,0x32,0x33,0x30,0x33,0x45,0x42,0x35,0x45,0x46,
0x31,0x31,0x45,0x30,0x42,0x41,0x41,0x33,0x42,0x42,
0x45,0x34,0x44,0x37,0x42,0x31,0x33,0x35,0x31,0x44,
0x22,0x20,0x78,0x6d,0x70,0x3a,0x43,0x72,0x65,0x61,
0x74,0x6f,0x72,0x54,0x6f,0x6f,0x6c,0x3d,0x22,0x41,
0x64,0x6f,0x62,0x65,0x20,0x50,0x68,0x6f,0x74,0x6f,
0x73,0x68,0x6f,0x70,0x20,0x43,0x53,0x35,0x20,0x57,
0x69,0x6e,0x64,0x6f,0x77,0x73,0x22,0x3e,0x20,0x3c,
0x78,0x6d,0x70,0x4d,0x4d,0x3a,0x44,0x65,0x72,0x69,
0x76,0x65,0x64,0x46,0x72,0x6f,0x6d,0x20,0x73,0x74,
0x52,0x65,0x66,0x3a,0x69,0x6e,0x73,0x74,0x61,0x6e,
0x63,0x65,0x49,0x44,0x3d,0x22,0x78,0x6d,0x70,0x2e,
0x69,0x69,0x64,0x3a,0x33,0x32,0x45,0x42,0x39,0x34,
0x34,0x44,0x45,0x46,0x42,0x35,0x45,0x30,0x31,0x31,
0x39,0x41,0x42,0x34,0x41,0x46,0x44,0x39,0x30,0x33,
0x39,0x46,0x33,0x35,0x45,0x42,0x22,0x20,0x73,0x74,
0x52,0x65,0x66,0x3a,0x64,0x6f,0x63,0x75,0x6d,0x65,
0x6e,0x74,0x49,0x44,0x3d,0x22,0x75,0x75,0x69,0x64,
0x3a,0x41,0x45,0x33,0x46,0x34,0x36,0x30,0x38,0x38,
0x46,0x38,0x39,0x44,0x45,0x31,0x31,0x41,0x38,0x36,
0x30,0x39,0x39,0x46,0x31,0x32,0x31,0x41,0x32,0x30,
0x34,0x32,0x39,0x22,0x2f,0x3e,0x20,0x3c,0x2f,0x72,
0x64,0x66,0x3a,0x44,0x65,0x73,0x63,0x72,0x69,0x70,
0x74,0x69,0x6f,0x6e,0x3e,0x20,0x3c,0x2f,0x72,0x64,
0x66,0x3a,0x52,0x44,0x46,0x3e,0x20,0x3c,0x2f,0x78,
0x3a,0x78,0x6d,0x70,0x6d,0x65,0x74,0x61,0x3e,0x20,
0x3c,0x3f,0x78,0x70,0x61,0x63,0x6b,0x65,0x74,0x20,
0x65,0x6e,0x64,0x3d,0x22,0x72,0x22,0x3f,0x3e,0x4b,
0x8c,0xae,0xe9,0x00,0x00,0x01,0x6f,0x49,0x44,0x41,
0x54,0x78,0xda,0xec,0x97,0x4b,0x4e,0xc3,0x40,0x10,
0x44,0x7b,0x82,0x25,0x04,0x12,0x62,0x41,0x94,0xec,
0xd8,0x71,0x09,0x2e,0x00,0x57,0x86,0x03,0x70,0x09,
0x4e,0x00,0x1b,0x08,0xf9,0xb8,0x7b,0xbe,0x74,0xdb,
0xf1,0x26,0x8a,0xc8,0xb4,0x32,0xce,0x48,0x80,0x25,
0xcb,0xb2,0x17,0x7e,0xd5,0x35,0x55,0x23,0xdb,0xa4,
0x94,0xa0,0xe6,0x31,0x81,0xca,0x47,0xb3,0xfb,0xe0,
0xe5,0xf5,0x3d,0xcd,0xaf,0x2f,0x8a,0x42,0xde,0x16,
0x2d,0xdc,0xdf,0xcd,0x4c,0x96,0x80,0x19,0xc3,0x6f,
0xa7,0x57,0x45,0x05,0x24,0x8d,0x03,0x92,0x89,0xd2,
0xb9,0xf8,0xe9,0x7d,0xcd,0x3e,0xb9,0xc5,0x83,0x99,
0x34,0x0e,0x1c,0xb0,0xec,0xf7,0x2f,0x41,0x82,0x11,
0x04,0x80,0x2a,0x03,0x69,0x84,0x0c,0x28,0x04,0xc4,
0x24,0x67,0x59,0x01,0x51,0x15,0x42,0x69,0x41,0x3c,
0x99,0x01,0xfb,0x43,0x18,0xab,0x86,0xb0,0x7a,0x0b,
0x6a,0x3b,0xd0,0x85,0x30,0x56,0x0d,0xe1,0xff,0x12,
0xb0,0x80,0x58,0xb6,0x87,0xde,0xfb,0x7a,0x3b,0x61,
0x08,0x41,0x27,0x20,0x16,0x5c,0x02,0x81,0x87,0xe0,
0xc1,0x39,0xab,0xcc,0x40,0x81,0x16,0xc8,0xd4,0x31,
0x06,0x86,0x8b,0x00,0xa7,0xdc,0x8a,0x8f,0x74,0x20,
0x30,0xdc,0xcb,0xf4,0x7c,0xdd,0x20,0xc1,0xe7,0x0a,
0x4f,0xb3,0x04,0xf3,0xe9,0x4d,0x3f,0x3d,0xc3,0x1d,
0xc3,0x89,0x2c,0x9c,0x23,0x42,0xd3,0x9c,0x3d,0xe5,
0x87,0xf0,0x88,0xef,0x81,0xc0,0xed,0x11,0xeb,0xc9,
0x3a,0x68,0x19,0x6c,0xad,0x1d,0xce,0xc7,0x51,0x1d,
0xe8,0x72,0xc3,0x53,0xb7,0x6c,0xf7,0x00,0x96,0xe9,
0x89,0x10,0x10,0xc9,0x20,0xe1,0x78,0x19,0x18,0xe0,
0x62,0xfb,0xd7,0x72,0xb9,0x85,0x93,0x80,0x81,0xc1,
0x86,0x58,0x90,0xdc,0x2b,0xbf,0x8a,0xf3,0xe1,0x52,
0x33,0xef,0xa5,0xeb,0xae,0x0b,0xdd,0x16,0x0c,0x0c,
0x36,0x02,0x26,0x16,0xa4,0x6b,0x41,0x66,0x06,0x76,
0xe1,0x1b,0xb4,0x12,0xff,0xde,0x76,0x22,0x63,0x19,
0x6e,0x5d,0x2f,0x4a,0xf5,0x6b,0xf6,0xb1,0xb6,0x07,
0x05,0xc8,0x56,0x1d,0xbb,0xc0,0x09,0x40,0xea,0xe6,
0x60,0xb1,0x46,0xb8,0x8c,0xab,0x67,0x30,0x93,0x07,
0x27,0x53,0x0b,0x38,0x63,0x10,0xf3,0xe7,0xff,0x8e,
0xbf,0x05,0x18,0x00,0x85,0x16,0x96,0xc3,0x26,0x01,
0xc6,0x25,0x00,0x00,0x00,0x00,0x49,0x45,0x4e,0x44,
0xae,0x42,0x60,0x82,

  };

 

 class VFLL13: public FXMainWindow
 {
  FXDECLARE(VFLL13)
 public:
  enum { ID_FIRST= FXMainWindow::ID_LAST,
         ID_SELF,
         
ID_WORKER,

 YQRZ_4,


 YQRZ_10,


 YQRZ_16,


 YQRZ_22,


 


 


  YQRZ_40,


 YQRZ_47,


 YQRZ_53,


 YQRZ_59,


 


 YQRZ_72,


  YQRZ_78,

 };
  
  long YQRZ_1(FXObject *object,FXSelector sel, void *arg);
 

  long YQRZ_7(FXObject *object,FXSelector sel, void *arg);
  



  long YQRZ_13(FXObject *object,FXSelector sel, void *arg);
  



  long YQRZ_19(FXObject *object,FXSelector sel, void *arg);
  



  long YQRZ_25(FXObject *object,FXSelector sel, void *arg);
  



  
  



  
  


  FXList *playlist;


  long YQRZ_43(FXObject *object,FXSelector sel,void* arg);



  long YQRZ_50(FXObject *object,FXSelector sel, void *arg);
  



  long YQRZ_56(FXObject *object,FXSelector sel, void *arg);
  



  long YQRZ_62(FXObject *object,FXSelector sel, void *arg);
  



  
  FXToggleButton *remove_btn;



  long YQRZ_75(FXObject *object,FXSelector sel, void *arg);
  


   FXFileList *files;
   long YQRZ_81(FXObject *object,FXSelector sel,void *arg);


  VFLL13(FXApp* a,const FXString& name,
         FXIcon *ic,FXIcon *mi,FXuint opts,
         FXint x,FXint y,FXint w,FXint h,
         FXint pl,FXint pr,FXint pt,FXint pb,
         FXint hs,FXint vs);
  virtual ~VFLL13();
  VFLL13() : FXMainWindow() { }
  long handle_close(FXObject *object,FXSelector sel, void *arg);
 };

 FXDEFMAP(VFLL13) YQRZ_87 []= {
  FXMAPFUNC(SEL_CLOSE, VFLL13::ID_SELF, VFLL13::handle_close),
  
  FXMAPFUNC(SEL_IO_READ, VFLL13::ID_WORKER, VFLL13::YQRZ_1),
 
 
    FXMAPFUNC(SEL_COMMAND, VFLL13::YQRZ_4, VFLL13::YQRZ_7),


 
    FXMAPFUNC(SEL_COMMAND, VFLL13::YQRZ_10, VFLL13::YQRZ_13),


 
    FXMAPFUNC(SEL_COMMAND, VFLL13::YQRZ_16, VFLL13::YQRZ_19),


 
    FXMAPFUNC(SEL_COMMAND, VFLL13::YQRZ_22, VFLL13::YQRZ_25),


 


 


  
     FXMAPFUNC(SEL_CLICKED,VFLL13::YQRZ_40,VFLL13::YQRZ_43),


 
    FXMAPFUNC(SEL_COMMAND, VFLL13::YQRZ_47, VFLL13::YQRZ_50),


 
    FXMAPFUNC(SEL_COMMAND, VFLL13::YQRZ_53, VFLL13::YQRZ_56),


 
    FXMAPFUNC(SEL_COMMAND, VFLL13::YQRZ_59, VFLL13::YQRZ_62),


 


 
    FXMAPFUNC(SEL_COMMAND, VFLL13::YQRZ_72, VFLL13::YQRZ_75),



  
    FXMAPFUNC(SEL_COMMAND, VFLL13::YQRZ_78, VFLL13::YQRZ_81), 


 }; 

 FXIMPLEMENT(VFLL13, FXMainWindow, YQRZ_87, ARRAYNUMBER(YQRZ_87))

 // FIXME: the argument names must be changed so that
 // we can have widgets called 'y' etc.
 VFLL13::VFLL13(
         FXApp* a,const FXString& name,
         FXIcon *ic,FXIcon *mi,FXuint opts,
         FXint x,FXint y,FXint w,FXint h,
         FXint pl,FXint pr,FXint pt,FXint pb,
         FXint hs,FXint vs) :
               FXMainWindow(a,name, ic,mi,opts,
                            x,y,w,h, pl,pr,pt,pb, hs,vs)

 {
   setTarget(this);
   setSelector(ID_SELF);
   

    FXApp::instance()->addInput(mbox_fd(dpy_mbox), SEL_IO_READ,
                                                    this, ID_WORKER);
 

  FXVerticalFrame *YQRZ_86= new
    FXVerticalFrame( this,
        LAYOUT_LEFT |
        LAYOUT_TOP |
        LAYOUT_SIDE_TOP |
        LAYOUT_FILL_X |
        LAYOUT_FILL_Y |
        0 |
        0 | 
        0 ,
        0,0,0,0, // x,y,w,h
        DEFAULT_PAD,
        DEFAULT_PAD,
        DEFAULT_PAD,
        DEFAULT_PAD,
        DEFAULT_SPACING,
        DEFAULT_SPACING);


  FXHorizontalFrame *YQRZ_38= new
    FXHorizontalFrame( YQRZ_86,
        LAYOUT_LEFT |
        LAYOUT_TOP |
        LAYOUT_SIDE_TOP |
        LAYOUT_FILL_X |
        0 |
        0 |
        0 | 
        0 ,
        0,0,0,0, // x,y,w,h
        DEFAULT_PAD,
        DEFAULT_PAD,
        DEFAULT_PAD,
        DEFAULT_PAD,
        DEFAULT_SPACING,
        DEFAULT_SPACING);



  FXButton *YQRZ_3;

  YQRZ_3= 
    new FXButton(YQRZ_38,
      "",
      VFLL1,
      this ,
      YQRZ_4 ,

        LAYOUT_LEFT |
        LAYOUT_TOP |
        LAYOUT_SIDE_TOP |
        0 |
        0 |

        FRAME_LINE |
        ICON_BEFORE_TEXT |
        JUSTIFY_CENTER_X |
        JUSTIFY_CENTER_Y 

        

        
        
        ,

        0,0,0,0, // x,y,w,h
        DEFAULT_PAD,
        DEFAULT_PAD,
        DEFAULT_PAD,
        DEFAULT_PAD);

  

  

  

  


   

         



  FXButton *YQRZ_9;

  YQRZ_9= 
    new FXButton(YQRZ_38,
      "",
      VFLL2,
      this ,
      YQRZ_10 ,

        LAYOUT_LEFT |
        LAYOUT_TOP |
        LAYOUT_SIDE_TOP |
        0 |
        0 |

        FRAME_LINE |
        ICON_BEFORE_TEXT |
        JUSTIFY_CENTER_X |
        JUSTIFY_CENTER_Y 

        

        
        
        ,

        0,0,0,0, // x,y,w,h
        DEFAULT_PAD,
        DEFAULT_PAD,
        DEFAULT_PAD,
        DEFAULT_PAD);

  

  

  

  


   

         



  FXButton *YQRZ_15;

  YQRZ_15= 
    new FXButton(YQRZ_38,
      "",
      VFLL3,
      this ,
      YQRZ_16 ,

        LAYOUT_LEFT |
        LAYOUT_TOP |
        LAYOUT_SIDE_TOP |
        0 |
        0 |

        FRAME_LINE |
        ICON_BEFORE_TEXT |
        JUSTIFY_CENTER_X |
        JUSTIFY_CENTER_Y 

        

        
        
        ,

        0,0,0,0, // x,y,w,h
        DEFAULT_PAD,
        DEFAULT_PAD,
        DEFAULT_PAD,
        DEFAULT_PAD);

  

  

  

  


   

         



  FXButton *YQRZ_21;

  YQRZ_21= 
    new FXButton(YQRZ_38,
      "",
      VFLL4,
      this ,
      YQRZ_22 ,

        LAYOUT_LEFT |
        LAYOUT_TOP |
        LAYOUT_SIDE_TOP |
        0 |
        0 |

        FRAME_LINE |
        ICON_BEFORE_TEXT |
        JUSTIFY_CENTER_X |
        JUSTIFY_CENTER_Y 

        

        
        
        ,

        0,0,0,0, // x,y,w,h
        DEFAULT_PAD,
        DEFAULT_PAD,
        DEFAULT_PAD,
        DEFAULT_PAD);

  

  

  

  


   

         



  FXButton *YQRZ_27;

  YQRZ_27= 
    new FXButton(YQRZ_38,
      "",
      VFLL5,
       NULL,
       0,

        LAYOUT_RIGHT |
        LAYOUT_TOP |
        LAYOUT_SIDE_TOP |
        0 |
        0 |

        FRAME_LINE |
        ICON_BEFORE_TEXT |
        JUSTIFY_CENTER_X |
        JUSTIFY_CENTER_Y 

        

        
        
        ,

        0,0,0,0, // x,y,w,h
        DEFAULT_PAD,
        DEFAULT_PAD,
        DEFAULT_PAD,
        DEFAULT_PAD);

  

  

  

  


   

         



  FXButton *YQRZ_33;

  YQRZ_33= 
    new FXButton(YQRZ_38,
      "",
      VFLL6,
       NULL,
       0,

        LAYOUT_RIGHT |
        LAYOUT_TOP |
        LAYOUT_SIDE_TOP |
        0 |
        0 |

        FRAME_LINE |
        ICON_BEFORE_TEXT |
        JUSTIFY_CENTER_X |
        JUSTIFY_CENTER_Y 

        

        
        
        ,

        0,0,0,0, // x,y,w,h
        DEFAULT_PAD,
        DEFAULT_PAD,
        DEFAULT_PAD,
        DEFAULT_PAD);

  

  

  

  


   

         


   FXList *YQRZ_39; 

   YQRZ_39= playlist = new FXList(YQRZ_86,
        this ,
        YQRZ_40 ,

        LAYOUT_LEFT |
        LAYOUT_TOP |
        LAYOUT_SIDE_TOP |
        LAYOUT_FILL_X |
        LAYOUT_FILL_Y 
                 ,0,0,0,0); // xywh


  

  

  

  



  FXHorizontalFrame *YQRZ_76= new
    FXHorizontalFrame( YQRZ_86,
        LAYOUT_LEFT |
        LAYOUT_TOP |
        LAYOUT_SIDE_TOP |
        LAYOUT_FILL_X |
        0 |
        0 |
        0 | 
        0 ,
        0,0,0,0, // x,y,w,h
        DEFAULT_PAD,
        DEFAULT_PAD,
        DEFAULT_PAD,
        DEFAULT_PAD,
        DEFAULT_SPACING,
        DEFAULT_SPACING);



  FXButton *YQRZ_46;

  YQRZ_46= 
    new FXButton(YQRZ_76,
      "",
      VFLL7,
      this ,
      YQRZ_47 ,

        LAYOUT_LEFT |
        LAYOUT_TOP |
        LAYOUT_SIDE_TOP |
        0 |
        0 |

        FRAME_LINE |
        ICON_BEFORE_TEXT |
        JUSTIFY_CENTER_X |
        JUSTIFY_CENTER_Y 

        

        
        
        ,

        0,0,0,0, // x,y,w,h
        DEFAULT_PAD,
        DEFAULT_PAD,
        DEFAULT_PAD,
        DEFAULT_PAD);

  

  

  

  


   

         



  FXButton *YQRZ_52;

  YQRZ_52= 
    new FXButton(YQRZ_76,
      "",
      VFLL8,
      this ,
      YQRZ_53 ,

        LAYOUT_LEFT |
        LAYOUT_TOP |
        LAYOUT_SIDE_TOP |
        0 |
        0 |

        FRAME_LINE |
        ICON_BEFORE_TEXT |
        JUSTIFY_CENTER_X |
        JUSTIFY_CENTER_Y 

        

        
        
        ,

        0,0,0,0, // x,y,w,h
        DEFAULT_PAD,
        DEFAULT_PAD,
        DEFAULT_PAD,
        DEFAULT_PAD);

  

  

  

  


   

         



  FXButton *YQRZ_58;

  YQRZ_58= 
    new FXButton(YQRZ_76,
      "",
      VFLL9,
      this ,
      YQRZ_59 ,

        LAYOUT_LEFT |
        LAYOUT_TOP |
        LAYOUT_SIDE_TOP |
        0 |
        0 |

        FRAME_LINE |
        ICON_BEFORE_TEXT |
        JUSTIFY_CENTER_X |
        JUSTIFY_CENTER_Y 

        

        
        
        ,

        0,0,0,0, // x,y,w,h
        DEFAULT_PAD,
        DEFAULT_PAD,
        DEFAULT_PAD,
        DEFAULT_PAD);

  

  

  

  


   

         



  FXToggleButton *YQRZ_65;

  YQRZ_65= remove_btn =
    new FXToggleButton(YQRZ_76,
      "", "", 
      VFLL10, VFLL11,
       NULL,
       0,

        LAYOUT_RIGHT |
        LAYOUT_TOP |
        LAYOUT_SIDE_TOP |
        0 |
        0 |

        FRAME_LINE |
        ICON_BEFORE_TEXT |
        JUSTIFY_CENTER_X |
        JUSTIFY_CENTER_Y 

        

        
        
        ,

        0,0,0,0, // x,y,w,h
        DEFAULT_PAD,
        DEFAULT_PAD,
        DEFAULT_PAD,
        DEFAULT_PAD);

  

  

  

  


   

         



  FXButton *YQRZ_71;

  YQRZ_71= 
    new FXButton(YQRZ_76,
      "",
      VFLL12,
      this ,
      YQRZ_72 ,

        LAYOUT_RIGHT |
        LAYOUT_TOP |
        LAYOUT_SIDE_TOP |
        0 |
        0 |

        FRAME_LINE |
        ICON_BEFORE_TEXT |
        JUSTIFY_CENTER_X |
        JUSTIFY_CENTER_Y 

        

        
        
        ,

        0,0,0,0, // x,y,w,h
        DEFAULT_PAD,
        DEFAULT_PAD,
        DEFAULT_PAD,
        DEFAULT_PAD);

  

  

  

  


   

         



  FXFileList *YQRZ_77;
  YQRZ_77= files = new FXFileList(YQRZ_86,
           this ,
           YQRZ_78 ,
             FILELIST_NO_OWN_ASSOC |
             0 |
             FILELIST_NO_PARENT |

             LAYOUT_LEFT |
             LAYOUT_TOP |
             LAYOUT_SIDE_TOP |
             LAYOUT_FILL_X |
             LAYOUT_FILL_Y |
             FRAME_RAISED|FRAME_THICK 
           ,0,0,0,0);  // xywh

  

  

  

  



 }
 VFLL13:: ~ VFLL13() { }
 long VFLL13::handle_close
    (FXObject *object,FXSelector sel, void *arg)
 {
     close_player(); 
    return 0;
 } 
 
  long VFLL13::YQRZ_1(FXObject *object,FXSelector sel, void *arg)
  {
     dpy_dispatch((dpy_t*) mbox_get(dpy_mbox)); return 0; 
  }



  long VFLL13::YQRZ_7
    (FXObject *object,FXSelector sel, void *arg)
       {  play_prev();  return 1; }





  long VFLL13::YQRZ_13
    (FXObject *object,FXSelector sel, void *arg)
       {  action_play();  return 1; }





  long VFLL13::YQRZ_19
    (FXObject *object,FXSelector sel, void *arg)
       {  action_pause();  return 1; }





  long VFLL13::YQRZ_25
    (FXObject *object,FXSelector sel, void *arg)
       {  play_next();  return 1; }












 long VFLL13::YQRZ_43
    (FXObject *object,FXSelector sel, void *arg)
    {
      FXList *YQRZ_44= (FXList*)(object);
      int clicked_item= (intptr_t) (arg);
      FYListItem *item= (FYListItem*) (YQRZ_44->getItem(clicked_item));

      int clicked_key = item->data_i;
      
      
      
      
      if (remove_btn->getState()) {
        playlist_remove_file(clicked_key);
      } else {
        play_file(clicked_key);
      }
    
      return 1;
    }






  long VFLL13::YQRZ_50
    (FXObject *object,FXSelector sel, void *arg)
       {  files->setDirectory(get_home_dir());  return 1; }





  long VFLL13::YQRZ_56
    (FXObject *object,FXSelector sel, void *arg)
       { 
        FXString cdir= files->getDirectory();
        char *k= strdup(cdir.text());
        char *Z;
        for(Z=k+strlen(k)-1;Z!=k && Z[0]!='/';Z--)
           ;
        if (Z==k) { files->setDirectory("/"); }
        else {
          Z[0]= 0;
          files->setDirectory(k);
        }
        free(k);
        return 1; }





  long VFLL13::YQRZ_62
    (FXObject *object,FXSelector sel, void *arg)
       { 
        int i;
        for(i=0;i<files->getNumItems();i++)
           if (files->isItemFile(i))
           {
             FXString fn, pa;
             fn= files->getItemFilename(i);
             pa= files->getItemPathname(i);
             playlist_add_file(fn.text(), pa.text());
           }
       return 1; }









  long VFLL13::YQRZ_75
    (FXObject *object,FXSelector sel, void *arg)
       {  playlist_remove_all();  return 1; }




  
long VFLL13::YQRZ_81
     (FXObject *object, FXSelector sel, void *arg)
{
   FXFileList *YQRZ_82= (FXFileList*) object;
   int YQRZ_83= (intptr_t) arg;
   FXString YQRZ_84= YQRZ_82->getItemPathname(YQRZ_83);
   const char *clicked_path=YQRZ_84.text(); 
   FXString YQRZ_85= YQRZ_82->getItemFilename(YQRZ_83);
   const char *clicked_name= YQRZ_85.text();
   if (YQRZ_82->isItemFile(YQRZ_83))
   {
       playlist_add_file(clicked_name, clicked_path); 
      return 1;
   }
   if (YQRZ_82->isItemDirectory(YQRZ_83))
   {
       files->setDirectory(clicked_path); 
      return 1;
   }
   return 1;
}
  





  

 gui_main_t *gui_c_main()
 {
    // ffo -> fox form object
    // ifo -> interface form object 
   VFLL13 *ffo;
   gui_main_t * ifo;
   ffo= new VFLL13(
       FXApp::instance(),
       "BlueHead",
       NULL, NULL, // FIXME: icons
       DECOR_ALL ,
       0, 0,
       400, 400,
       0, 0,
       0, 0,
       0, 0);
   // If the application has not been created, this will done
   // when the application gets created. Otherwise, we have
   // to explicitly do this.
   if (app_created) 
     ffo->create();
   
   ffo->show(PLACEMENT_SCREEN);

   ifo= (typeof(ifo)) malloc(sizeof(*ifo));
   ifo->self= ffo;
    




 



 



 



 



 




  
    ifo->playlist.cls= & gui_intlist_class;
    ifo->playlist.widget= ffo->playlist;
   

 



 



 



 




  
    ifo->files.widget= ffo->files;
    ifo->files.cls= &gui_file_list_class;
  


   return ifo;
 }




static FXApp *app;

void gui_init(int *argc,char **argv)
{
  app= new FXApp("YAK", "KUT");
  app->init(*argc, argv);
  app->create();
  
  VFLL1= new FXPNGIcon(app, YQRZ_2);
  VFLL1->create();

  VFLL2= new FXPNGIcon(app, YQRZ_8);
  VFLL2->create();

  VFLL3= new FXPNGIcon(app, YQRZ_14);
  VFLL3->create();

  VFLL4= new FXPNGIcon(app, YQRZ_20);
  VFLL4->create();

  VFLL5= new FXPNGIcon(app, YQRZ_26);
  VFLL5->create();

  VFLL6= new FXPNGIcon(app, YQRZ_32);
  VFLL6->create();

  VFLL7= new FXPNGIcon(app, YQRZ_45);
  VFLL7->create();

  VFLL8= new FXPNGIcon(app, YQRZ_51);
  VFLL8->create();

  VFLL9= new FXPNGIcon(app, YQRZ_57);
  VFLL9->create();

  VFLL10= new FXPNGIcon(app, YQRZ_63);
  VFLL10->create();

  VFLL11= new FXPNGIcon(app, YQRZ_64);
  VFLL11->create();

  VFLL12= new FXPNGIcon(app, YQRZ_70);
  VFLL12->create();

  app_created= 1;
}

void gui_run()
{
  app->run();
}

void gui_flush()
{
  app->runWhileEvents();
}



